#!/bin/sh

if [ -n "$1" ]; then
    STOP_ON_ERR=$1
else
    STOP_ON_ERR=1
fi

if [ -n "$2" ]; then
    DURATION=$2
fi

if [ -n "$3" ]; then
    SourceIP=$3
fi

if [ -n "$4" ]; then
    DestinationIP=$4
fi

Result=0

mkdir -p /tmp/log
rm -fr /tmp/log/ethernet_stress_result_$SourceIP

#Get local IP list
IP_count=$(ifconfig |grep "inet addr:" | awk 'BEGIN {FS=" +|:"} {print $4}')
   
#Check local IP address and decide remote IP address
for i in ${IP_count[@]}; do
   if [ "${i}" == "192.168.8.101" ]; then
      RemoteIP="192.168.8.100"
      break;
   fi
   if [ "${i}" == "192.168.8.100" ]; then
      RemoteIP="192.168.8.101"
      break;
   fi
done


##########################################
# Make IP and NIC mapping and driver file
##########################################
#local
rm -fr /tmp/ip_dev_map.txt
NIC_count=$(ifconfig |grep "inet addr:" | wc -l)

for ((i=1; i<=$NIC_count; i++)); do
   ifconfig |grep encap | awk '{if(NR=='"${i}"') print $1}' >> /tmp/ip_dev_map.txt
   ifconfig |grep "inet addr:"| awk '{if(NR=='"${i}"') print $2}' >> /tmp/ip_dev_map.txt
   #Get nic driver name
   NIC_name=$(ifconfig |grep encap | awk '{if(NR=='"${i}"') print $1}')
   if [ $NIC_name != "lo" ]; then
      ethtool -i $NIC_name | grep driver | awk '{print $1$2}' >> /tmp/ip_dev_map.txt
   fi
done
sed -i ':a;N;$!ba;s/\naddr:/ /g' /tmp/ip_dev_map.txt
sed -i ':a;N;$!ba;s/\ndriver:/ /g' /tmp/ip_dev_map.txt

Client_NIC=$(grep $SourceIP /tmp/ip_dev_map.txt | awk '{print $1}')
Client_NIC_PCI_BDF=$(ethtool -i $Client_NIC | grep bus-info | awk '{print $2}')

#Remote
rcmd -i $RemoteIP -c "rm -fr /tmp/ip_dev_map.txt"
NIC_count=$(rcmd -i $RemoteIP -c "ifconfig |grep \"inet addr:\" | wc -l" -s)

for ((i=1; i<=$NIC_count; i++)); do
   rcmd -i $RemoteIP -c "ifconfig | grep encap | awk '{if(NR==${i}) print \$1}' >> /tmp/ip_dev_map.txt"
   sleep 0.01
   rcmd -i $RemoteIP -c "ifconfig |grep \"inet addr:\"| awk '{if(NR==${i}) print \$2}' >> /tmp/ip_dev_map.txt"
   sleep 0.01
   #Get nic driver name
   NIC_name=$(rcmd -i $RemoteIP -c "ifconfig |grep encap | awk '{if(NR=='"${i}"') print \$1}'" -s)
   sleep 0.01
   if [ $NIC_name != "lo" ]; then
      rcmd -i $RemoteIP -c "ethtool -i $NIC_name | grep driver | awk '{print \$1\$2}' >> /tmp/ip_dev_map.txt"
      sleep 0.01
   fi
done
rcmd -i $RemoteIP -c "sed -i ':a;N;\$!ba;s/\naddr:/ /g' /tmp/ip_dev_map.txt"
rcmd -i $RemoteIP -c "sed -i ':a;N;\$!ba;s/\ndriver:/ /g' /tmp/ip_dev_map.txt"

Server_NIC=$(rcmd -i $RemoteIP -c "grep $DestinationIP /tmp/ip_dev_map.txt | awk '{print \$1}'" -s)
Server_NIC_PCI_BDF=$(rcmd -i $RemoteIP -c "ethtool -i $Server_NIC | grep bus-info | awk '{print \$2}'" -s)



############################
#Check ethernet statistics
############################
#Client side
Statistics_line_count=$(ethtool -S $Client_NIC | grep error |wc -l)
for ((i=1; i<=$Statistics_line_count; i++)); do
   Statistics_item=$(ethtool -S $Client_NIC | grep error | awk '{if(NR=='$i') print $1}')
   Statistics_value=$(ethtool -S $Client_NIC | grep error | awk '{if(NR=='$i') print $2}')

   if [ "$Statistics_value" != 0 ]; then
      Result=1
      break
   fi
done

if [ "$Result" -eq 1 ]; then
   #######################
   #Clean nic statistics
   #######################
   #Client side
   NIC_driver=$(grep $SourceIP /tmp/ip_dev_map.txt | awk '{print $3}')
   NIC_driver_count=$(grep $NIC_driver /tmp/ip_dev_map.txt | wc -l)
   reset_command="modprobe -r $NIC_driver; modprobe $NIC_driver; "

   for ((i=1; i<=$NIC_driver_count; i++)); do
      ifconfig_string=$(grep $NIC_driver /tmp/ip_dev_map.txt |awk '{if(NR=='$i')print $1" "$2}')
      reset_command=$reset_command"ifconfig "$ifconfig_string"; "
   done
   eval $reset_command
fi
Result=0


#Server side
Statistics_line_count=$(rcmd -i $RemoteIP -c "ethtool -S $Server_NIC | grep error |wc -l" -s)
for ((i=1; i<=$Statistics_line_count; i++)); do
   Statistics_item=$(rcmd -i $RemoteIP -c "ethtool -S $Server_NIC | grep error | awk '{if(NR=='$i') print \$1}'" -s)
   Statistics_value=$(rcmd -i $RemoteIP -c "ethtool -S $Server_NIC | grep error | awk '{if(NR=='$i') print \$2}'" -s)

   if [ "$Statistics_value" != 0 ]; then
      Result=1
      break
   fi
done

if [ "$Result" -eq 1 ]; then
   #######################
   #Clean nic statistics
   #######################
   #Server side
   NIC_driver=$(rcmd -i $RemoteIP -c "grep $DestinationIP /tmp/ip_dev_map.txt | awk '{print \$3}'" -s)
   NIC_driver_count=$(rcmd -i $RemoteIP -c "grep $NIC_driver /tmp/ip_dev_map.txt | wc -l" -s)
   reset_command="modprobe -r $NIC_driver; modprobe $NIC_driver; "

   for ((i=1; i<=$NIC_driver_count; i++)); do
      ifconfig_string=$(rcmd -i $RemoteIP -c "grep $NIC_driver /tmp/ip_dev_map.txt |awk '{if(NR==${i})print \$1\" \"\$2}'" -s)
      reset_command=$reset_command"ifconfig "$ifconfig_string"; "
   done
   rcmd -i $RemoteIP -c "$reset_command"

   #wait for driver reinsert ready
   sleep 5
fi
Result=0


#######################################
#Check remote iperf execute or not
#######################################
PS=$(rcmd -i $RemoteIP -c "ps | grep \"iperf -s -B $DestinationIP\"| head -n 1" -s | awk '{print $1"\t"$3"\t"$4"\t"$5"\t"$6"\t"$7"\t"$8}')

#splite the result to array
IFS=' ' read -r -a PS_ARRAY <<< ${PS}
PID=${PS_ARRAY[0]}
COMMAND=${PS_ARRAY[3]}
IperfServerIP=${PS_ARRAY[6]}

if [ "${COMMAND}" != "iperf" ]; then
   rcmd -i $RemoteIP -c "mkdir -p /tmp/log"

   #print the remote start time
   now="[$(date '+%D %T')]"
   rcmd -i $RemoteIP -c "printf \"$now Start ethernet stress\n\" >> /tmp/log/ethernet_stress_server_$DestinationIP.log"

   #Star iperf ethernet server
   rcmd -i $RemoteIP -c "iperf -s -B $DestinationIP >> /tmp/log/ethernet_stress_server_$DestinationIP.log  2>&1 &"
   sleep 1
fi


rm -f /tmp/log/ethernet_stress_result_$SourceIP
now="[$(date '+%D %T')]"
printf "$now Start ethernet stress\n" >> /tmp/log/ethernet_stress_client_$SourceIP.log
printf "$now Start ethernet stress\n" >> /tmp/log/ethernet_stress_result_$SourceIP


#########################
#Get stress log length
#########################
Client_log_length=$(wc -l /tmp/log/ethernet_stress_client_$SourceIP.log | awk '{print $1}')
Server_log_length=$(rcmd -i $RemoteIP -c "wc -l /tmp/log/ethernet_stress_server_$DestinationIP.log | awk '{print \$1}'" -s)


#####################
#Start client iperf
#####################
iperf -c $DestinationIP -t 604800 -P 8 >> /tmp/log/ethernet_stress_client_$SourceIP.log 2>&1 &
#iperf -c $DestinationIP -t 604800 >> /tmp/log/ethernet_stress_current.log &
#PS=$(ps |grep "iperf -c $DestinationIP" | head -n 1)
PS=$(ps |grep "iperf -c $DestinationIP" | grep -v "grep")
IFS=' ' read -r -a PS_ARRAY <<< ${PS}
ClientPID=${PS_ARRAY[0]}


#################
# Clear PCI AER
#################
lemur pci error clear $Client_NIC_PCI_BDF > /dev/null
rcmd -i $RemoteIP -c "lemur pci error clear $Server_NIC_PCI_BDF"



######################################################
#Start pulling the flag and check duration and error
######################################################
START="$(date +%s)"

while [ 1 ]; do

   #Update current time
   CURRENT="$(date +%s)"
   TIME_DIFF=$(( $CURRENT - $START ))

   ###############################
   #Check /tmp/stop_flag exist
   ###############################
   if [ -f /tmp/stop_flag ]; then
      break;
   fi


   ############################
   #Check ethernet statistics
   ############################
   #Client side
   Statistics_line_count=$(ethtool -S $Client_NIC | grep error |wc -l)
   for ((i=1; i<=$Statistics_line_count; i++)); do
      Statistics_item=$(ethtool -S $Client_NIC | grep error | awk '{if(NR=='$i') print $1}')
      Statistics_value=$(ethtool -S $Client_NIC | grep error | awk '{if(NR=='$i') print $2}')
      
      if [ "$Statistics_value" != 0 ]; then
         Result=1
         now="[$(date '+%D %T')]"
         printf "$now Ethernet client statistics error: $Statistics_item $Statistics_value\n" >> /tmp/log/ethernet_stress_client_$SourceIP.log
         if [ "$STOP_ON_ERR" -eq 1 ]; then
            break;
         fi
      fi
   done
   
   #Server side
   Statistics_line_count=$(rcmd -i $RemoteIP -c "ethtool -S $Server_NIC | grep error |wc -l" -s)
   for ((i=1; i<=$Statistics_line_count; i++)); do
      Statistics_item=$(rcmd -i $RemoteIP -c "ethtool -S $Server_NIC | grep error | awk '{if(NR=='$i') print \$1}'" -s)
      Statistics_value=$(rcmd -i $RemoteIP -c "ethtool -S $Server_NIC | grep error | awk '{if(NR=='$i') print \$2}'" -s)

      if [ "$Statistics_value" != 0 ]; then
         Result=1
         now="[$(date '+%D %T')]"
         printf "$now Ethernet Server statistics error: $Statistics_item $Statistics_value\n" >> /tmp/log/ethernet_stress_client_$SourceIP.log
         if [ "$STOP_ON_ERR" -eq 1 ]; then
            break;
         fi
      fi
   done


   ###################
   #Check PCI AER
   ###################
   #Client side
   Client_AER_Uncorrectable=$(lemur pci error aer $Client_NIC_PCI_BDF | grep "Uncorrectable" | awk '{print $2}')
   Client_AER_Correctable=$(lemur pci error aer $Client_NIC_PCI_BDF | grep "Correctable" | awk '{print $2}')

   if [ "$Client_AER_Uncorrectable" != "0x00000000" ]; then
      Result=1
      now="[$(date '+%D %T')]"
      printf "$now Client PCI AER Uncorrectable erroe: $Client_AER_Uncorrectable\n" >> /tmp/log/ethernet_stress_client_$SourceIP.log
      #Stop on error check
      if [ "$STOP_ON_ERR" -eq 1 ]; then
         break;
      fi
   fi
   if [ "$Client_AER_Correctable" != "0x00000000" ]; then
      Result=1
      now="[$(date '+%D %T')]"
      printf "$now Client PCI AER Correctable erroe: $Client_AER_Correctable\n" >> /tmp/log/ethernet_stress_client_$SourceIP.log
      #Stop on error check
      if [ "$STOP_ON_ERR" -eq 1 ]; then
         break;
      fi
   fi
   #Server side
   Server_AER_Uncorrectable=$(rcmd -i $RemoteIP -c "lemur pci error aer $Server_NIC_PCI_BDF | grep "Uncorrectable" | awk '{print \$2}'" -s)
   Server_AER_Correctable=$(rcmd -i $RemoteIP -c "lemur pci error aer $Server_NIC_PCI_BDF | grep "Correctable" | awk '{print \$2}'" -s)

   if [ "$Server_AER_Uncorrectable" != "0x00000000" ]; then
      Result=1
      now="[$(date '+%D %T')]"
      printf "$now Server PCI AER Uncorrectable erroe: $Server_AER_Uncorrectable\n" >> /tmp/log/ethernet_stress_client_$SourceIP.log
      #Stop on error check
      if [ "$STOP_ON_ERR" -eq 1 ]; then
         break;
      fi
   fi
   if [ "$Server_AER_Correctable" != "0x00000000" ]; then
      Result=1
      now="[$(date '+%D %T')]"
      printf "$now Server PCI AER Correctable erroe: $Server_AER_Correctable\n" >> /tmp/log/ethernet_stress_client_$SourceIP.log
      #Stop on error check
      if [ "$STOP_ON_ERR" -eq 1 ]; then
         break;
      fi
   fi


   ###########################
   #Check iperf report error
   ###########################
   #Client side
   iperf_error=$(tail -n +$Client_log_length /tmp/log/ethernet_stress_client_$SourceIP.log | grep -E "error|FATAL|fail")
   if [ "$iperf_error" != "" ]; then
      Result=1
      now="[$(date '+%D %T')]"
      printf "$now iperf client error\n" >> /tmp/log/ethernet_stress_client_$SourceIP_$DestinationIP.log
      if [ "$STOP_ON_ERR" -eq 1 ]; then
         break;
      fi
   fi
   #Server side
   iperf_error=$(rcmd -i $RemoteIP -c "tail -n +$Server_log_length /tmp/log/ethernet_stress_server_$DestinationIP.log | grep -E \"error|FATAL|fail\"" -s)
   if [ "$iperf_error" != "" ]; then
      Result=1
      now="[$(date '+%D %T')]"
      printf "$now iperf server error\n" >> /tmp/log/ethernet_stress_client_$SourceIP.log
      if [ "$STOP_ON_ERR" -eq 1 ]; then
         break;
      fi
   fi


   #################
   #Check duration
   #################
   if [ "$DURATION" -ne 0 ]; then
      if [ $TIME_DIFF -ge $DURATION ]; then
         break;
      fi
   fi

done


now="[$(date '+%D %T')]"
##########################
#Check and repport result
##########################
if [ "$Result" -eq 1 ]; then
   echo "$now FAILD" >> /tmp/log/ethernet_stress_result_$SourceIP
   touch /tmp/stop_flag
else
   echo "$now PASS" >> /tmp/log/ethernet_stress_result_$SourceIP
fi


#Check remote iperf execute or not
#PS=$(rcmd -i $RemoteIP -c "ps | grep \"iperf -s -B $DestinationIP\"| head -n 1" -s | awk '{print $1"\t"$3"\t"$4"\t"$5"\t"$6"\t"$7"\t"$8}')
PS=$(rcmd -i $RemoteIP -c "ps | grep \"iperf -s -B $DestinationIP\"| grep -v 'grep'" -s | awk '{print $1"\t"$3"\t"$4"\t"$5"\t"$6"\t"$7"\t"$8}')


#splite the result to array
IFS=' ' read -r -a PS_ARRAY <<< ${PS}
PID=${PS_ARRAY[0]}

#Kill client/server iperf
kill $ClientPID > /dev/null 2>&1
rcmd -i $RemoteIP -c "kill -9 $PID"

sleep 1

#print the client end time
now="[$(date '+%D %T')]"
printf "$now Stop ethernet\n" >> /tmp/log/ethernet_stress_client_$SourceIP.log

#print the remote end time
now="[$(date '+%D %T')]"
rcmd -i $RemoteIP -c "printf \"$now Stop ethernet stress\n\" >> /tmp/log/ethernet_stress_server_$DestinationIP.log"

rm -fr /tmp/ip_dev_map.txt
rcmd -i $RemoteIP -c "rm /tmp/ip_dev_map.txt"
