#include "CPUOperation.hpp"



CPUOperation::CPUOperation(OPERATION_TYPE oprationType, string helpContent) : OperationBaseClass(oprationType, helpContent)
{

}



CPUOperation::~CPUOperation()
{

}



int
CPUOperation::executeOperation(StringParser command_line)
{
	  string parameter_1;
	  int    err = 0;

    command_line.getParameter(&parameter_1, 0);

    //If the first parameter is "help" or space, show help message
    if (parameter_1 == HELP || parameter_1 == "")
    {
        Help_parameter(command_line);
        err = -1;
        return err;
    }

    //Execute operation
    switch (getOperationType())
    {
        case OPERATION_ALL:
        case OPERATION_SHOW_CPU:
        case OPERATION_SHOW_VENDOR:
        case OPERATION_SHOW_MODEL:
        case OPERATION_SHOW_FAMILY:
        case OPERATION_SHOW_CORE:
        case OPERATION_SHOW_FREQUENCY:
        case OPERATION_SHOW_STEPPING:
        case OPERATION_SHOW_CACHE:
        case OPERATION_SHOW_TEMPERATUR:
            err = show(command_line);
            break;
        case OPERATION_STABILITY:
            err = stability(command_line);
            break;
        case OPERATION_READ:
            err = read(command_line);
            break;
        case OPERATION_WRITE:
            err = write(command_line);
            break;
        case OPERATION_UTILITY_GETMCE:
            err = getMCE(command_line);
            break;
        default:
            cout << "Unsupport operation" << endl;
            err = -1; 
            break;
    }

    return err;
}



int
CPUOperation::Help_parameter(StringParser command_line)
{
    string module, command, operation;

    command_line.getModule(&module);
    command_line.getCommand(&command);
    getName(&operation);

    cout << "Usage: lemur " << module << " " << command << " " << operation << " ";
    //Show parameter help message depend on operation type
    switch (getOperationType())
    {
        case OPERATION_ALL:
            cout << "P1" << endl;
            cout << "Displays CPU all information" << endl;
            cout << "Syntax: show all P1" << endl;
            cout << " P1: Device name" << endl;
            break;
        case OPERATION_SHOW_CPU:
            cout << "P1" << endl;
            cout << "Displays CPU name" << endl;
            cout << "Syntax: show cpu P1" << endl;
            cout << " P1: Device name" << endl;
            break;
        case OPERATION_SHOW_VENDOR:
            cout << "P1" << endl;
            cout << "Displays CPU vendor name" << endl;
            cout << "Syntax: show vendor P1" << endl;
            cout << " P1: Device name" << endl;
            break;
        case OPERATION_SHOW_MODEL:
            cout << "P1" << endl;
            cout << "Displays CPU model type" << endl;
            cout << "Syntax: show model P1" << endl;
            cout << " P1: Device name" << endl;
            break;
        case OPERATION_SHOW_FAMILY:
            cout << "P1" << endl;
            cout << "Displays CPU family" << endl;
            cout << "Syntax: show family P1" << endl;
            cout << " P1: Device name" << endl;
            break;
        case OPERATION_SHOW_CORE:
            cout << "P1" << endl;
            cout << "Displays CPU core number" << endl;
            cout << "Syntax: show core P1" << endl;
            cout << " P1: Device name" << endl;
            break;
        case OPERATION_SHOW_FREQUENCY:
            cout << "P1" << endl;
            cout << "Displays CPU frequency" << endl;
            cout << "Syntax: show frequency P1" << endl;
            cout << " P1: Device name" << endl;
            break;
        case OPERATION_SHOW_STEPPING:
            cout << "P1" << endl;
            cout << "Displays CPU stepping" << endl;
            cout << "Syntax: show stepping P1" << endl;
            cout << " P1: Device name" << endl;
            break;
        case OPERATION_SHOW_CACHE:
            cout << "P1" << endl;
            cout << "Displays CPU cache size" << endl;
            cout << "Syntax: show cache P1" << endl;
            cout << " P1: Device name" << endl;
            break;
        case OPERATION_SHOW_TEMPERATUR:
            cout << "P1" << endl;
            cout << "Displays CPU temperature" << endl;
            cout << "Syntax: show temperature P1" << endl;
            cout << " P1: Device name" << endl;
            break;
        case OPERATION_STABILITY:
            cout << "P1 P2" << endl;
            cout << "Calculating pi value and verified the value by checksum" << endl;
            cout << "Syntax: stability P1 P2" << endl;
            cout << " P1: PI calculate level(Dec: 1~9)" << endl;
            cout << " P2: Device name" << endl;
            break;
        case OPERATION_READ:
            cout << "P1 P2 P3" << endl;
            cout << "Read register operation of MSR(x86)" << endl;
            cout << "Syntax: read P1 P2 P3" << endl;
            cout << " P1: Register address(Hex)" << endl;
            cout << " P2: Register count" << endl;
            cout << " P3: Device name" << endl;
            break;
        case OPERATION_WRITE:
            cout << "P1 P2 P3" << endl;
            cout << "Write register operation of MSR(x86)" << endl;
            cout << "Syntax: write P1 P2 P3" << endl;
            cout << " P1: Register address(Hex)" << endl;
            cout << " P2: Data(Hex)" << endl;
            cout << " P3: Device name" << endl;
            break;
        case OPERATION_UTILITY_GETMCE:
            cout << "P1" << endl;
            cout << "Get the machine check exception message" << endl;
            cout << "Syntax: getmce P1" << endl;
            cout << " P1: Device name" << endl;
            break;
       default:
            cout << "\nNot define parameter help message" << endl;
            break;
    }

    return 0;
}



int
CPUOperation::show(StringParser command_line)
{
    int            err = 0;
    ErrorMsg       *error_code;
    string         WellKnownName, Operation, result;
    ProcessorDAPI  *pCPUDAPI = NULL;
    CPU_ITEM       option;

    error_code = NO_ERROR_MSG;

    err = Syntax_check(command_line);

    if (err == 0)
    {
        //Get I2C device name
        command_line.getParameter(&WellKnownName, 0);
        //cout << "Name: " <<  WellKnownName << endl;

        //Get show operation
        command_line.getOperation(&Operation);
        //cout << "Operation: " << Operation << endl;

        if (Operation == "all")
        {
            option = CPU_ITEM_ALL;
        }
        else if (Operation == "cpu")
        {
            option = CPU_ITEM_CPU;
        }
        else if (Operation == "vendor")
        {
            option = CPU_ITEM_VENDOR;
        }
        else if (Operation == "model")
        {
            option = CPU_ITEM_MODEL;
        }
        else if (Operation == "family")
        {
            option = CPU_ITEM_FAMILY;
        }
        else if (Operation == "core")
        {
            option = CPU_ITEM_CORE;
        }
        else if (Operation == "frequency")
        {
            option = CPU_ITEM_FREQUENCY;
        }
        else if (Operation == "stepping")
        {
            option = CPU_ITEM_STEPPING;
        }
        else if (Operation == "cache")
        {
            option = CPU_ITEM_CACHE;
        }
        else if (Operation == "temperature")
        {
            option = CPU_ITEM_TEMPERATURE;
        }
        else
        {
            cout << "Not supported operation" << endl;
            return -1;
        }


        //Creat CPU DAPI object
        pCPUDAPI = new ProcessorDAPI("CPU_DAPI");

        //Get CPU information by API
        error_code = pCPUDAPI->GetInformation(WellKnownName, option, &result);

        if (error_code != NO_ERROR_MSG)
        {
            printError(error_code);
            err = -1;
        }

        cout << result << endl;

        delete pCPUDAPI;
    }

    return err;
}



int
CPUOperation::stability(StringParser command_line)
{
    int                 err = 0;
    int                 loop;
    ProcessorDAPI       *pCPUDAPI = NULL;
    ErrorMsg            *error_code;
    string              WellKnownName, temp_string, error_massage;
    int                 result = 0;
    StabilityTestObject st("Pi test");

    error_code = NO_ERROR_MSG;

    err = Syntax_check(command_line);

    if (err == 0)
    {
        //Get CPU device name
        command_line.getParameter(&WellKnownName, 1);
        //cout << "Name: " <<  WellKnownName << endl;

        //Get Start address
        command_line.getParameter(&temp_string, 0);
        command_line.convertStringToDec(temp_string, &loop);
        //cout << "Start: " << StartAddress << endl;

        //Switch level to loop
        switch (loop)
        {
            case 1:
                loop = 16;
                break;
            case 2:
                loop = 17;
                break;
            case 3:
                loop = 18;
                break;
            case 4:
                loop = 19;
                break;
            case 5:
                loop = 20;
                break;
            case 6:
                loop = 21;
                break;
            case 7:
                loop = 22;
                break;
            case 8:
                loop = 23;
                break;
            case 9:
                loop = 24;
                break;
        }

        //Creat CPU DAPI object
        pCPUDAPI = new ProcessorDAPI("CPU_DAPI");

        pCPUDAPI->SetExecutionCPUCore(WellKnownName);

        //Start pi calculate by API
        error_code = st.PiCalculate(&result, 0, loop);

        if (error_code != NO_ERROR_MSG)
        {
            printError(error_code);
            err = -1;
        }
        delete pCPUDAPI;
    }

    return err;
}



int
CPUOperation::read(StringParser command_line)
{
    int      err = 0;
    int      RegisterAddress, Count, index = 0;
    list<unsigned long long> val;
    ProcessorDAPI  *pCPUDAPI = NULL;
    ErrorMsg *error_code;
    string   WellKnownName, temp_string, error_massage;

    error_code = NO_ERROR_MSG;

    err = Syntax_check(command_line);

    if (err == 0)
    {
        //Get CPU device name
        command_line.getParameter(&WellKnownName, 2);
        //cout << "Name: " <<  WellKnownName << endl;

        //Get Start register address
        command_line.getParameter(&temp_string, 0);
        command_line.convertStringToHex(temp_string, &RegisterAddress);
        //cout << "Start: " << RegisterAddress << endl;

        //Get register count
        command_line.getParameter(&temp_string, 1);
        command_line.convertStringToHex(temp_string, &Count);
        //cout << "Count: " << Count << endl;

        //Creat CPU DAPI object
        pCPUDAPI = new ProcessorDAPI("CPU_DAPI");

        //Read CPU register value by API
        error_code = pCPUDAPI->Read(WellKnownName, RegisterAddress, Count, &val);

        if (error_code != NO_ERROR_MSG)
        {
            printError(error_code);
            err = -1;
        }
        else
        {
            while (!val.empty())
            {
                printf("0x%x: %06lx\n",RegisterAddress + index , (long unsigned int)val.front());
                val.pop_front();
                index++;
            }
        }
        delete pCPUDAPI;
    }

    return err;
}



int
CPUOperation::write(StringParser command_line)
{
    int      err = 0;
    int      RegisterAddress, index = 0;
    unsigned long long Data;
    list<unsigned long long> val;
    ProcessorDAPI  *pCPUDAPI = NULL;
    ErrorMsg *error_code;
    string   WellKnownName, temp_string, error_massage;


    error_code = NO_ERROR_MSG;

    err = Syntax_check(command_line);

    if (err == 0)
    {
        //Get CPU device name
        command_line.getParameter(&WellKnownName, 2);
        //cout << "Name: " <<  WellKnownName << endl;

        //Get Start register address
        command_line.getParameter(&temp_string, 0);
        command_line.convertStringToHex(temp_string, &RegisterAddress);
        //cout << "Start: " << RegisterAddress << endl;

        //Get write data
        command_line.getParameter(&temp_string, 1);
        command_line.convertStringToHex(temp_string, &Data);
        //cout << "Data: " << Data << endl;

        //Creat CPU DAPI object
        pCPUDAPI = new ProcessorDAPI("CPU_DAPI");

        //Write data to CPU register by API
        error_code = pCPUDAPI->Write(WellKnownName, RegisterAddress, Data);

        if (error_code != NO_ERROR_MSG)
        {
            printError(error_code);
            err = -1;
        }
        else
        {
            error_code = pCPUDAPI->Read(WellKnownName, RegisterAddress, 1, &val);
            if (error_code != NO_ERROR_MSG)
            {
                printError(error_code);
                err = -1;
            }
            else
            {
                printf("0x%x: %06lx\n",RegisterAddress + index , (long unsigned int)val.front());
            }
        }
        delete pCPUDAPI;
    }

    return err;
}



int
CPUOperation::getMCE(StringParser command_line)
{
    int           err = 0;
    ProcessorDAPI *pCPUDAPI = NULL;
    ErrorMsg      *error_code;
    string        WellKnownName, MCE_message, error_massage;


    error_code = NO_ERROR_MSG;

    err = Syntax_check(command_line);

    if (err == 0)
    {
        //Get CPU device name
        command_line.getParameter(&WellKnownName, 0);
        //cout << "Name: " <<  WellKnownName << endl;


        //Creat CPU DAPI object
        pCPUDAPI = new ProcessorDAPI("CPU_DAPI");

        //Get MCE message by API
        error_code = pCPUDAPI->GetMCE(WellKnownName, &MCE_message);

        if (error_code != NO_ERROR_MSG)
        {
            printError(error_code);
            err = -1;
        }
        else
        {
            printf("%s\n", MCE_message.c_str());
        }
        delete pCPUDAPI;
    }

    return err;
}



int
CPUOperation::Syntax_check(StringParser command_line)
{
    int    i, parameter_number = 0;
    string parameter[7];
    int    err = 0;
    string operation_name;

    getName(&operation_name);

    command_line.getParameterNumber(&parameter_number);

    //Get parameter
    for (i=0; i<parameter_number; i++)
    {
        command_line.getParameter(&parameter[i], i);
    }

    switch (getOperationType())
    {
        case OPERATION_STABILITY:

            //Check parameter number
            if (parameter_number != 2)
            {
                cout << "Parameter numbers isn't 2" << endl;
                return -1;
            }

            //Check parameter 1 is numeric
            err = command_line.isHEXString(parameter[0]);
            if (err != 0)
            {
                cout << "parameter 1 isn't numeric." << endl;
                return -1;
            }
            break;
        case OPERATION_ALL:
        case OPERATION_SHOW_CPU:
        case OPERATION_SHOW_VENDOR:
        case OPERATION_SHOW_MODEL:
        case OPERATION_SHOW_FAMILY:
        case OPERATION_SHOW_CORE:
        case OPERATION_SHOW_FREQUENCY:
        case OPERATION_SHOW_STEPPING:
        case OPERATION_SHOW_CACHE:
        case OPERATION_SHOW_TEMPERATUR:
            //Check parameter number
            if (parameter_number != 1)
            {
                cout << "Parameter numbers isn't 1" << endl;
                return -1;
            }
            break;
        case OPERATION_READ:
        case OPERATION_WRITE:
            //Check parameter number
            if (parameter_number != 3)
            {
                cout << "Parameter numbers isn't 3" << endl;
                return -1;
            }

            //Check parameter 1 is numeric
            err = command_line.isHEXString(parameter[0]);
            if (err != 0)
            {
                cout << "parameter 1 isn't numeric." << endl;
                return -1;
            }
            //Check parameter 2 is numeric
            err = command_line.isHEXString(parameter[1]);
            if (err != 0)
            {
                cout << "parameter 2 isn't numeric." << endl;
                return -1;
            }
            break;
        case OPERATION_UTILITY_GETMCE:
            //Check parameter number
            if (parameter_number != 1)
            {
                cout << "Parameter numbers isn't 1" << endl;
                return -1;
            }
            break;
       default:
            cout << "Not define syntax_check for " << operation_name << endl;
            err = -1;
            break;
    }

    return err;
}
