#include "EthernetCommand.hpp"

EthernetCommand::EthernetCommand(COMMAND_TYPE CommandType, string HelpContent) : CommandBaseClass(CommandType, HelpContent)
{
    //Base on command type to add different operations
    switch (CommandType)
    {
        case CMD_SHOW:
            addOperation(new EthernetOperation(OPERATION_ALL, "Shows ethernet device all information"));
            addOperation(new EthernetOperation(OPERATION_NODE, "Shows ethernet node name in system"));
            addOperation(new EthernetOperation(OPERATION_LINK, "Shows ethernet link status"));
            addOperation(new EthernetOperation(OPERATION_SHOW_SPEED, "Shows ethernet current link speed"));
            addOperation(new EthernetOperation(OPERATION_FW, "Shows ethernet F.W version"));
            break;
        case CMD_TEST:
            addOperation(new EthernetOperation(OPERATION_MDI, "Check the link is up or not"));
            addOperation(new EthernetOperation(OPERATION_INTERRUPT, "Test the NIC interrupt work status"));
            addOperation(new EthernetOperation(OPERATION_INTERNAL_LOOPBACK_TEST, "Test the NIC internal loopback link"));
            addOperation(new EthernetOperation(OPERATION_SINGLE_LOOPBACK_TEST, "Test the NIC single loopback link"));
            addOperation(new EthernetOperation(OPERATION_CROSS_LOOPBACK_TEST, "Test the NIC cross loopback link"));
            break;
        case CMD_MAC:
            addOperation(new EthernetOperation(OPERATION_READ_MAC, "Read a MAC address from Etherent"));
            addOperation(new EthernetOperation(OPERATION_WRITE_MAC, "Write a MAC address to Ethernet"));
            break;
        case CMD_EEPROM:
            addOperation(new EthernetOperation(OPERATION_READ_EEPROM, "Read data from EEPROM"));
            addOperation(new EthernetOperation(OPERATION_WRITE_EEPROM, "Write data to EEPROM"));
            addOperation(new EthernetOperation(OPERATION_VALIDATE_EEPROM, "Validate the content by calculate checksum"));
            addOperation(new EthernetOperation(OPERATION_PROGRAM, "Program EEPROM Image"));
            break;
        case CMD_SFP:
            addOperation(new EthernetOperation(OPERATION_READ_SFP, "Read data from SFP"));
            addOperation(new EthernetOperation(OPERATION_WRITE_SFP, "Write data to SFP"));
            break;
        case CMD_ERROR:
            addOperation(new EthernetOperation(OPERATION_ERROR_ALL, "show all error (statistics/pcie)"));
            addOperation(new EthernetOperation(OPERATION_ERROR_STATISTICS, "show component error statistics"));
            addOperation(new EthernetOperation(OPERATION_ERROR_PCI, "show pci error include PCIe/PCI-Compatible/AER"));
            addOperation(new EthernetOperation(OPERATION_CLEAR, "Clear all error statistics and PCI error count"));
            break;
        //Add NEW command and operation here
        default:
            break;
    }
}



EthernetCommand::~EthernetCommand()
{

}



int
EthernetCommand::List(StringParser command_line)
{
    EthernetDAPI  *pEthernetDAPI = NULL;
    ErrorMsg      *error_code = NULL;
    list<struct ethernet_data> eth_list;

    pEthernetDAPI = new EthernetDAPI("Ethernet_DAPI");

    //Get ethernet device list by API
    error_code = pEthernetDAPI->GetDeviceNames(&eth_list);
    if (error_code != NO_ERROR_MSG)
    {
        printError(error_code);
        return -1;
    }
    else
    {
        while (!eth_list.empty())
        {
            cout << setw(15) << setiosflags(ios::left) << eth_list.front().name \
                 << setw(10) << eth_list.front().BDF << eth_list.front().controller << endl;

            eth_list.pop_front();
        }
    }

    return 0;
}



int
EthernetCommand::Show(StringParser command_line)
{
   // cout << "I2C Show command execution" << endl;
    string parser_operation_name;
	  int i, err = 0;
	  string operation_name;

    command_line.getOperation(&parser_operation_name);

    //Check help or not operation than list supported operation
    if (parser_operation_name == HELP || parser_operation_name == "")
    {
        listOperation(command_line);
        return 0;
    }

    //Search for operation
    for (i = 0; i< m_operation_count; i++)
    {
        m_operation[i]->getName(&operation_name);
        if (operation_name == parser_operation_name)
        {
            err = m_operation[i]->executeOperation(command_line);
            break;
        }
    }

    if (i == m_operation_count)
    {
        cout << parser_operation_name << " operation not found" << endl;
        return -1;
    }
    else
    {
        return err;
    }

    return 0;
}



int
EthernetCommand::Test(StringParser command_line)
{
    string parser_operation_name;
    int i, err = 0;
    string operation_name;

    command_line.getOperation(&parser_operation_name);

    //Check help or not operation than list supported operation
    if (parser_operation_name == HELP || parser_operation_name == "")
    {
        listOperation(command_line);
        return 0;
    }

    //Search for operation
    for (i = 0; i< m_operation_count; i++)
    {
        m_operation[i]->getName(&operation_name);
        if (operation_name == parser_operation_name)
        {
            err = m_operation[i]->executeOperation(command_line);
            break;
        }
    }

    if (i == m_operation_count)
    {
        cout << parser_operation_name << " operation not found" << endl;
        return -1;
    }
    else
    {
        return err;
    }
}



int
EthernetCommand::MAC(StringParser command_line)
{
    string parser_operation_name;
    int i, err = 0;
    string operation_name;

    command_line.getOperation(&parser_operation_name);

    //Check help or not operation than list supported operation
    if (parser_operation_name == HELP || parser_operation_name == "")
    {
        listOperation(command_line);
        return 0;
    }

    //Search for operation
    for (i = 0; i< m_operation_count; i++)
    {
        m_operation[i]->getName(&operation_name);
        if (operation_name == parser_operation_name)
        {
            err = m_operation[i]->executeOperation(command_line);
            break;
        }
    }

    if (i == m_operation_count)
    {
        cout << parser_operation_name << " operation not found" << endl;
        return -1;
    }
    else
    {
        return err;
    }
}



int
EthernetCommand::EEPROM(StringParser command_line)
{
    int    i, err = 0;
    string parser_operation_name;
    string operation_name;

    command_line.getOperation(&parser_operation_name);

    //Check help or not operation than list supported operation
    if (parser_operation_name == HELP || parser_operation_name == "")
    {
        listOperation(command_line);
        return 0;
    }

    //Search for operation
    for (i = 0; i< m_operation_count; i++)
    {
        m_operation[i]->getName(&operation_name);
        if (operation_name == parser_operation_name)
        {
            err = m_operation[i]->executeOperation(command_line);
            break;
        }
    }

    if (i == m_operation_count)
    {
        cout << parser_operation_name << " operation not found" << endl;
        return -1;
    }
    else
    {
        return err;
    }
}



int
EthernetCommand::SFP(StringParser command_line)
{
    int    i, err = 0;
    string parser_operation_name;
    string operation_name;

    command_line.getOperation(&parser_operation_name);

    //Check help or not operation than list supported operation
    if (parser_operation_name == HELP || parser_operation_name == "")
    {
        listOperation(command_line);
        return 0;
    }

    //Search for operation
    for (i = 0; i< m_operation_count; i++)
    {
        m_operation[i]->getName(&operation_name);
        if (operation_name == parser_operation_name)
        {
            err = m_operation[i]->executeOperation(command_line);
            break;
        }
    }

    if (i == m_operation_count)
    {
        cout << parser_operation_name << " operation not found" << endl;
        return -1;
    }
    else
    {
        return err;
    }
}



int
EthernetCommand::ERROR(StringParser command_line)
{
    int    i, err = 0;
    string parser_operation_name;
    string operation_name;

    command_line.getOperation(&parser_operation_name);

    //Check help or not operation than list supported operation
    if (parser_operation_name == HELP || parser_operation_name == "")
    {
        listOperation(command_line);
        return 0;
    }

    //Search for operation
    for (i = 0; i< m_operation_count; i++)
    {
        m_operation[i]->getName(&operation_name);
        if (operation_name == parser_operation_name)
        {
            err = m_operation[i]->executeOperation(command_line);
            break;
        }
    }

    if (i == m_operation_count)
    {
        cout << parser_operation_name << " operation not found" << endl;
        return -1;
    }
    else
    {
        return err;
    }
}



int
EthernetCommand::executeCommand(StringParser command_line)
{
	  int err = 0;
	  COMMAND_TYPE command_type = NO_COMMAND;

    command_line.getCommandType(&command_type);

    switch (command_type)
    {
        case CMD_LIST:
            err = List(command_line);
            break;
        case CMD_SHOW:
            err = Show(command_line);
            break;
        case CMD_TEST:
            err = Test(command_line);
            break;
        case CMD_MAC:
            err = MAC(command_line);
            break;
        case CMD_EEPROM:
            err = EEPROM(command_line);
            break;
        case CMD_SFP:
            err = SFP(command_line);
            break;
        case CMD_ERROR:
            err = ERROR(command_line);
            break;
        default:
            cout << "Not supported command" << endl;
            err = -1;
            break;
    }

    return err;
}


