#include "GPIOOperation.hpp"



GPIOOperation::GPIOOperation(OPERATION_TYPE OprationType, string helpContent) : OperationBaseClass(OprationType, helpContent)
{
    m_pGPIODAPI = NULL;
    switch (OprationType)
    {
        case OPERATION_GET:
            setName("get");
            setHelp(helpContent);
            m_OperationType = OprationType;
            break;
        case OPERATION_SET:
            setName("set");
            setHelp(helpContent);
            m_OperationType = OprationType;
            break;
        default:
            break;
    }
}



GPIOOperation::~GPIOOperation()
{

  if (m_pGPIODAPI != NULL)
      delete m_pGPIODAPI;

}



int
GPIOOperation::executeOperation(StringParser CommandLine)
{
    //printf("%s\n", __func__);
    string parameter1;
    COMMAND_TYPE CommandType;
    int    err = 0;

    CommandLine.getParameter(&parameter1, 0);
    CommandLine.getCommandType(&CommandType);

    //If the first parameter is "help" or space, show help message
    if (CommandType == CMD_LIST)
    {
        parameter1 = NO_OPRATION;

    }
    else
    {
        if (parameter1 == HELP || parameter1 == "")
        {
            helpParameter();
            err = -1;
            return err;
        }
    }

    //Execute operation
    switch (m_OperationType)
    {
        case NO_OPRATION:
            err = executeList(CommandLine);
            break;
        case OPERATION_GET:
        case OPERATION_SET:
            err = executeUtility(CommandLine);
            break;
        default:
            cout << "Unsupport operation" << endl;
            err = -1; 
            break;
    }

    return err;
}



int
GPIOOperation::helpParameter()
{
    //Show parameter help message depend on operation type
    switch (m_OperationType)
    {
        case OPERATION_GET:
            cout << "Get GPIO Status" << endl;
            cout << "Syntax: get P1" << endl;
            cout << "P1: GPIO Name" << endl;
            break;
        case OPERATION_SET:
            cout << "Set GPIO Status" << endl;
            cout << "Syntax: set P1 P2" << endl;
            cout << "P1: On or OFF" << endl;
            cout << "P2: GPIO Name" << endl;
            break;
       default:
            break;
    }
    
    return 0;
}


int GPIOOperation::executeList(StringParser CommandLine)
{
    INT32 err = 0;
    //printf("%s\n", __func__);

    ErrorMsg*   pErrorMsg = NO_ERROR_MSG;
    list<string>  GPIODevicelist;
    GPIODevicelist.clear();
    string  strErrorMessage;
    strErrorMessage.clear();

    m_pGPIODAPI = new GPIODAPI("GPIODAPI");

    if((pErrorMsg=m_pGPIODAPI->getErrorMsg()))
    {
    	getErrorMessage(pErrorMsg, &strErrorMessage);
    	cout << "Error:" << strErrorMessage.c_str() << endl;
    	//getLastErrorMessage(pErrorMsg, &strErrorMessage);
    	//printError(pErrorMsg);
    	return -1;
    }

    err = checkSyntax(CommandLine);
    if (err == 0)
    {

        pErrorMsg = m_pGPIODAPI->displayGPIOList(&GPIODevicelist);
        if( pErrorMsg == NO_ERROR_MSG)
        {
            while (!GPIODevicelist.empty())
            {
                cout << GPIODevicelist.front().c_str() << endl;
                GPIODevicelist.pop_front();
            }
        }
        else
        {
            getErrorMessage(pErrorMsg, &strErrorMessage);
            cout << "Error:" << strErrorMessage.c_str() << endl;
            //getErrorMsg();
        }


    }

    return err;
}


int
GPIOOperation::executeUtility(StringParser CommandLine)
{
    int err = 0;

    //printf("%s\n", __PRETTY_FUNCTION__);


    string TargetName;
    TargetName.clear();

    string SetValue;
    SetValue.clear();

    CHAR    TempString[256];

    BOOL    Status = TRUE;
    //UINT32 intLogicalID = -1;
    ErrorMsg*   pErrorMsg = NO_ERROR_MSG;

    string  strErrorMessage;
    strErrorMessage.clear();

    m_pGPIODAPI = new GPIODAPI("GPIODAPI");


    err = checkSyntax(CommandLine);
    if (err == 0)
    {

        //printf("2nd Parameter:%s\n", TargetName.c_str());

        switch(m_OperationType)
        {
            case OPERATION_GET:
                CommandLine.getParameter(&TargetName, 0);


                pErrorMsg = m_pGPIODAPI->getStatus(TargetName, &Status);
                if(pErrorMsg != NO_ERROR_MSG)
                {
                    getErrorMessage(pErrorMsg, &strErrorMessage);
                    cout << strErrorMessage.c_str() << endl;
                    return -1;
                }

                break;
            case OPERATION_SET:
                CommandLine.getParameter(&SetValue, 0);
                CommandLine.getParameter(&TargetName, 1);

                if(strcmp(SetValue.c_str(), "0") == 0 || strcmp(SetValue.c_str(), "off") == 0 || strcmp(SetValue.c_str(), "OFF") == 0 )
                {
                	SetValue="off";
                }
                else if(strcmp(SetValue.c_str(), "1") == 0 || strcmp(SetValue.c_str(), "on") == 0 || strcmp(SetValue.c_str(), "ON") == 0)
                {
                	SetValue="on";
                }

                sprintf(TempString, "%s", SetValue.c_str());
                ConvertStringLowerCase(TempString);

                //printf("TempString=%s\n", TempString);

                // Set
                if(strcmp(TempString, "on") == 0)
                {
                    pErrorMsg = m_pGPIODAPI->setStatus(TargetName, ON);
                }
                else
                {
                    pErrorMsg = m_pGPIODAPI->setStatus(TargetName, OFF);
                }


                if(pErrorMsg != NO_ERROR_MSG)
                {
                    getErrorMessage(pErrorMsg, &strErrorMessage);
                    cout << strErrorMessage.c_str() << endl;
                    return -1;
                }

                // read back
                pErrorMsg = m_pGPIODAPI->getStatus(TargetName, &Status);
                if(pErrorMsg != NO_ERROR_MSG)
                {
                    getErrorMessage(pErrorMsg, &strErrorMessage);
                    cout << strErrorMessage.c_str() << endl;
                    return -1;
                }

                break;
            default:
                return -1;
                break;
        }

        TargetName.clear();
        SetValue.clear();

        switch(m_OperationType)
        {
            case OPERATION_GET:
                //TargetName.append(Infolist.front().c_str());
                if(Status)
                {
                    cout << "ON" << endl;
                }
                else
                {
                    cout << "OFF" << endl;
                }
                //cout << Status;

                break;
            case OPERATION_SET:
                //TargetName.append(Infolist.front().c_str() );
                if(Status)
                {
                    cout << "ON" << endl;
                }
                else
                {
                    cout << "OFF" << endl;
                }
                break;
            default:
                break;
        }
        cout << TargetName.c_str() << endl;
        //cout << Infolist.front().c_str() << endl;
        //Infolist.pop_front();

        TargetName.clear();
    }


    return err;
}

/*
int
GPIOOperation::executeTest(StringParser CommandLine)
{
    INT32 err = 0;
    string TempString;
    ErrorMsg*   pErrorMsg = NO_ERROR_MSG;


    INT32 StartSector = 0;
    INT32 TestSize = 0;
    string TargetName;

    TempString.clear();
    TargetName.clear();

    string  strErrorMessage;
    strErrorMessage.clear();

    UINT32 intLogicalID = 0;
    list<string>  Infolist;

    m_pGPIODAPI = new GPIODAPI("GPIODAPI");

    err = checkSyntax(CommandLine);
    if (err == 0)
    {
        CommandLine.getParameter(&TempString, 0);
        //printf("1st Parameter:%s\n", TempString.c_str());
        CommandLine.convertStringToHex(TempString, &StartSector);
        //printf("1st ParameterToHex:%d\n", StartSector);
        TempString.clear();

        CommandLine.getParameter(&TempString, 1);
        //printf("2nd Parameter:%s\n", TempString.c_str());
        CommandLine.convertStringToHex(TempString, &TestSize);
        //printf("2nd ParameterToHex:%d\n", TestSize);
        TempString.clear();

        CommandLine.getParameter(&TargetName, 2);
        //printf("3rd Parameter:%s\n", TargetName.c_str());

        m_pGPIODAPI->getLogicalIdByName(TargetName.c_str(), &intLogicalID);
        //printf("Operation-executeUtility-LogicalID=%d\n",intLogicalID );
        //m_pGPIODAPI->getDataByID(intLogicalID, GPIODataElement, &Infolist);


        switch(m_OperationType)
        {
            case READPFMC:
                pErrorMsg = m_pGPIODAPI->readPerformanceTest(intLogicalID, &Infolist, (UINT32)StartSector, (UINT32)TestSize);
                break;
            case WRITEPFMC:
                pErrorMsg = m_pGPIODAPI->writePerformanceTest(intLogicalID, &Infolist, (UINT32)StartSector, (UINT32)TestSize);
                break;
            case WRC:
                pErrorMsg = m_pGPIODAPI->readwriteTest(intLogicalID, &Infolist, (UINT32)StartSector, (UINT32)TestSize);
                break;
            default:
                cout << "Unsupport operation" << endl;
                err = -1;
                break;
        }

        if(pErrorMsg != NO_ERROR_MSG)
        {
            getErrorMessage(pErrorMsg, &strErrorMessage);
            cout <<  strErrorMessage.c_str() << endl;
            return -1;
        }
        else
        {
            while(!Infolist.empty())
            {
                cout << Infolist.front().c_str() << endl;
                Infolist.pop_front();
            }
        }

        //cout << "Execute Test operation." << endl;
    }

    return err;
}
*/


int
GPIOOperation::checkSyntax(StringParser CommandLine)
{
    int    intParameterNumber = 0;
    string SetValue="";
    int    err = 0;

    CommandLine.getParameterNumber(&intParameterNumber);
    CommandLine.getParameter(&SetValue, 0);
    //printf("intParameterNumber=%d\n", intParameterNumber);

    switch (m_OperationType)
    {
        case NO_OPRATION:
            if (intParameterNumber > 0)
            {
                cout << "Not Necessary Operation Parameter" << endl;
                helpParameter();
                return -1;
            }
            break;
        case OPERATION_GET:
             //Check parameter number
            if (intParameterNumber != 1)
            {
                cout << "Parameter numbers isn't 1" << endl;
                helpParameter();
                return -1;
            }
            break;
        case OPERATION_SET:
        	//Check parameter number
            if (intParameterNumber != 2)
            {
                cout << "Parameter numbers isn't 2" << endl;
                helpParameter();
                return -1;
            }

           if(strcmp(SetValue.c_str(), "0") == 0 || strcmp(SetValue.c_str(), "off") == 0 || strcmp(SetValue.c_str(), "OFF") == 0 )
           {
           	SetValue="off";
           }
           else if(strcmp(SetValue.c_str(), "1") == 0 || strcmp(SetValue.c_str(), "on") == 0 || strcmp(SetValue.c_str(), "ON") == 0)
           {
           	SetValue="on";
           }
           else
           {
           	cout << "The 1 Parameter necessary 1 | 0 | ON | OFF";
           	helpParameter();
           	return -1;
           }
            break;
       default:
            break;
    }

    return err;
}
