#include "NTBCommand.hpp"

NTBCommand::NTBCommand(COMMAND_TYPE eCommandType, string strHelpContent) : CommandBaseClass(eCommandType, strHelpContent)
{
    m_pNTBDAPI = NULL;
    m_s32Result = 0; //default to no error

    switch(eCommandType)
    {
        case CMD_SHOW:
            addOperation(new NTBOperation(OPERATION_ALL,        "Show all information"));
            addOperation(new NTBOperation(OPERATION_SHOW_LINK,  "Show link status"));
            addOperation(new NTBOperation(OPERATION_SHOW_SPEED, "Show link speed"));
            addOperation(new NTBOperation(OPERATION_SHOW_WIDTH, "Show link width"));
            break;
        case CMD_TEST:
            addOperation(new NTBOperation(OPERATION_SERVER,         "Perform ntb server"));
            addOperation(new NTBOperation(OPERATION_CLIENT,         "Perform ntb client"));
            addOperation(new NTBOperation(OPERATION_PERFORMANCE,    "Performance test"));
            break;
        case CMD_ERROR:
            addOperation(new NTBOperation(OPERATION_CLEAR,          "Clear error counters"));
            addOperation(new NTBOperation(OPERATION_ERROR_ALL,      "Show all kinds of errors"));
            addOperation(new NTBOperation(OPERATION_PCI_COMPATIBLE, "Show PCI-compatible errors"));
            addOperation(new NTBOperation(OPERATION_PCIE,           "Show PCIE errors"));
            addOperation(new NTBOperation(OPERATION_AER,            "Show AER errors"));
            break;
        case CMD_UTILITY:
            addOperation(new NTBOperation(OPERATION_INIT,           "reset windows"));
            break;
        default:
            break;
    }
}

NTBCommand::~NTBCommand()
{

}


int 
NTBCommand::List()
{
    vector<NTB_INFO>            ntbList;
    vector<NTB_INFO>::iterator  it;

    m_pNTBDAPI = new NTBDAPI("NTBDAPI");
    
    m_pErrorMsg = m_pNTBDAPI->searchDevice(&ntbList);
    if(m_pErrorMsg != NO_ERROR_MSG)
    {
        printError(NTB_DEBUG_LEVEL, m_pErrorMsg);
        return -1;
    }

    for(it = ntbList.begin(); it != ntbList.end(); ++it)
    {
        cout << it->ntbName << endl;
    }

    delete m_pNTBDAPI;
    m_pNTBDAPI = NULL;
    return 0; 
}


int
NTBCommand::executeCommand(StringParser in_classCommandLine)
{
    COMMAND_TYPE    eCommandType = NO_COMMAND;

    in_classCommandLine.getCommandType(&eCommandType);
    switch (eCommandType)
    {
        case CMD_LIST:
            return List();
        case CMD_SHOW:
        case CMD_TEST:
        case CMD_ERROR:
        case CMD_UTILITY:
            return findAndExecuteOperation(in_classCommandLine);

        default:
            cout << UNKNOWN_COMMAND_MSG << endl;
            return -1;
    }
}










