#include "PCICommand.hpp"

PCICommand::PCICommand(COMMAND_TYPE eCommandType, string strHelpContent) : CommandBaseClass(eCommandType, strHelpContent)
{
    switch (eCommandType)
    {
        case CMD_LIST:
            addOperation(new PCIOperation(OPERATION_PCI, "List PCI/PCIe devices"));
            addOperation(new PCIOperation(OPERATION_EEPROM, "List PCIe switch eeproms"));
            addOperation(new PCIOperation(OPERATION_HBA, "List PCI HBAs"));
            addOperation(new PCIOperation(OPERATION_SWITCH, "List PCIe Switches"));
            break;
       
        case CMD_SHOW:
            addOperation(new PCIOperation(OPERATION_ALL, "Show all information of a device"));
            addOperation(new PCIOperation(OPERATION_ID, "Show vendor ID and device ID of a device"));
            addOperation(new PCIOperation(OPERATION_CLASS_TYPE, "Show class type of a device"));
            addOperation(new PCIOperation(OPERATION_PORT_TYPE, "Show port type of a device"));
            addOperation(new PCIOperation(OPERATION_SHOW_LINK, "Show link status(up/down) of a device"));
            addOperation(new PCIOperation(OPERATION_SHOW_SPEED, "Show link speed of a device"));
            addOperation(new PCIOperation(OPERATION_SHOW_WIDTH, "Show link width of a device"));
            addOperation(new PCIOperation(OPERATION_RETRAIN_LINK, "Show all available retrain path"));
            break;
        
        case CMD_TEST:
            addOperation(new PCIOperation(OPERATION_RETRAIN, "perform retrain test"));
            break;

        case CMD_UTILITY:
            addOperation(new PCIOperation(OPERATION_CONFIG_READ, "read config data"));
            addOperation(new PCIOperation(OPERATION_CONFIG_WRITE, "write config data"));
            addOperation(new PCIOperation(OPERATION_SEARCH, "search hidden PCI devices"));
            break;
        
        case CMD_EEPROM:
            addOperation(new PCIOperation(OPERATION_DUMP, "dump eeprom content"));
            addOperation(new PCIOperation(OPERATION_READ, "read eeprom data"));
            addOperation(new PCIOperation(OPERATION_WRITE, "write data to eeprom"));
            addOperation(new PCIOperation(OPERATION_TEST, "do eeprom test"));
            addOperation(new PCIOperation(OPERATION_PROGRAM, "program eeprom"));
            addOperation(new PCIOperation(OPERATION_SAVE, "save eeprom content to a file"));
            addOperation(new PCIOperation(OPERATION_FW, "show EEPROM Image version"));
            break;

        case CMD_ERROR:
            addOperation(new PCIOperation(OPERATION_CLEAR, "clear errors of a device"));
            addOperation(new PCIOperation(OPERATION_ERROR_ALL, "show all kinds of error statistics of a device"));
            addOperation(new PCIOperation(OPERATION_PCI_COMPATIBLE, "show pci-compatible error statistics"));
            addOperation(new PCIOperation(OPERATION_PCIE, "show PCIE error statistics"));
            addOperation(new PCIOperation(OPERATION_AER, "show AER"));
            break;
        
        default:
            break;
    }
}


PCICommand::~PCICommand()
{

}

int 
PCICommand::executeCommand(StringParser in_classCommandLine)
{
    COMMAND_TYPE    eCommandType = NO_COMMAND;
   
    in_classCommandLine.getCommandType(&eCommandType);
    switch (eCommandType)
    {
        case CMD_SHOW:
        case CMD_LIST: 
        case CMD_TEST:
        case CMD_UTILITY:
        case CMD_EEPROM:
        case CMD_ERROR:
            return findAndExecuteOperation(in_classCommandLine);
    
        default:
            cout << UNKNOWN_COMMAND_MSG << endl;
            return -1;
    }
}
