#include "StorageCommand.hpp"

StorageCommand::StorageCommand(COMMAND_TYPE CommandType, string strHelpContent) : CommandBaseClass(CommandType, strHelpContent)
{
     //Base on command type to add different operations
     switch (CommandType)
     {
     case CMD_LIST:
	  addOperation(new StorageOperation(OPERATION_DRIVE, "Controller-Drive Name."));
	  break;
     case CMD_SHOW:
	  addOperation(new StorageOperation(OPERATION_ALL, "Display Storage Drive Information."));
	  addOperation(new StorageOperation(OPERATION_NODE, "Display Storage Drive Device Node Name."));
	  addOperation(new StorageOperation(OPERATION_SN, "Display Storage Drive Serial Number."));
	  addOperation(new StorageOperation(OPERATION_MODEL, "Display Storage Drive Model Number."));
	  addOperation(new StorageOperation(OPERATION_FW, "Display Storage Drive Firmware Version."));
	  addOperation(new StorageOperation(OPERATION_LOGICAL_BLOCK_SIZE, "Display Storage Drive Logical Block Size."));
	  addOperation(new StorageOperation(OPERATION_PHYSICAL_BLOCK_SIZE, "Display Storage Drive Physical Block Size."));
	  addOperation(new StorageOperation(OPERATION_ADDRESS, "Display SAS Storage Drive SAS Address."));
	  addOperation(new StorageOperation(OPERATION_CAPACITY, "Display Storage Drive Capacity (size in MB)."));
	  addOperation(new StorageOperation(OPERATION_PORT, "Display Storage Drive Port Number."));
	  addOperation(new StorageOperation(OPERATION_SPEED, "Display Storage Drive Link Speed."));
	  addOperation(new StorageOperation(OPERATION_TEMP, "Display Storage Drive Temperature."));
	  break;
     case CMD_TEST:
	  addOperation(new StorageOperation(OPERATION_READPFMC, "Storage Read Performance Test"));
	  addOperation(new StorageOperation(OPERATION_WRITEPFMC, "Storage Write Performance Test"));
	  addOperation(new StorageOperation(OPERATION_WRC, "Storage Read/Write Test"));
	  break;
     case CMD_UTILITY:
          addOperation(new StorageOperation(OPERATION_SMART, "Get Storage Drive SMART Information"));
          addOperation(new StorageOperation(OPERATION_WZERO, "Write zero to Drive"));
          addOperation(new StorageOperation(OPERATION_GLIST, "Get Storage Drive G-List Data"));
          addOperation(new StorageOperation(OPERATION_PLIST, "Get Sorage Drive P-List Data"));
	  break;
	  //Add NEW command and operation here
     default:
	  break;
     }
}

StorageCommand::~StorageCommand()
{

}

INT32 StorageCommand::executeCommand(StringParser CommandLine)
{
     INT32 s32Err = 0;
     string strOperationName = "";

     string strParserstrOperationName = "";    
     CommandLine.getOperation(&strParserstrOperationName);

     COMMAND_TYPE eCommandType = (COMMAND_TYPE)0;
     CommandLine.getCommandType(&eCommandType);
     switch(eCommandType)
     {
     case CMD_LIST:
     case CMD_SHOW:
     case CMD_TEST:
     case CMD_ERROR:
     case CMD_UTILITY:
	  if(strParserstrOperationName=="help" || strParserstrOperationName=="")
	  {
	       listOperation(CommandLine);
	       return -1;
	  }
	  break;
     default:
	  cout << "Not supported command" << endl;
	  return -1;
	  break;
     }

     INT32 s32OperationIndex = 0;
     for(s32OperationIndex=0; s32OperationIndex<m_operation_count; s32OperationIndex++)
     {
	  m_operation[s32OperationIndex]->getName(&strOperationName);
	  //printf("m_operation[%d]=%s\n", s32OperationIndex, strOperationName.c_str());
	  if (strOperationName == strParserstrOperationName)
	  {
	       s32Err = m_operation[s32OperationIndex]->executeOperation(CommandLine);
	       break;
	  }
     }

     if(s32OperationIndex == m_operation_count)
     {
	  cout << strParserstrOperationName << " operation not found" << endl;
	  return -1;
     }

     return s32Err;
}
