/// @file       LEDDAPI.hpp
/// @brief      LEDDAPI class
/// @details    Provide diagnostic and utility APIs for LEDs
/// @author     KenHu
/// @copyright  Copyright 2015-2016 Hon Hai Technology Group. All Rights Reserved.

#ifndef LED_DAPI_HPP
#define LED_DAPI_HPP

#include "BaseClassDAPI.hpp"
#include "BaseClassDDAPI.hpp"

#define ARRAY_SIZE(a)   (sizeof((a))/sizeof(*(a)))

enum LED_MODULE_INDEX {
    CONTROLLER,
    ENCLOSURE,
    ETHERNET,
    PSU,
    SAS,
    NUMBER_OF_MODULES,
};

/// @class      LEDDAPI LEDDAPI.hpp "LEDDAPI.hpp"
/// @brief      Provide diagnostic and utility APIs for LEDs
/// @param      m_deviceList is a list of pointers to DDAPI base class
class LEDDAPI : public BaseClassDAPI
{
    public: 
                            LEDDAPI(STRING in_szClassName);
        virtual             ~LEDDAPI(void);

        ErrorMsg*           getDeviceList(list<string>  *out_pList);
        ErrorMsg*           getLEDStatus(string in_strLEDName,
                                         string *out_strStatus);
        ErrorMsg*           setLEDStatus(string in_strLEDName,
                                         string in_strStatus);
    protected:

    private:
        BaseClassDDAPI*     createModuleClassByName(string in_strModuleName);
        BaseClassDDAPI*     createModuleClassByID(LED_MODULE_INDEX in_eModuleIndex);
        string              getModuleName(string in_strLEDName);
};

#endif

