/// @file       PSUDAPI.hpp
/// @brief      PSUDAPI class
/// @details    Provide diagnostic and utility APIs for PSU module
/// @author     KenHu
/// @copyright  Copyright 2015-2016 Hon Hai Technology Group. All Rights Reserved.

#ifndef PSU_DAPI_HPP
#define PSU_DAPI_HPP

#include "BaseClassDAPI.hpp"
#include "PSUDDAPI.hpp"

typedef enum {
    PSU_TYPE,
    PSU_FW,
    PSU_MAX_POWER,
    PSU_PART_NUMBER,
    PSU_SERIAL_NUMBER,
    PSU_VENDOR,
    PSU_LED,
    PSU_TEMPERATURE,
    PSU_TEMPERATURE_THRESHOLD,
    PSU_FAN,
    PSU_UNKNOWN,
} PSU_ELEMENT_TYPE;

/// @class      PSUDAPI PSUDAPI.hpp "PSUDAPI.hpp"
/// @brief      Provide diagnostic and utility APIs for PSU module 
/// @param      m_PSUList is a list of pointers to PSU DDAPI base class
class PSUDAPI : public BaseClassDAPI
{
    public: 
                            PSUDAPI(STRING in_szClassName);
        virtual             ~PSUDAPI(void);
        
        ErrorMsg*           getDeviceList(string        in_strDeviceType,
                                          list<string>  *out_pList);
        ErrorMsg*           getDataByElement(PSU_ELEMENT_TYPE   in_eElementType,
                                             string             in_strDevice,
                                             string             *out_strpValue);
        ErrorMsg*           getSupportValueByElement(PSU_ELEMENT_TYPE   in_eElementType,
                                                     string             in_strDevice,
                                                     list<string>       *out_pList);

        ErrorMsg*           setDataByElement(PSU_ELEMENT_TYPE   in_eElementType,
                                             string             in_strDevice,
                                             string             in_strValue);

        ErrorMsg*           updateFWByFile(string in_strDevice,
                                           UINT8  in_u8Region,
                                           string in_strFile);
    protected:

    private:
        PSUDDAPI*           m_pPSUDDAPI;
};

#endif

