/// @file       StorageDAPI.cpp
/// @brief      StorageDAPI className
/// @details    Provide diagnostic and utility APIs for Storage module
/// @version    1.0.0
/// @date       2015.12.01
/// @author     Ben Yang
/// @copyright  Copyright 2014 Hon Hai Technology Group. All Rights Reserved.
#include "StorageDAPI.hpp"
using namespace::std;
/// @fn         StorageDAPI::StorageDAPI(STRING in_szClassName) : BaseClassDAPI(in_szClassName)
/// @brief      StorageDAPI constructor
/// @param      in_szClassName is a char string for class name
/// @exception  Nothing
/// @return     Nothing
StorageDAPI::StorageDAPI(STRING in_szClassName) : BaseClassDAPI(in_szClassName)
{
     ErrorMsg *pErrorMsg = NULL;
     m_pStorageDDAPIOperation = NULL;
//     m_pSCSIOperation = NULL;
     m_strOutputResultString = "";
     m_u32StorageSectorCountPerRDWRAccess = STORAGE_SECTOR_COUNT_PER_RDWR_ACCESS;

     m_pStorageDDAPIOperation = new StorageDDAPIOperation("StorageDDAPIOperation");
     if(m_pStorageDDAPIOperation == NULL)
     {
	  addErrorMsg((ERRORCODE)(ERROR_STORAGE_DAPI_CONSTRUCTOR | ERROR_DIAGNOSTIC_API_CONSTRUCTOR1_ERROR), "Create storage DDAPI fail");
//	  pErrorMsg = m_pStorageDDAPIOperation->getErrorMsg();
     }
     if(m_pStorageDDAPIOperation->getErrorMsg())
     {
	  addErrorMsg((ERRORCODE)(ERROR_STORAGE_DAPI_CONSTRUCTOR | ERROR_DIAGNOSTIC_API_CONSTRUCTOR1_ERROR), "Create storage DDAPI fail");

     }
/*
     if(pErrorMsg != NULL)
     {
	  addErrorMsg(ERROR_STORAGE_DAPI_CONSTRUCTOR, NULL, pErrorMsg);
	  }*/
/*
     m_pSCSIOperation = new SCSIOperation("SCSIOperation");
     if(m_pSCSIOperation == NULL)
     {
	  pErrorMsg = m_pSCSIOperation->getErrorMsg();
     }

     if(pErrorMsg != NULL)
     {
	  addErrorMsg(ERROR_STORAGE_DAPI_CONSTRUCTOR, NULL, pErrorMsg);
     }
*/

     // generate storage list
/*     if((pErrorMsg = generateStorageList()) != NO_ERROR_MSG)
     {
	  addErrorMsg(ERROR_STORAGE_DAPI_CONSTRUCTOR, NULL, pErrorMsg);
//	  return returnErrorMsg(ERROR_STORAGE_DAPI_IOCTL, NULL, pErrorMsg);
     }
*/
     m_pSCSIFactory = new SCSIFactory("SCSIFactory");
     if(m_pSCSIFactory == NULL)
     {
	  pErrorMsg = m_pSCSIFactory->getErrorMsg();
     }
     if(pErrorMsg != NULL)
     {
	  addErrorMsg(ERROR_STORAGE_DAPI_CONSTRUCTOR, NULL, pErrorMsg);
     }

     m_pSCSISenseAnalyzer = new SCSISenseAnalyzer("SCSISenseAnalyzer");
     if(m_pSCSISenseAnalyzer == NULL)
     {
	  pErrorMsg = m_pSCSISenseAnalyzer->getErrorMsg();
     }
     if(pErrorMsg != NULL)
     {
	  addErrorMsg(ERROR_STORAGE_DAPI_CONSTRUCTOR, NULL, pErrorMsg);
     }

     m_pNVMeDAPI = new NVMeDAPI("NVMeDAPI");
     if(m_pNVMeDAPI == NULL)
     {
	  pErrorMsg = m_pNVMeDAPI->getErrorMsg();
     }
     if(pErrorMsg != NULL)
     {
	  addErrorMsg(ERROR_STORAGE_DAPI_CONSTRUCTOR, NULL, pErrorMsg);
     }
}

/// @fn         StorageDAPI::~StorageDAPI(void)
/// @brief      StorageDAPI destructor
/// @exception  Nothing
/// @return     Nothing
StorageDAPI::~StorageDAPI()
{
     m_strOutputResultString = "";
     m_u32StorageSectorCountPerRDWRAccess = STORAGE_SECTOR_COUNT_PER_RDWR_ACCESS;

     if(m_pStorageDDAPIOperation != NULL)
     {
	  delete m_pStorageDDAPIOperation;
     }
/*
     if(m_pSCSIOperation != NULL)
     {
	  delete m_pSCSIOperation;
     }
*/
     if(m_pSCSIFactory != NULL)
     {
	  delete m_pSCSIFactory;
     }
     if(m_pSCSISenseAnalyzer != NULL)
     {
	  delete m_pSCSISenseAnalyzer;
     }
     if(m_pNVMeDAPI != NULL)
     {
	  delete m_pNVMeDAPI;
     }
}

/// @fn         ErrorMsg *StorageDAPI::IOCTL(UINT32 in_u32DefinedIndex, StorageIOCTLOperation in_eOperation)
/// @brief      Function for abstraction of DAPI IOCTL operation
/// @param      in_u32DefinedIndex is a index for Storage device node name (eg. 1)
/// @param      in_eOperation desired driver opertaion
/// @param      inout_pXMLData is a pointer to XML data for Storage information
/// @exception  Nothing
/// @return     A pointer to ErrorMsg for error message list
ErrorMsg *StorageDAPI::IOCTL(UINT32 in_u32DefinedIndex, StorageIOCTLOperation in_eOperation)
{
     ERRORCODE errorCode = NO_ERROR;
     ErrorMsg *pErrorMsg = NO_ERROR_MSG;
/*
     if(in_eOperation != eStorageGenerateList)
     {
	  if((pErrorMsg = generateStorageList()) != NO_ERROR_MSG)
	  {
	       return returnErrorMsg(ERROR_STORAGE_DAPI_IOCTL, NULL, pErrorMsg);
	  }
     }
*/
     switch(in_eOperation)
     {
     case eStorageGenerateList:
	  if((pErrorMsg = m_pStorageDDAPIOperation->IOCTL(0, eStorageGenerateList)) != NO_ERROR_MSG)
	  {
	       return returnErrorMsg(ERROR_STORAGE_DAPI_IOCTL, NULL, pErrorMsg);
	  }
	  break;

     case eStorageListDrive:
	  if((pErrorMsg = m_pStorageDDAPIOperation->IOCTL(0, eStorageListDrive)) != NO_ERROR_MSG)
	  {
	       return returnErrorMsg(ERROR_STORAGE_DAPI_IOCTL, NULL, pErrorMsg);
	  }
	  break;

     default:
	  errorCode = ERROR_STORAGE_DAPI_NO_SUCH_IOCTL_OPERATION;
	  return returnErrorMsg((ERRORCODE)(ERROR_STORAGE_DAPI_IOCTL | errorCode), NULL, NULL);
	  break;
     }
     return returnErrorMsg(errorCode, NULL, NULL);
}

/// @fn         ErrorMsg *StorageDAPI::generateStorageList()
/// @brief      generate Storage devie list
/// @exception  Nothing
/// @return     A pointer to ErrorMsg for error message list
ErrorMsg *StorageDAPI::generateStorageList()
{
     ERRORCODE errorCode = NO_ERROR;
     ErrorMsg *pErrorMsg = NO_ERROR_MSG;
     if((pErrorMsg = m_pStorageDDAPIOperation->generateStorageList()) != NO_ERROR_MSG)
     {
	  return returnErrorMsg(ERROR_STORAGE_DAPI_GENERATE_STORAGE_LIST, NULL, pErrorMsg);
     }

     return returnErrorMsg(errorCode, NULL, NULL);
}

ErrorMsg *StorageDAPI::getIndexByName(string in_strDefinedName, UINT32 *out_u32DefinedIndex)
{
     ERRORCODE errorCode = NO_ERROR;
     ErrorMsg *pErrorMsg = NO_ERROR_MSG;
/*
     if((pErrorMsg = generateStorageList()) != NO_ERROR_MSG)
     {
	  return returnErrorMsg(ERROR_STORAGE_DAPI_IOCTL, NULL, pErrorMsg);
     }
*/
     if((pErrorMsg = m_pStorageDDAPIOperation->getIndexByName(in_strDefinedName, out_u32DefinedIndex)) != NO_ERROR_MSG)
     {
	  return returnErrorMsg(ERROR_STORAGE_DAPI_GET_DEFINED_INDEX_BY_NAME, NULL, pErrorMsg);
     }

   return returnErrorMsg(errorCode, NULL, NULL);
}

ErrorMsg *StorageDAPI::getElementDataByIndex(UINT32 in_u32DefinedIndex, StorageDataElement in_eStorageDataElement, string *out_strpReturnData)
{
     ERRORCODE errorCode = NO_ERROR;
     ErrorMsg *pErrorMsg = NO_ERROR_MSG;
/*
     if((pErrorMsg = generateStorageList()) != NO_ERROR_MSG)
     {
	  return returnErrorMsg(ERROR_STORAGE_DAPI_IOCTL, NULL, pErrorMsg);
     }
*/
     if((pErrorMsg = m_pStorageDDAPIOperation->getElementDataByIndex(in_u32DefinedIndex, in_eStorageDataElement, out_strpReturnData)) != NO_ERROR_MSG)
     {
	  return returnErrorMsg(ERROR_STORAGE_DAPI_GET_ELEMENT_DATA_BY_INDEX, NULL, pErrorMsg);
     }

     return returnErrorMsg(errorCode, NULL, NULL);
}

ErrorMsg *StorageDAPI::getSMARTByIndex(UINT32 in_u32DefinedIndex)
{
     ERRORCODE errorCode = NO_ERROR;
     ErrorMsg *pErrorMsg = NO_ERROR_MSG;

     UINT32 u32ControllerType = 0;
     if((pErrorMsg = m_pStorageDDAPIOperation->getControllerTypeByIndex(in_u32DefinedIndex, &u32ControllerType)) != NO_ERROR_MSG)
     {
	  return returnErrorMsg(ERROR_STORAGE_DAPI_GET_SMART_BY_INDEX, NULL, pErrorMsg);
     }

     vector<NVMe_DEVICE_INFO> nvmeList;
     string strDefinedName = "";
     UINT32 u32DefinedIndex = 0;
     INT32 s32NSID = 0;
     INT32 c;
     unsigned int temperature;
     
     stringstream ssTemp;
     ssTemp.str("");
     ssTemp.clear();

     string strReturnData = "";
     switch(u32ControllerType)
     {
     case eStorageControllerTypeSAS:
	  if((pErrorMsg = m_pStorageDDAPIOperation->getSMARTByIndex(in_u32DefinedIndex, &strReturnData)) != NO_ERROR_MSG)
	  {
	       return returnErrorMsg(ERROR_STORAGE_DAPI_GET_SMART_BY_INDEX, NULL, pErrorMsg);
	  }
	  break;
     case eStorageControllerTypeNVME:
	  if((pErrorMsg = m_pStorageDDAPIOperation->getNameByIndex(in_u32DefinedIndex, &strDefinedName)) != NO_ERROR_MSG)
	  {
	       return returnErrorMsg(ERROR_STORAGE_DAPI_GET_SMART_BY_INDEX, NULL, pErrorMsg);	       
	  }

	  if((pErrorMsg = m_pNVMeDAPI->getInformationList(&nvmeList)) == NO_ERROR_MSG)
	  {
	       if((pErrorMsg = m_pNVMeDAPI->getLogicalIDByName(strDefinedName, (UINT8*)&u32DefinedIndex)) == NO_ERROR_MSG)
	       {
		    if((pErrorMsg = m_pNVMeDAPI->getBarResourceByController(u32DefinedIndex, &nvmeList[u32DefinedIndex].barInfo)) == NO_ERROR_MSG)
		    {
			 if((pErrorMsg = m_pNVMeDAPI->getIdentityCTRLByController(u32DefinedIndex, &nvmeList[u32DefinedIndex].idCtrl)) == NO_ERROR_MSG)
			 {
			      if((pErrorMsg = m_pNVMeDAPI->getIdentityNSByController(u32DefinedIndex, &nvmeList[u32DefinedIndex].idNs)) == NO_ERROR_MSG)
			      {
				   if((pErrorMsg = m_pNVMeDAPI->getFirmwareLogByController(u32DefinedIndex, &nvmeList[u32DefinedIndex].fwLog)) == NO_ERROR_MSG)
				   {
					if((pErrorMsg = m_pNVMeDAPI->getSmartLogByController(u32DefinedIndex, &nvmeList[u32DefinedIndex].smartLog)) == NO_ERROR_MSG)
					{
					     pErrorMsg = m_pNVMeDAPI->getNameSpaceIDByController(u32DefinedIndex, &s32NSID);
					}
				   }
			      }
			 }
		    }
	       }
	  }

	  ssTemp.str("");
	  ssTemp.clear();
/*
	  temperature = ((nvmeList[u32DefinedIndex].smartLog.temperature[1] << 8) | nvmeList[u32DefinedIndex].smartLog.temperature[0]) - 273;
	  printf("Smart Log for NVME device:%s\n", strDefinedName.c_str());
	  printf("critical_warning                    : %#x\n", nvmeList[u32DefinedIndex].smartLog.critical_warning);
	  printf("temperature                         : %u C\n", temperature);
	  printf("available_spare                     : %u%%\n", nvmeList[u32DefinedIndex].smartLog.avail_spare);
	  printf("available_spare_threshold           : %u%%\n", nvmeList[u32DefinedIndex].smartLog.spare_thresh);
	  printf("percentage_used                     : %u%%\n", nvmeList[u32DefinedIndex].smartLog.percent_used);
	  printf("data_units_read                     : %'.0Lf\n",
		 int128_to_double(nvmeList[u32DefinedIndex].smartLog.data_units_read));
	  printf("data_units_written                  : %'.0Lf\n",
		 int128_to_double(nvmeList[u32DefinedIndex].smartLog.data_units_written));
	  printf("host_read_commands                  : %'.0Lf\n",
		 int128_to_double(nvmeList[u32DefinedIndex].smartLog.host_reads));
	  printf("host_write_commands                 : %'.0Lf\n",
		 int128_to_double(nvmeList[u32DefinedIndex].smartLog.host_writes));
	  printf("controller_busy_time                : %'.0Lf\n",
		 int128_to_double(nvmeList[u32DefinedIndex].smartLog.ctrl_busy_time));
	  printf("power_cycles                        : %'.0Lf\n",
		 int128_to_double(nvmeList[u32DefinedIndex].smartLog.power_cycles));
	  printf("power_on_hours                      : %'.0Lf\n",
		 int128_to_double(nvmeList[u32DefinedIndex].smartLog.power_on_hours));
	  printf("unsafe_shutdowns                    : %'.0Lf\n",
		 int128_to_double(nvmeList[u32DefinedIndex].smartLog.unsafe_shutdowns));
	  printf("media_errors                        : %'.0Lf\n",
		 int128_to_double(nvmeList[u32DefinedIndex].smartLog.media_errors));
	  printf("num_err_log_entries                 : %'.0Lf\n",
		 int128_to_double(nvmeList[u32DefinedIndex].smartLog.num_err_log_entries));
	  printf("Warning Temperature Time            : %u\n", le32toh(nvmeList[u32DefinedIndex].smartLog.warning_temp_time));
	  printf("Critical Composite Temperature Time : %u\n", le32toh(nvmeList[u32DefinedIndex].smartLog.critical_comp_time));
	  for (c=0; c < 8; c++) {
	       __u16 temp = le16toh(nvmeList[u32DefinedIndex].smartLog.temp_sensor[c]);
	       printf("Temperature Sensor %d                : %u C\n", c + 1,
		      temp ? temp - 273 : 0);
	  }*/

	  ssTemp << "Smart Log for NVME device: " << strDefinedName << endl;
	  ssTemp << "critical_warning                    : " << hex << nvmeList[u32DefinedIndex].smartLog.critical_warning << endl;
	  ssTemp << "temperature                         : " << dec << temperature << " C" << endl;
	  ssTemp << "available_spare                     : " << (UINT32)nvmeList[u32DefinedIndex].smartLog.avail_spare << "%" << endl;
	  ssTemp << "available_spare_threshold           : " << (UINT32)nvmeList[u32DefinedIndex].smartLog.spare_thresh << "%" << endl;
	  ssTemp << "percentage_used                     : " << (UINT32)nvmeList[u32DefinedIndex].smartLog.percent_used << "%" << endl;
	  ssTemp << "data_units_read                     : " << (UINT32)int128_to_double(nvmeList[u32DefinedIndex].smartLog.data_units_read) << endl;
	  ssTemp << "data_units_written                  : " << (UINT32)int128_to_double(nvmeList[u32DefinedIndex].smartLog.data_units_written) << endl;
	  ssTemp << "host_read_commands                  : " << (UINT32)int128_to_double(nvmeList[u32DefinedIndex].smartLog.host_reads) << endl;
          ssTemp << "host_write_commands                 : " << (UINT32)int128_to_double(nvmeList[u32DefinedIndex].smartLog.host_writes) << endl;
          ssTemp << "controller_busy_time                : " << (UINT32)int128_to_double(nvmeList[u32DefinedIndex].smartLog.ctrl_busy_time) << endl;
	  ssTemp << "power_cycles                        : " << (UINT32)int128_to_double(nvmeList[u32DefinedIndex].smartLog.power_cycles) << endl;
	  ssTemp << "power_on_hours                      : " << (UINT32)int128_to_double(nvmeList[u32DefinedIndex].smartLog.power_on_hours) << endl;
	  ssTemp << "unsafe_shutdowns                    : " << (UINT32)int128_to_double(nvmeList[u32DefinedIndex].smartLog.unsafe_shutdowns) << endl;
	  ssTemp << "media_errors                        : " << (UINT32)int128_to_double(nvmeList[u32DefinedIndex].smartLog.media_errors) << endl;
	  ssTemp << "num_err_log_entries                 : " << (UINT32)int128_to_double(nvmeList[u32DefinedIndex].smartLog.num_err_log_entries) << endl;
	  ssTemp << "Warning Temperature Time            : " << le32toh(nvmeList[u32DefinedIndex].smartLog.warning_temp_time) << endl;
	  ssTemp << "Critical Composite Temperature Time : " << le32toh(nvmeList[u32DefinedIndex].smartLog.critical_comp_time) << endl;
	  for(c=0; c < 8; c++)
	  {
	       __u16 temp = le16toh(nvmeList[u32DefinedIndex].smartLog.temp_sensor[c]);

	       if(temp)
	       {
		    ssTemp << "Temperature Sensor " << c + 1 << "                : " << (UINT32)temp - 273 << " C" << endl;
	       }
	       else
	       {

		    ssTemp << "Temperature Sensor                  : " << 0 << " C" << endl;
	       }
//	       ssTemp << ": "<< temp ? temp - 273 : 0 << "C" << endl;
	  }
	  strReturnData = ssTemp.str();
	  if(pErrorMsg != NO_ERROR_MSG)
	  {
	       return returnErrorMsg(ERROR_STORAGE_DAPI_GET_SMART_BY_INDEX, NULL, pErrorMsg);
	  }
	  break;
     default:
	  errorCode = ERROR_STORAGE_DAPI_DEVICE_NOT_SUPPORT_SMART;
	  return returnErrorMsg((ERRORCODE)(ERROR_STORAGE_DAPI_GET_SMART_BY_INDEX | errorCode), NULL, NULL);
     }

     m_strOutputResultString = strReturnData;

     return returnErrorMsg(errorCode, NULL, NULL);
}

ErrorMsg *StorageDAPI::getElementDataByName(string in_strDefinedName, StorageDataElement in_eStorageDataElement, string *out_strpReturnData)
{
     ERRORCODE errorCode = NO_ERROR;
     ErrorMsg *pErrorMsg = NO_ERROR_MSG;
/*
     if((pErrorMsg = generateStorageList()) != NO_ERROR_MSG)
     {
	  return returnErrorMsg(ERROR_STORAGE_DAPI_IOCTL, NULL, pErrorMsg);
     }
*/
     if((pErrorMsg = m_pStorageDDAPIOperation->getElementDataByName(in_strDefinedName, in_eStorageDataElement, out_strpReturnData)) != NO_ERROR_MSG)
     {
	  return returnErrorMsg(ERROR_STORAGE_DAPI_GET_ELEMENT_DATA_BY_INDEX, NULL, pErrorMsg);
     }

     return returnErrorMsg(errorCode, NULL, NULL);
}

/// @fn         ErrorMsg *StorageDAPI::readPerformanceTest(UINT32 in_u32DefinedIndex, UINT64 in_u64StartSector, UINT32 in_u32TestSize)
/// @brief      Read specific size(MB) of data from starting logical sector on hard disk through index to get read performance (MB/s)
/// @param      in_u32DefinedIndex is a index for Storage device node name (eg. 1)
/// @param      in_u64StartSector is starting sector to test
/// @param      in_u32TestSizeMB is test size by MB
/// @exception  Nothing
/// @return     A pointer to ErrorMsg for error message list
ErrorMsg *StorageDAPI::readPerformanceTest(UINT32 in_u32DefinedIndex, UINT64 in_u64StartSector, UINT32 in_u32TestSizeMB, BOOL in_blsMultiModeOutputFormat)
{
     ERRORCODE errorCode = NO_ERROR;
     ErrorMsg *pErrorMsg = NO_ERROR_MSG;

     if((pErrorMsg = setSectorCountPerRDWRAccess(in_u32DefinedIndex)) != NO_ERROR_MSG)
     {
	  return returnErrorMsg(ERROR_STORAGE_DAPI_READ_PERFORMANCE_TEST, NULL, pErrorMsg);
     }

     string strSGDeviceNodeName = "";
     if((pErrorMsg = m_pStorageDDAPIOperation->getElementDataByIndex(in_u32DefinedIndex, eStorageDriveSGDeviceNodeName, &strSGDeviceNodeName)) != NO_ERROR_MSG)
     {
	  return returnErrorMsg(ERROR_STORAGE_DAPI_READ_PERFORMANCE_TEST, NULL, pErrorMsg);
     }

     string strTempString = "";
     if((pErrorMsg = m_pStorageDDAPIOperation->getElementDataByIndex(in_u32DefinedIndex, eStorageDriveCapacity, &strTempString)) != NO_ERROR_MSG)
     {
	  return returnErrorMsg(ERROR_STORAGE_DAPI_READ_PERFORMANCE_TEST, NULL, pErrorMsg);
     }
     //strTempString = strTempString.substr(0, strTempString.size()-2);

     stringstream ssTemp;
     ssTemp.str("");
     ssTemp.clear();
     ssTemp << strTempString;
     UINT64 u64CapacityMB = 0;
     ssTemp >> u64CapacityMB;

     strTempString = "";
     if((pErrorMsg = m_pStorageDDAPIOperation->getElementDataByIndex(in_u32DefinedIndex, eStorageDriveLogicalBlockSize, &strTempString)) != NO_ERROR_MSG)
     {
	  return returnErrorMsg(ERROR_STORAGE_DAPI_READ_PERFORMANCE_TEST, NULL, pErrorMsg);
     }
     ssTemp.str("");
     ssTemp.clear();
     ssTemp << strTempString;
     UINT64 u64LogicalBlockSize = 0;
     ssTemp >> u64LogicalBlockSize;

     UINT64 u64StartSector = in_u64StartSector;
     UINT64 u64LastSector = (u64CapacityMB-1)*1024*1024/u64LogicalBlockSize;//2048;

     if(u64StartSector > u64LastSector)
     {
	  u64StartSector = u64StartSector - u64LastSector;
     }

     struct itimerval startTime, endTime, setting;
     UINT64 u64TotalByte = 0;
     double dElapsed = 0;
     UINT32 *u32pBuffer = (UINT32 *)malloc(m_u32StorageSectorCountPerRDWRAccess*u64LogicalBlockSize);
     INT32 s32FileDescriptor = open(strSGDeviceNodeName.c_str(), O_RDWR | O_SYNC | O_NONBLOCK);
     if(s32FileDescriptor < 0)
     {
	  free(u32pBuffer);
	  errorCode = ERROR_STORAGE_DAPI_OPEN_DEVICE_FAIL;
	  return returnErrorMsg((ERRORCODE)(ERROR_STORAGE_DAPI_READ_PERFORMANCE_TEST | errorCode), NULL, NULL);
     }

     setting.it_interval.tv_sec = numeric_limits<time_t>::max();//1000;
     setting.it_interval.tv_usec = 0;
     setting.it_value.tv_sec = numeric_limits<time_t>::max();//1000;
     setting.it_value.tv_usec = 0;

     // setitimer(ITIMER_REAL, &(struct itimerval){{1000,0},{1000,0}}, NULL);
     setitimer(ITIMER_REAL, &setting, NULL);
     // Now do the timings for real
     initProgress(in_u32TestSizeMB);
     getitimer(ITIMER_REAL, &startTime);
     do
     {
	  getitimer(ITIMER_REAL, &startTime);
	  pErrorMsg = m_pSCSIFactory->executeSCSIRead(s32FileDescriptor, u64StartSector, m_u32StorageSectorCountPerRDWRAccess, (UINT8*)u32pBuffer, m_u32StorageSectorCountPerRDWRAccess*u64LogicalBlockSize);
	  getitimer(ITIMER_REAL, &endTime);
	  if(pErrorMsg != NO_ERROR_MSG)
	  {
	       free(u32pBuffer);
	       updateProgress(in_u32TestSizeMB);
	       return returnErrorMsg(ERROR_STORAGE_DAPI_READ_PERFORMANCE_TEST, NULL, pErrorMsg);
	  }

	  dElapsed = dElapsed + (startTime.it_value.tv_sec - endTime.it_value.tv_sec) + ((startTime.it_value.tv_usec - endTime.it_value.tv_usec) / 1000000.0);

	  if((pErrorMsg = m_pSCSISenseAnalyzer->setSenseDataByIndex(in_u32DefinedIndex, (UCHAR *)m_pSCSIFactory->getCurrentSenseBuffer())) != NO_ERROR_MSG)
	  {
	       free(u32pBuffer);
	       updateProgress(in_u32TestSizeMB);
	       return returnErrorMsg(ERROR_STORAGE_DAPI_READ_PERFORMANCE_TEST, NULL, pErrorMsg);
	  }

	  u64StartSector += m_u32StorageSectorCountPerRDWRAccess;
	  if(u64StartSector > u64LastSector)
	  {
	       u64StartSector = in_u64StartSector;
	  }
	  u64TotalByte += m_u32StorageSectorCountPerRDWRAccess*u64LogicalBlockSize;
	  if((u64TotalByte % (2048*u64LogicalBlockSize)) == 0)
	  {
	       updateProgress(1);//every 1 MB update progress
	  }
     }while(u64TotalByte < ((UINT64)(in_u32TestSizeMB))*2048*u64LogicalBlockSize);

     // release resource
     free(u32pBuffer);
     if(close(s32FileDescriptor) < 0)
     {
	  errorCode = ERROR_STORAGE_DAPI_CLOSE_DEVICE_FAIL;
	  return returnErrorMsg((ERRORCODE)(ERROR_STORAGE_DAPI_READ_PERFORMANCE_TEST | errorCode), NULL, NULL);
     }

     ssTemp.str("");
     ssTemp.clear();
     if(in_blsMultiModeOutputFormat == 1)
     {
	  string strDefinedName = "";
	  if((pErrorMsg = m_pStorageDDAPIOperation->getNameByIndex(in_u32DefinedIndex, &strDefinedName)) != NO_ERROR_MSG)
	  {
	       return returnErrorMsg(ERROR_STORAGE_DAPI_READ_PERFORMANCE_TEST, NULL, pErrorMsg);	       
	  }

	  ssTemp << strDefinedName << ": ";
	  if((u64TotalByte / dElapsed) > 1.0)//more than 1MB/s
	  {
	       ssTemp << fixed << setprecision(2) << dElapsed << "s\t";
	       ssTemp << fixed << setprecision(2) << u64TotalByte/(2048*u64LogicalBlockSize)/dElapsed << "MB/s\t";
	       m_strOutputResultString = ssTemp.str();
	  }
	  else
	  {
	       ssTemp << fixed << setprecision(2) << dElapsed << "s\t";
	       ssTemp << fixed << setprecision(2) << u64TotalByte/1024/dElapsed << "KB/s\t";
	       m_strOutputResultString = ssTemp.str();
	  }
     }
     else
     {
	  if((u64TotalByte / dElapsed) > 1.0)//more than 1MB/s
	  {
	       ssTemp << "Spend time : " << fixed << setprecision(2) << dElapsed << " s" << endl;
	       ssTemp << "Speed      : " << fixed << setprecision(2) << u64TotalByte/(2048*u64LogicalBlockSize)/dElapsed << " MB/s" << endl;
	       m_strOutputResultString = ssTemp.str();
	  }
	  else
	  {
	       ssTemp << "Spend time : " << fixed << setprecision(2) << dElapsed << " s" << endl;
	       ssTemp << "Speed      : " << fixed << setprecision(2) << u64TotalByte/1024/dElapsed << " KB/s" << endl;
	       m_strOutputResultString = ssTemp.str();
	  }

     }

     if((pErrorMsg = m_pSCSISenseAnalyzer->listSenseDataByIndex(in_u32DefinedIndex, in_blsMultiModeOutputFormat)) != NO_ERROR_MSG)
     {
	  m_strOutputResultString += m_pSCSISenseAnalyzer->m_strOutputResultString;
	  return returnErrorMsg(ERROR_STORAGE_DAPI_READ_PERFORMANCE_TEST, NULL, pErrorMsg);
     }
     m_strOutputResultString += m_pSCSISenseAnalyzer->m_strOutputResultString;

     return returnErrorMsg(errorCode, NULL, NULL);
}

/// @fn         ErrorMsg *StorageDAPI::writePerformanceTest(UINT32 in_u32DefinedIndex, UINT64 in_u64StartSector, UINT32 in_u32TestSize)
/// @brief      Write specific size(MB) of data from starting logical sector on hard disk through index to get read performance (MB/s)
/// @param      in_u32DefinedIndex is a index for Storage device node name (eg. 1)
/// @param      in_u64StartSector is starting sector to test
/// @param      in_u32TestSizeMB is test size by MB
/// @exception  Nothing
/// @return     A pointer to ErrorMsg for error message list
ErrorMsg *StorageDAPI::writePerformanceTest(UINT32 in_u32DefinedIndex, UINT64 in_u64StartSector, UINT32 in_u32TestSizeMB, BOOL in_blsMultiModeOutputFormat)
{
     ERRORCODE errorCode = NO_ERROR;
     ErrorMsg *pErrorMsg = NO_ERROR_MSG;

     if((pErrorMsg = setSectorCountPerRDWRAccess(in_u32DefinedIndex)) != NO_ERROR_MSG)
     {
	  return returnErrorMsg(ERROR_STORAGE_DAPI_WRITE_PERFORMANCE_TEST, NULL, pErrorMsg);
     }

     string strSGDeviceNodeName = "";
     if((pErrorMsg = m_pStorageDDAPIOperation->getElementDataByIndex(in_u32DefinedIndex, eStorageDriveSGDeviceNodeName, &strSGDeviceNodeName)) != NO_ERROR_MSG)
     {
	  return returnErrorMsg(ERROR_STORAGE_DAPI_WRITE_PERFORMANCE_TEST, NULL, pErrorMsg);
     }

     string strTempString = "";
     if((pErrorMsg = m_pStorageDDAPIOperation->getElementDataByIndex(in_u32DefinedIndex, eStorageDriveCapacity, &strTempString)) != NO_ERROR_MSG)
     {
	  return returnErrorMsg(ERROR_STORAGE_DAPI_WRITE_PERFORMANCE_TEST, NULL, pErrorMsg);
     }
     //strTempString = strTempString.substr(0, strTempString.size()-3);
     
     stringstream ssTemp;
     ssTemp.str("");
     ssTemp.clear();
     ssTemp << strTempString;
     UINT64 u64CapacityMB = 0;
     ssTemp >> u64CapacityMB;

     strTempString = "";
     if((pErrorMsg = m_pStorageDDAPIOperation->getElementDataByIndex(in_u32DefinedIndex, eStorageDriveLogicalBlockSize, &strTempString)) != NO_ERROR_MSG)
     {
	  return returnErrorMsg(ERROR_STORAGE_DAPI_WRITE_PERFORMANCE_TEST, NULL, pErrorMsg);
     }
     ssTemp.str("");
     ssTemp.clear();
     ssTemp << strTempString;
     UINT64 u64LogicalBlockSize = 0;
     ssTemp >> u64LogicalBlockSize;

     UINT64 u64StartSector = in_u64StartSector;
     UINT64 u64LastSector = (u64CapacityMB-1)*1024*1024/u64LogicalBlockSize;//2048;

     if(u64StartSector > u64LastSector)
     {
	  u64StartSector = u64StartSector - u64LastSector;
     }

     struct itimerval startTime, endTime, setting;
     UINT64 u64TotalByte = 0;
     double dElapsed = 0;
     UINT32 *u32pBuffer = (UINT32 *)malloc(m_u32StorageSectorCountPerRDWRAccess*u64LogicalBlockSize);
     for(UINT32 u32CNT=0; u32CNT<m_u32StorageSectorCountPerRDWRAccess*u64LogicalBlockSize/4; u32CNT++)
     {
	  u32pBuffer[u32CNT] = 0x5555AAAA;// test pattern is 0xAAAA5555 (little-endian)
     }

     INT32 s32FileDescriptor = open(strSGDeviceNodeName.c_str(), O_RDWR | O_SYNC | O_NONBLOCK);
     if(s32FileDescriptor < 0)
     {
	  free(u32pBuffer);
	  errorCode = ERROR_STORAGE_DAPI_OPEN_DEVICE_FAIL;
	  return returnErrorMsg((ERRORCODE)(ERROR_STORAGE_DAPI_WRITE_PERFORMANCE_TEST | errorCode), NULL, NULL);
     }

     setting.it_interval.tv_sec = numeric_limits<time_t>::max();//1000;
     setting.it_interval.tv_usec = 0;
     setting.it_value.tv_sec = numeric_limits<time_t>::max();//1000;
     setting.it_value.tv_usec = 0;

     // setitimer(ITIMER_REAL, &(struct itimerval){{1000,0},{1000,0}}, NULL);
     setitimer(ITIMER_REAL, &setting, NULL);
     // Now do the timings for real
     initProgress(in_u32TestSizeMB);
     do
     {
	  getitimer(ITIMER_REAL, &startTime);
	  pErrorMsg = m_pSCSIFactory->executeSCSIWrite(s32FileDescriptor, u64StartSector, m_u32StorageSectorCountPerRDWRAccess, (UINT8*)u32pBuffer, m_u32StorageSectorCountPerRDWRAccess*u64LogicalBlockSize);
	  getitimer(ITIMER_REAL, &endTime);
	  if(pErrorMsg != NO_ERROR_MSG)
	  {
	       free(u32pBuffer);
	       updateProgress(in_u32TestSizeMB);
	       return returnErrorMsg(ERROR_STORAGE_DAPI_WRITE_PERFORMANCE_TEST, NULL, pErrorMsg);
	  }

	  dElapsed = dElapsed + (startTime.it_value.tv_sec - endTime.it_value.tv_sec) + ((startTime.it_value.tv_usec - endTime.it_value.tv_usec) / 1000000.0);

	  if((pErrorMsg = m_pSCSISenseAnalyzer->setSenseDataByIndex(in_u32DefinedIndex, (UCHAR *)m_pSCSIFactory->getCurrentSenseBuffer())) != NO_ERROR_MSG)
	  {
	       free(u32pBuffer);
	       updateProgress(in_u32TestSizeMB);
	       return returnErrorMsg(ERROR_STORAGE_DAPI_WRITE_PERFORMANCE_TEST, NULL, pErrorMsg);
	  }

	  u64StartSector += m_u32StorageSectorCountPerRDWRAccess;
	  if(u64StartSector > u64LastSector)
	  {
	       u64StartSector = in_u64StartSector;
	  }
	  u64TotalByte += m_u32StorageSectorCountPerRDWRAccess*u64LogicalBlockSize;
	  if((u64TotalByte % (2048*u64LogicalBlockSize)) == 0)
	  {
	       updateProgress(1);//every 1 MB update progress
	  }
     }while(u64TotalByte < ((UINT64)(in_u32TestSizeMB))*2048*u64LogicalBlockSize);

     // release resource
     free(u32pBuffer);
     if(close(s32FileDescriptor) < 0)
     {
	  errorCode = ERROR_STORAGE_DAPI_CLOSE_DEVICE_FAIL;
	  return returnErrorMsg((ERRORCODE)(ERROR_STORAGE_DAPI_WRITE_PERFORMANCE_TEST | errorCode), NULL, NULL);
     }

     ssTemp.str("");
     ssTemp.clear();
     if(in_blsMultiModeOutputFormat == 1)
     {
	  string strDefinedName = "";
	  if((pErrorMsg = m_pStorageDDAPIOperation->getNameByIndex(in_u32DefinedIndex, &strDefinedName)) != NO_ERROR_MSG)
	  {
	       return returnErrorMsg(ERROR_STORAGE_DAPI_WRITE_PERFORMANCE_TEST, NULL, pErrorMsg);	       
	  }

	  ssTemp << strDefinedName << ": ";
	  if((u64TotalByte / dElapsed) > 1.0)//more than 1MB/s
	  {
	       ssTemp << fixed << setprecision(2) << dElapsed << "s\t";
	       ssTemp << fixed << setprecision(2) << u64TotalByte/(2048*u64LogicalBlockSize)/dElapsed << "MB/s\t";
	       m_strOutputResultString = ssTemp.str();
	  }
	  else
	  {
	       ssTemp << fixed << setprecision(2) << dElapsed << "s\t";
	       ssTemp << fixed << setprecision(2) << u64TotalByte/1024/dElapsed << "KB/s\t";
	       m_strOutputResultString = ssTemp.str();
	  }
     }
     else
     {
	  if((u64TotalByte / dElapsed) > 1.0)//more than 1MB/s
	  {
	       ssTemp << "Spend time : " << fixed << setprecision(2) << dElapsed << " s" << endl;
	       ssTemp << "Speed      : " << fixed << setprecision(2) << u64TotalByte/(2048*u64LogicalBlockSize)/dElapsed << " MB/s" << endl;
	       m_strOutputResultString = ssTemp.str();
	  }
	  else
	  {
	       ssTemp << "Spend time : " << fixed << setprecision(2) << dElapsed << " s" << endl;
	       ssTemp << "Speed      : " << fixed << setprecision(2) << u64TotalByte/1024/dElapsed << " KB/s" << endl;
	       m_strOutputResultString = ssTemp.str();
	  }
     }

     if((pErrorMsg = m_pSCSISenseAnalyzer->listSenseDataByIndex(in_u32DefinedIndex, in_blsMultiModeOutputFormat)) != NO_ERROR_MSG)
     {
	  m_strOutputResultString += m_pSCSISenseAnalyzer->m_strOutputResultString;
	  return returnErrorMsg(ERROR_STORAGE_DAPI_WRITE_PERFORMANCE_TEST, NULL, pErrorMsg);
     }
     m_strOutputResultString += m_pSCSISenseAnalyzer->m_strOutputResultString;

     return returnErrorMsg(errorCode, NULL, NULL);
}

/// @fn         ErrorMsg *StorageDAPI::readwriteTest(UINT32 in_u32DefinedIndex, UINT64 in_u64StartSector, UINT32 in_u32TestSizeMB, BOOL in_blsMultiModeOutputFormat)
/// @brief      Write specific size(MB) of data from starting logical sector on hard disk through index and read it back to compare data
/// @param      in_u32DefinedIndex is a index for Storage device node name (eg. 1)
/// @param      in_u64StartSector is starting sector to test
/// @param      in_u32TestSizeMB is test size by MB
/// @exception  Nothing
/// @return     A pointer to ErrorMsg for error message list
ErrorMsg *StorageDAPI::readwriteTest(UINT32 in_u32DefinedIndex, UINT64 in_u64StartSector, UINT32 in_u32TestSizeMB, BOOL in_blsMultiModeOutputFormat)
{
     ERRORCODE errorCode = NO_ERROR;
     ErrorMsg *pErrorMsg = NO_ERROR_MSG;

     if((pErrorMsg = setSectorCountPerRDWRAccess(in_u32DefinedIndex)) != NO_ERROR_MSG)
     {
	  return returnErrorMsg(ERROR_STORAGE_DAPI_READ_WRITE_TEST, NULL, pErrorMsg);
     }

     string strSGDeviceNodeName = "";
     if((pErrorMsg = m_pStorageDDAPIOperation->getElementDataByIndex(in_u32DefinedIndex, eStorageDriveSGDeviceNodeName, &strSGDeviceNodeName)) != NO_ERROR_MSG)
     {
	  return returnErrorMsg(ERROR_STORAGE_DAPI_READ_WRITE_TEST, NULL, pErrorMsg);
     }

     string strTempString = "";
     if((pErrorMsg = m_pStorageDDAPIOperation->getElementDataByIndex(in_u32DefinedIndex, eStorageDriveCapacity, &strTempString)) != NO_ERROR_MSG)
     {
	  return returnErrorMsg(ERROR_STORAGE_DAPI_READ_WRITE_TEST, NULL, pErrorMsg);
     }
     //strTempString = strTempString.substr(0, strTempString.size()-2);
     
     stringstream ssTemp;
     ssTemp.str("");
     ssTemp.clear();
     ssTemp << strTempString;
     UINT64 u64CapacityMB = 0;
     ssTemp >> u64CapacityMB;

     strTempString = "";
     if((pErrorMsg = m_pStorageDDAPIOperation->getElementDataByIndex(in_u32DefinedIndex, eStorageDriveLogicalBlockSize, &strTempString)) != NO_ERROR_MSG)
     {
	  return returnErrorMsg(ERROR_STORAGE_DAPI_READ_WRITE_TEST, NULL, pErrorMsg);
     }
     ssTemp.str("");
     ssTemp.clear();
     ssTemp << strTempString;
     UINT64 u64LogicalBlockSize = 0;
     ssTemp >> u64LogicalBlockSize;

     UINT64 u64StartSector = in_u64StartSector;
     UINT64 u64LastSector = (u64CapacityMB-1)*1024*1024/u64LogicalBlockSize;//2048;

     if(u64StartSector > u64LastSector)
     {
	  u64StartSector = u64StartSector - u64LastSector;
     }

     struct itimerval startTime, endTime, setting;
     UINT64 u64TotalByte = 0;
     double dElapsed = 0;
     UINT32 *u32pBuffer = (UINT32 *)malloc(m_u32StorageSectorCountPerRDWRAccess*u64LogicalBlockSize);
     for(UINT32 u32CNT=0; u32CNT<m_u32StorageSectorCountPerRDWRAccess*u64LogicalBlockSize/4;u32CNT++)
     {
	  u32pBuffer[u32CNT] = 0xAAAA5555;// test pattern is 0x5555AAAA (little-endian)
     }

     UINT32 *u32pBufferRead = (UINT32 *)malloc(m_u32StorageSectorCountPerRDWRAccess*u64LogicalBlockSize);
     INT32 s32FileDescriptor = open(strSGDeviceNodeName.c_str(), O_RDWR | O_SYNC | O_NONBLOCK);
     if(s32FileDescriptor < 0)
     {
	  free(u32pBuffer);
	  free(u32pBufferRead);
	  errorCode = ERROR_STORAGE_DAPI_OPEN_DEVICE_FAIL;
	  return returnErrorMsg((ERRORCODE)(ERROR_STORAGE_DAPI_READ_WRITE_TEST | errorCode), NULL, NULL);
     }

     setting.it_interval.tv_sec = numeric_limits<time_t>::max();//1000;
     setting.it_interval.tv_usec = 0;
     setting.it_value.tv_sec = numeric_limits<time_t>::max();//1000;
     setting.it_value.tv_usec = 0;

     // setitimer(ITIMER_REAL, &(struct itimerval){{1000,0},{1000,0}}, NULL);
     setitimer(ITIMER_REAL, &setting, NULL);
     // Now do the timings for real
     initProgress(in_u32TestSizeMB);
     do
     {
	  getitimer(ITIMER_REAL, &startTime);
	  pErrorMsg = m_pSCSIFactory->executeSCSIWrite(s32FileDescriptor, u64StartSector, m_u32StorageSectorCountPerRDWRAccess, (UINT8*)u32pBuffer, m_u32StorageSectorCountPerRDWRAccess*u64LogicalBlockSize);
	  if(pErrorMsg != NULL)
	  {
	       free(u32pBuffer);
	       free(u32pBufferRead);
	       updateProgress(in_u32TestSizeMB);
	       return returnErrorMsg(ERROR_STORAGE_DAPI_READ_WRITE_TEST, NULL, pErrorMsg);
	  }

	  if((pErrorMsg = m_pSCSISenseAnalyzer->setSenseDataByIndex(in_u32DefinedIndex, (UCHAR *)m_pSCSIFactory->getCurrentSenseBuffer())) != NO_ERROR_MSG)
	  {
	       free(u32pBuffer);
	       updateProgress(in_u32TestSizeMB);
	       return returnErrorMsg(ERROR_STORAGE_DAPI_READ_PERFORMANCE_TEST, NULL, pErrorMsg);
	  }

	  pErrorMsg = m_pSCSIFactory->executeSCSIRead(s32FileDescriptor, u64StartSector, m_u32StorageSectorCountPerRDWRAccess, (UINT8*)u32pBufferRead, m_u32StorageSectorCountPerRDWRAccess*u64LogicalBlockSize);
	  getitimer(ITIMER_REAL, &endTime);
	  if(pErrorMsg != NULL)
	  {
	       free(u32pBuffer);
	       free(u32pBufferRead);
	       updateProgress(in_u32TestSizeMB);
	       return returnErrorMsg(ERROR_STORAGE_DAPI_READ_WRITE_TEST, NULL, pErrorMsg);
	  }

	  if((pErrorMsg = m_pSCSISenseAnalyzer->setSenseDataByIndex(in_u32DefinedIndex, (UCHAR *)m_pSCSIFactory->getCurrentSenseBuffer())) != NO_ERROR_MSG)
	  {
	       free(u32pBuffer);
	       updateProgress(in_u32TestSizeMB);
	       return returnErrorMsg(ERROR_STORAGE_DAPI_READ_PERFORMANCE_TEST, NULL, pErrorMsg);
	  }

	  dElapsed = dElapsed + (startTime.it_value.tv_sec - endTime.it_value.tv_sec) + ((startTime.it_value.tv_usec - endTime.it_value.tv_usec) / 1000000.0);


	  if(memcmp(u32pBuffer, u32pBufferRead, m_u32StorageSectorCountPerRDWRAccess*u64LogicalBlockSize) != 0)
	  {
	       free(u32pBuffer);
	       free(u32pBufferRead);
	       updateProgress(in_u32TestSizeMB);
	       errorCode = ERROR_STORAGE_DAPI_COMPARE_DATA_ERROR;
	       return returnErrorMsg((ERRORCODE)(ERROR_STORAGE_DAPI_READ_WRITE_TEST | errorCode), NULL, NULL);
	  }


	  u64StartSector += m_u32StorageSectorCountPerRDWRAccess;
	  if(u64StartSector > u64LastSector)
	  {
	       u64StartSector = in_u64StartSector;
	  }
	  u64TotalByte += m_u32StorageSectorCountPerRDWRAccess*u64LogicalBlockSize;
	  if((u64TotalByte % (2048*u64LogicalBlockSize)) == 0)
	  {
	       updateProgress(1);//every 1 MB update progress
	  }
     }while(u64TotalByte < ((UINT64)(in_u32TestSizeMB))*2048*u64LogicalBlockSize);

     // release resource
     free(u32pBuffer);
     free(u32pBufferRead);
     if(close(s32FileDescriptor) < 0)
     {
	  errorCode = ERROR_STORAGE_DAPI_CLOSE_DEVICE_FAIL;
	  return returnErrorMsg((ERRORCODE)(ERROR_STORAGE_DAPI_READ_WRITE_TEST | errorCode), NULL, NULL);
     }

     ssTemp.str("");
     ssTemp.clear();
     if(in_blsMultiModeOutputFormat == 1)
     {
	  string strDefinedName = "";
	  if((pErrorMsg = m_pStorageDDAPIOperation->getNameByIndex(in_u32DefinedIndex, &strDefinedName)) != NO_ERROR_MSG)
	  {
	       return returnErrorMsg(ERROR_STORAGE_DAPI_READ_WRITE_TEST, NULL, pErrorMsg);	       
	  }

	  ssTemp << strDefinedName << ": ";
	  if((u64TotalByte / dElapsed) > 1.0)//more than 1MB/s
	  {
	       ssTemp << fixed << setprecision(2) << dElapsed << "s\t";
	       ssTemp << fixed << setprecision(2) << u64TotalByte/(2048*u64LogicalBlockSize)/dElapsed << "MB/s\t";
	       m_strOutputResultString = ssTemp.str();
	  }
	  else
	  {
	       ssTemp << fixed << setprecision(2) << dElapsed << "s\t";
	       ssTemp << fixed << setprecision(2) << u64TotalByte/1024/dElapsed << "MB/s\t";
	       m_strOutputResultString = ssTemp.str();
	  }
     }
     else
     {
	  if((u64TotalByte / dElapsed) > 1.0)//more than 1MB/s
	  {
	       ssTemp << "Spend time : " << fixed << setprecision(2) << dElapsed << " s" << endl;
	       ssTemp << "Speed      : " << fixed << setprecision(2) << u64TotalByte/(2048*u64LogicalBlockSize)/dElapsed << " MB/s" << endl;
	       m_strOutputResultString = ssTemp.str();
	  }
	  else
	  {
	       ssTemp << "Spend time : " << fixed << setprecision(2) << dElapsed << " s" << endl;
	       ssTemp << "Speed      : " << fixed << setprecision(2) << u64TotalByte/1024/dElapsed << " MB/s" << endl;
	       m_strOutputResultString = ssTemp.str();
	  }
     }

     if((pErrorMsg = m_pSCSISenseAnalyzer->listSenseDataByIndex(in_u32DefinedIndex, in_blsMultiModeOutputFormat)) != NO_ERROR_MSG)
     {
	  m_strOutputResultString += m_pSCSISenseAnalyzer->m_strOutputResultString;
	  return returnErrorMsg(ERROR_STORAGE_DAPI_WRITE_PERFORMANCE_TEST, NULL, pErrorMsg);
     }
     m_strOutputResultString += m_pSCSISenseAnalyzer->m_strOutputResultString;

     return returnErrorMsg(errorCode, NULL, NULL);
}

/// @fn         ErrorMsg *StorageDAPI::writeZeroByIndex(UINT32 in_u32DefinedIndex, UINT64 in_u64StartSector, BOOL in_blsMultiModeOutputFormat)
/// @brief      Write specific size(MB) of data from starting logical sector on hard disk through index to get read performance (MB/s)
/// @param      in_u32DefinedIndex is a index for Storage device node name (eg. 1)
/// @param      in_u64StartSector is starting sector to test
/// @param      in_u32TestSizeMB is test size by MB
/// @exception  Nothing
/// @return     A pointer to ErrorMsg for error message list
ErrorMsg *StorageDAPI::writeZeroByIndex(UINT32 in_u32DefinedIndex, UINT64 in_u64StartSector, BOOL in_blsMultiModeOutputFormat)
{
     ERRORCODE errorCode = NO_ERROR;
     ErrorMsg *pErrorMsg = NO_ERROR_MSG;

     UINT32 u32ControllerType = eStorageControllerTypeUnknown;
     if((pErrorMsg = m_pStorageDDAPIOperation->getControllerTypeByIndex(in_u32DefinedIndex, &u32ControllerType)) != NO_ERROR_MSG)
     {
	  return returnErrorMsg(ERROR_STORAGE_DAPI_WRITE_ZERO_BY_INDEX, NULL, pErrorMsg);
     }

     if(u32ControllerType == (UINT32)eStorageControllerTypeUSB)
     {
	  m_u32StorageSectorCountPerRDWRAccess = 128;
     }
     else
     {
	  m_u32StorageSectorCountPerRDWRAccess = 4096;
     }

     string strSGDeviceNodeName = "";
     if((pErrorMsg = m_pStorageDDAPIOperation->getElementDataByIndex(in_u32DefinedIndex, eStorageDriveSGDeviceNodeName, &strSGDeviceNodeName)) != NO_ERROR_MSG)
     {
	  return returnErrorMsg(ERROR_STORAGE_DAPI_WRITE_ZERO_BY_INDEX, NULL, pErrorMsg);
     }

     string strTempString = "";
     if((pErrorMsg = m_pStorageDDAPIOperation->getElementDataByIndex(in_u32DefinedIndex, eStorageDriveCapacity, &strTempString)) != NO_ERROR_MSG)
     {
	  return returnErrorMsg(ERROR_STORAGE_DAPI_WRITE_ZERO_BY_INDEX, NULL, pErrorMsg);
     }

     stringstream ssTemp;
     ssTemp.str("");
     ssTemp.clear();
     ssTemp << strTempString;
     UINT64 u64CapacityMB = 0;
     ssTemp >> u64CapacityMB;

     strTempString = "";
     if((pErrorMsg = m_pStorageDDAPIOperation->getElementDataByIndex(in_u32DefinedIndex, eStorageDriveLogicalBlockSize, &strTempString)) != NO_ERROR_MSG)
     {
	  return returnErrorMsg(ERROR_STORAGE_DAPI_WRITE_ZERO_BY_INDEX, NULL, pErrorMsg);
     }

     ssTemp.str("");
     ssTemp.clear();
     ssTemp << strTempString;
     UINT64 u64LogicalBlockSize = 0;
     ssTemp >> u64LogicalBlockSize;

     UINT64 u64StartSector = in_u64StartSector;
     UINT64 u64LastSector = (u64CapacityMB-2)*1024*1024/u64LogicalBlockSize;//2048;

     if(u64StartSector > u64LastSector)
     {
	  u64StartSector = u64StartSector - u64LastSector;
     }

     UINT64 u64TotalByte = 0;
     double dElapsed = 0;
     UINT32 *u32pBuffer = (UINT32 *)malloc(m_u32StorageSectorCountPerRDWRAccess*u64LogicalBlockSize);
     for(UINT32 u32CNT=0; u32CNT<m_u32StorageSectorCountPerRDWRAccess*u64LogicalBlockSize/4; u32CNT++)
     {
	  u32pBuffer[u32CNT] = 0x0;
     }

     INT32 s32FileDescriptor = open(strSGDeviceNodeName.c_str(), O_RDWR | O_SYNC | O_NONBLOCK);
     if(s32FileDescriptor < 0)
     {
	  free(u32pBuffer);
	  errorCode = ERROR_STORAGE_DAPI_OPEN_DEVICE_FAIL;
	  return returnErrorMsg((ERRORCODE)(ERROR_STORAGE_DAPI_WRITE_ZERO_BY_INDEX | errorCode), NULL, NULL);
     }

     // Now do the timings for real
     initProgress(u64CapacityMB);
     do
     {
	  pErrorMsg = m_pSCSIFactory->executeSCSIWrite(s32FileDescriptor, u64StartSector, m_u32StorageSectorCountPerRDWRAccess, (UINT8*)u32pBuffer, m_u32StorageSectorCountPerRDWRAccess*u64LogicalBlockSize);

	  if(pErrorMsg != NO_ERROR_MSG)
	  {
	       free(u32pBuffer);
	       updateProgress(u64CapacityMB);
	       return returnErrorMsg(ERROR_STORAGE_DAPI_WRITE_ZERO_BY_INDEX, NULL, pErrorMsg);
	  }

	  if((pErrorMsg = m_pSCSISenseAnalyzer->setSenseDataByIndex(in_u32DefinedIndex, (UCHAR *)m_pSCSIFactory->getCurrentSenseBuffer())) != NO_ERROR_MSG)
	  {
	       free(u32pBuffer);
	       updateProgress(u64CapacityMB);
	       return returnErrorMsg(ERROR_STORAGE_DAPI_WRITE_ZERO_BY_INDEX, NULL, pErrorMsg);
	  }

	  u64StartSector += m_u32StorageSectorCountPerRDWRAccess;
	  if(u64StartSector > u64LastSector)
	  {
	       u64StartSector = in_u64StartSector;
	  }

	  u64TotalByte += m_u32StorageSectorCountPerRDWRAccess*u64LogicalBlockSize;
	  if((u64TotalByte % (4096*u64LogicalBlockSize)) == 0)
	  {
	       updateProgress(2);//every 2 MB update progress
	  }
     }while(u64TotalByte < ((UINT64)(u64CapacityMB))*2048*u64LogicalBlockSize);

     // release resource
     free(u32pBuffer);
     if(close(s32FileDescriptor) < 0)
     {
	  errorCode = ERROR_STORAGE_DAPI_CLOSE_DEVICE_FAIL;
	  return returnErrorMsg((ERRORCODE)(ERROR_STORAGE_DAPI_WRITE_ZERO_BY_INDEX | errorCode), NULL, NULL);
     }

     ssTemp.str("");
     ssTemp.clear();
     if(in_blsMultiModeOutputFormat == 1)
     {
	  string strDefinedName = "";
	  if((pErrorMsg = m_pStorageDDAPIOperation->getNameByIndex(in_u32DefinedIndex, &strDefinedName)) != NO_ERROR_MSG)
	  {
	       return returnErrorMsg(ERROR_STORAGE_DAPI_WRITE_ZERO_BY_INDEX, NULL, pErrorMsg);	       
	  }
	  m_strOutputResultString = strDefinedName + ": ";
/*
	  ssTemp << strDefinedName << ": ";
	  if((u64TotalByte / dElapsed) > 1.0)//more than 1MB/s
	  {
	       ssTemp << fixed << setprecision(2) << dElapsed << "s\t";
	       ssTemp << fixed << setprecision(2) << u64TotalByte/(2048*u64LogicalBlockSize)/dElapsed << "MB/s\t";
	       m_strOutputResultString = ssTemp.str();
	  }
	  else
	  {
	       ssTemp << fixed << setprecision(2) << dElapsed << "s\t";
	       ssTemp << fixed << setprecision(2) << u64TotalByte/1024/dElapsed << "KB/s\t";
	       m_strOutputResultString = ssTemp.str();
	  }
*/
     }
     else
     {
/*
	  if((u64TotalByte / dElapsed) > 1.0)//more than 1MB/s
	  {
//	       ssTemp << "Spend time : " << fixed << setprecision(2) << dElapsed << " s" << endl;
//	       ssTemp << "Speed      : " << fixed << setprecision(2) << u64TotalByte/(2048*u64LogicalBlockSize)/dElapsed << " MB/s" << endl;
	       m_strOutputResultString = ssTemp.str();
	  }
	  else
	  {
//	       ssTemp << "Spend time : " << fixed << setprecision(2) << dElapsed << " s" << endl;
//	       ssTemp << "Speed      : " << fixed << setprecision(2) << u64TotalByte/1024/dElapsed << " KB/s" << endl;
	       m_strOutputResultString = ssTemp.str();
	  }
*/
     }

     if((pErrorMsg = m_pSCSISenseAnalyzer->listSenseDataByIndex(in_u32DefinedIndex, in_blsMultiModeOutputFormat)) != NO_ERROR_MSG)
     {
	  return returnErrorMsg(ERROR_STORAGE_DAPI_WRITE_ZERO_BY_INDEX, NULL, pErrorMsg);
     }
     m_strOutputResultString += m_pSCSISenseAnalyzer->m_strOutputResultString;

     return returnErrorMsg(errorCode, NULL, NULL);
}

ErrorMsg *StorageDAPI::getPGListByIndex(UINT32 in_u32DefinedIndex, UINT32 in_u32RequestPList, UINT32 in_u32RequestGList, UINT32 in_u32Format, BOOL in_blsMultiModeOutputFormat)
{
     ERRORCODE errorCode = NO_ERROR;
     ErrorMsg *pErrorMsg = NO_ERROR_MSG;

     stringstream ssTemp;
     ssTemp.str("");
     ssTemp.clear();
     if(in_blsMultiModeOutputFormat == true)
     {
	  string strDefinedName = "";
	  if((pErrorMsg = m_pStorageDDAPIOperation->getNameByIndex(in_u32DefinedIndex, &strDefinedName)) != NO_ERROR_MSG)
	  {
	       return returnErrorMsg(ERROR_STORAGE_DAPI_GET_PGLIST_BY_INDEX, NULL, pErrorMsg);	       
	  }

	  ssTemp << strDefinedName << ": ";
	  m_strOutputResultString = ssTemp.str();
     }

     string strSGDeviceNodeName = "";
     if((pErrorMsg = m_pStorageDDAPIOperation->getElementDataByIndex(in_u32DefinedIndex, eStorageDriveSGDeviceNodeName, &strSGDeviceNodeName)) != NO_ERROR_MSG)
     {
	  return returnErrorMsg(ERROR_STORAGE_DAPI_GET_PGLIST_BY_INDEX, NULL, pErrorMsg);
     }

     UINT32 *u32pBuffer = (UINT32 *)malloc(STORAGE_SCSICMD_BUFFER_SIZE);
     for(UINT32 u32CNT=0; u32CNT<STORAGE_SCSICMD_BUFFER_SIZE/4; u32CNT++)
     {
	  u32pBuffer[u32CNT] = 0x0;
     }

     INT32 s32FileDescriptor = open(strSGDeviceNodeName.c_str(), O_RDWR | O_SYNC | O_NONBLOCK);
     if(s32FileDescriptor < 0)
     {
	  free(u32pBuffer);
	  errorCode = ERROR_STORAGE_DAPI_OPEN_DEVICE_FAIL;
	  return returnErrorMsg((ERRORCODE)(ERROR_STORAGE_DAPI_GET_PGLIST_BY_INDEX | errorCode), NULL, NULL);
     }

     if((pErrorMsg = m_pSCSIFactory->scsiReadDefect12(s32FileDescriptor, in_u32RequestPList, in_u32RequestGList, in_u32Format/*4/*in_u32DLFormat*/, 0/*UINT32 in_u32AddrDescIndex*/, (UINT8 *)u32pBuffer, STORAGE_SCSICMD_BUFFER_SIZE/*UINT32 in_u32BufferLength*/)) != NO_ERROR_MSG)
     {
	  free(u32pBuffer);
	  return returnErrorMsg(ERROR_STORAGE_DAPI_GET_PGLIST_BY_INDEX, NULL, pErrorMsg);
     }

     if(close(s32FileDescriptor) < 0)
     {
	  free(u32pBuffer);
	  errorCode = ERROR_STORAGE_DAPI_CLOSE_DEVICE_FAIL;
	  return returnErrorMsg((ERRORCODE)(ERROR_STORAGE_DAPI_GET_PGLIST_BY_INDEX | errorCode), NULL, NULL);
     }

     string strTranslateHexString = "";
     if((errorCode = translateHexToString(&strTranslateHexString, &(u32pBuffer[1]), sizeof(UINT32)/*STORAGE_SCSICMD_BUFFER_SIZE/*UINT32 in_u32DataLengthInBytes*/, false)) != NO_ERROR)
     {
	  free(u32pBuffer);
	  return returnErrorMsg((ERRORCODE)(ERROR_STORAGE_DAPI_GET_PGLIST_BY_INDEX | errorCode), NULL, NULL);
     }


     // get PorG-List Count
     UINT32 u32PGListCount = 0;
     ssTemp.str("");
     ssTemp.clear();
     ssTemp << hex << strTranslateHexString;
     ssTemp >> u32PGListCount;
     ssTemp.str("");
     ssTemp.clear();
     ssTemp << dec << u32PGListCount;

     if(in_u32RequestPList==1 && in_u32RequestGList==0)
     {
	  m_strOutputResultString += "P-ListCount: " + ssTemp.str();
     }
     else if(in_u32RequestPList==0 && in_u32RequestGList==1)
     {
	  m_strOutputResultString += "G-ListCount: " + ssTemp.str();
     }
     else if(in_u32RequestPList==1 && in_u32RequestGList==1)
     {
	  m_strOutputResultString += "P&G-ListCount: " + ssTemp.str();
     }
     else
     {
	  free(u32pBuffer);
	  errorCode = ERROR_STORAGE_DAPI_NO_SUCH_IOCTL_OPERATION;
	  return returnErrorMsg((ERRORCODE)(ERROR_STORAGE_DAPI_GET_PGLIST_BY_INDEX | errorCode), NULL, NULL);
     }
     ssTemp.str("");
     ssTemp.clear();

     // get PorG-List Data
/*
     if((errorCode = dumpHexDataToString(&strDumpHexString, &(u32pBuffer[2]), u32PGListCount, "HEX", false)) != NO_ERROR)
     {
	  free(u32pBuffer);
	  return returnErrorMsg((ERRORCODE)(ERROR_STORAGE_DAPI_GET_PGLIST_BY_INDEX | errorCode), NULL, NULL);
     }

     for(UINT32 u32I=0; u32I<strDumpHexString.size(); u32I++)// delete empty character
     {
	  strDumpHexString.erase(std::remove(strDumpHexString.begin(), strDumpHexString.end(), ' '), strDumpHexString.end());
     }
     ssTemp.str("");
     ssTemp.clear();

     UINT32 u32ListLength;
     UINT32 u32AddNumber = 0;
     if(in_u32Format == 0)
     {
	  u32ListLength = 8;      //short type seems to return 8 bytes as well
     }
     else if(in_u32Format==0x3 || in_u32Format==0x4 ||in_u32Format==0x5)
     {
	  u32ListLength = 8;
     }
     else
     {
	  u32ListLength = 16;   //user-defined type, print 16 bytes in a row;
     }

     for(UINT32 u32I=0; u32I<u32PGListCount/u32ListLength; u32I++)
     {
	  if(u32ListLength > 8) //add a space between 8-byte data.
	  {
	       strDumpHexString.insert((u32ListLength*2+2)*u32I + 8*2 , " ");
	       u32AddNumber++;
	       strDumpHexString.insert(u32ListLength*2*(u32I+1) + u32I + u32AddNumber , "\n");
	  }
	  else
	  {
	       strDumpHexString.insert(u32ListLength*2*(u32I+1) + u32I , "\n");
	  }
     }
     ssTemp << hex << strDumpHexString;
     m_strOutputResultString = m_strOutputResultString + ssTemp.str();
*/
     free(u32pBuffer);
     return returnErrorMsg(errorCode, NULL, NULL);
}

ErrorMsg *StorageDAPI::setSectorCountPerRDWRAccess(UINT32 in_u32DefinedIndex)
{
     ERRORCODE errorCode = NO_ERROR;
     ErrorMsg *pErrorMsg = NO_ERROR_MSG;

     UINT32 u32ControllerType = eStorageControllerTypeUnknown;

     if((pErrorMsg = m_pStorageDDAPIOperation->getControllerTypeByIndex(in_u32DefinedIndex, &u32ControllerType)) != NO_ERROR_MSG)
     {
	  return returnErrorMsg(ERROR_STORAGE_DAPI_SET_SECTOR_COUNT_PER_RDWR_ACCESS, NULL, pErrorMsg);
     }

     switch(u32ControllerType)
     {
     case eStorageControllerTypeSAS:
     case eStorageControllerTypeSATA:
     case eStorageControllerTypeNVME:
	  m_u32StorageSectorCountPerRDWRAccess = STORAGE_SECTOR_COUNT_PER_RDWR_ACCESS;
	  break;
     case eStorageControllerTypeUSB:
	  m_u32StorageSectorCountPerRDWRAccess = STORAGE_USB_SECTOR_COUNT_PER_RDWR_ACCESS;
	  break;
     default:
	  errorCode = ERROR_STORAGE_DAPI_UNKNOWN_CONTROLLER_TYPE;
	  return returnErrorMsg((ERRORCODE)(ERROR_STORAGE_DAPI_SET_SECTOR_COUNT_PER_RDWR_ACCESS | errorCode), NULL, NULL);
     }

     return returnErrorMsg(errorCode, NULL, NULL);
}

ERRORCODE StorageDAPI::translateHexToString(string *out_strpDataString, UINT32 *in_u32pDataBuffer, UINT32 in_u32DataLengthInBytes, BOOL in_blsDumpAddressHead)
{
     ERRORCODE errorCode = NO_ERROR;

     stringstream ssTemp;
     ssTemp.str("");
     ssTemp.clear();

     for(UINT32 u32I=0; u32I<in_u32DataLengthInBytes + ((in_u32DataLengthInBytes % HEXDUMP_COLUMS) ? (HEXDUMP_COLUMS - in_u32DataLengthInBytes % HEXDUMP_COLUMS) : 0); u32I++)
     {
	  // set offset
	  if(in_blsDumpAddressHead == 1)
	  {
	       if(u32I%HEXDUMP_COLUMS == 0)
	       {
		    if(u32I != 0)
		    {
			 ssTemp << "\n";
		    }
		    ssTemp << hex << uppercase << setw(8) << setfill('0') << u32I;
		    ssTemp << ": ";
	       }
	  }
	  // output hex data
	  if(u32I < in_u32DataLengthInBytes)
	  {
	       ssTemp << hex << uppercase << setfill('0') << setw(2) << (0xff & ((char*)in_u32pDataBuffer)[u32I]);
	  }

	  if(u32I%4 == 3)
	  {
	       ssTemp << " ";
	  }
     }

     *out_strpDataString = ssTemp.str();
     return errorCode;
}

long double StorageDAPI::int128_to_double(__u8 *data)
{
	int i;
	long double result = 0;

	for (i = 0; i < 16; i++) {
		result *= 256;
		result += data[15 - i];
	}
	return result;
}
