#include "UARTDAPI.hpp"

UARTDAPI::UARTDAPI(STRING className) : BaseClassDAPI(className)
{
    m_pUartDriverAPI = new UARTDDAPI("uartDriver");
    
}

UARTDAPI::~UARTDAPI()
{
    if(m_pUartDriverAPI != NULL)
    {
        delete m_pUartDriverAPI;
        m_pUartDriverAPI = NULL;
    }
}


ErrorMsg*
UARTDAPI::getUartList(vector<SERIAL_INFO>*  out_pUartList)
{
    m_pErrorMsg = m_pUartDriverAPI->scanUarts(out_pUartList);
    if(m_pErrorMsg != NO_ERROR_MSG)
    {
        return returnErrorMsg(ERROR_UART_DAPI_GET_LIST, NULL, m_pErrorMsg);
    }
    return NO_ERROR_MSG;
}


ErrorMsg*
UARTDAPI::getUartInfoByName(STRING          in_uartLogicalName,
                            SERIAL_INFO*    out_pUartInfo)
{
    vector<SERIAL_INFO>             uartList;
    vector<SERIAL_INFO>::iterator   it;

    //printf("UARTDAPI::getUartInfoByName, in_uartLogicalName: %s\n", in_uartLogicalName);//check 

    m_pErrorMsg = m_pUartDriverAPI->scanUarts(&uartList);
    if(m_pErrorMsg != NO_ERROR_MSG)
    {
        return returnErrorMsg(ERROR_UART_DAPI_GET_INFO, NULL, m_pErrorMsg);
    }

    for(it = uartList.begin(); it != uartList.end(); ++it)
    {
        if(it->strUartName.compare(in_uartLogicalName) == 0)
        {
            *out_pUartInfo = *it;
            return NO_ERROR_MSG;
        }
    }

    return returnErrorMsg((ERRORCODE)(ERROR_UART_DAPI_GET_INFO | ERROR_NO_DEVICE_FOUND), NULL, NULL);
}


ErrorMsg*
UARTDAPI::testScratchPad(STRING   in_uartLogicalName)
{
    m_pErrorMsg = m_pUartDriverAPI->testScratchPad(in_uartLogicalName);
    if(m_pErrorMsg != NO_ERROR_MSG)
    {
        return returnErrorMsg(ERROR_UART_DAPI_SCRATCHPAD_TEST, NULL, m_pErrorMsg);
    }
    return NO_ERROR_MSG;
}


ErrorMsg*
UARTDAPI::doInternalLoopbackTest(STRING     in_uartLogicalName, 
                                 bool       in_interruptMode,
                                 UINT32     in_u32BufferSize,
                                 bool       in_userDefinedPattern,
                                 UINT8      in_u8Pattern)
{
    if(in_interruptMode)
    {
        //do internal loopback with interrupt mode 
    }
    else //polling mode  
    {
        m_pErrorMsg = m_pUartDriverAPI->doInternalLoopbackTest(in_uartLogicalName,
                                                               in_u32BufferSize,
                                                               in_userDefinedPattern,
                                                               in_u8Pattern);
    }
    if(m_pErrorMsg != NO_ERROR_MSG)
    {
        return returnErrorMsg(ERROR_UART_DAPI_INTERNAL_LOOPBACK_TEST, NULL, m_pErrorMsg);
    }
    return NO_ERROR_MSG;    
}

