/// @file       IOCTL.hpp
/// @brief      IOCTL class definition
/// @details    Provide IOCTL class member and function definition
/// @version    1.0.0
/// @date       2014.02.20
/// @author     JerryLee
/// @copyright  Copyright 2014 Hon Hai Technology Group. All Rights Reserved.

#ifndef IOCTL_HPP_
#define IOCTL_HPP_

#include "error_code.hpp"
#include "common.hpp"
#include <fcntl.h>
#include <sys/stat.h>
#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <stdarg.h>
#include <unistd.h>
#include <strings.h>
#include <sys/ioctl.h>
#include <errno.h>
#include <limits.h>
#include <dirent.h>
#include <linux/stddef.h>
#include <linux/types.h>
#include <string>

using namespace std;

#define INVALID_TEMPERATURE 0xFF
#define INVALID_CONTROLLER_ID 0xFF
#define INVALID_PHY_NUMBER 0xFF

// Error count definitions
typedef struct ERROR_COUNT_INFO {
        UINT32 InvalidDwordError;
        UINT32 DisparityError;
        UINT32 LossDwordSync;
        UINT32 PhyResetProblem;
} ErrorCountInfo;

class IOCTL
{
    public:
        IOCTL(STRING in_fileName);
        virtual ~IOCTL(void);

        enum IOCTL_OP_CODE
        {
            UnspportIOCTL
        };

        ERRORCODE getFileHandle(INT8* fileHandle);
        ERRORCODE getFileHandle(STRING fileName, INT8* fileHandle);
        UINT8 getControllerID(void);
        UINT8 getPHYNumber(void);
        void setControllerID(UINT8 iocNumber);
        void setPHYNumber(UINT8 phyNumber);
        ERRORCODE setFileHandle(STRING fileName);
        ERRORCODE closeFileHandle();

        virtual ERRORCODE allocateIOCTLArgument(UINT8 IOCTLOperation);
        virtual ERRORCODE setIOCTLOperationArgument(UINT8 IOCTLOperation, void* data = NULL);
        virtual ERRORCODE executeIOCTL(UINT8 IOCTLOperation);
        virtual ERRORCODE executeIOCTL(UINT8 IOCTLOperation, void* IOCTLResult);


    protected:
        INT8 myFileHandle;
        UINT8 targetControllerID;
        UINT8 targetPHYNumber;
        void* myIOCTLArgument;

    private:

    };

#endif /* IOCTL_HPP_ */
