#include "EthernetDDAPI.hpp"


EthernetDDAPI::EthernetDDAPI(STRING className) : EthernetBaseClass(className)
{
    m_ethernet_name = className;
}


EthernetDDAPI::~EthernetDDAPI()
{

}



ErrorMsg* EthernetDDAPI::GetEthernetMAC(string ethernet_name,
                                        string *mac_address)
{
    string vendor_ID;
    string device_ID;
    EthernetBaseClass *nic_interface = NULL;

    m_ethernet_name = ethernet_name;

    //Get ethernet PCI configuration vendor ID and device ID
    m_errorCode = get_vendor_ID(&vendor_ID);
    if (m_errorCode != NO_ERROR_MSG)
    {
       return returnErrorMsg(ERROR_ETHERNET_DDAPI_SENDIOCTL, "Get vender ID fail", m_errorCode);
    }

    m_errorCode = get_device_ID(&device_ID);
    if (m_errorCode != NO_ERROR_MSG)
    {
       return returnErrorMsg(ERROR_ETHERNET_DDAPI_SENDIOCTL, "Get device ID fail", m_errorCode);
    }
    //cout << "vendor_ID: " <<vendor_ID << "  device_ID:" << device_ID << endl;

    //Select which type of ethernet to create object
    if (vendor_ID.compare("0x8086") == 0 && (device_ID.compare("0x1533") == 0 || device_ID.compare("0x1537") == 0)) //INTEL I210
    {
        nic_interface = new INTEL_I210(m_ethernet_name.c_str());
    }
    if (vendor_ID.compare("0x8086") == 0 && device_ID.compare("0x15ab") == 0) //INTEL X552(KR)
    {
        nic_interface = new INTEL_X552(m_ethernet_name.c_str());
    }
    if (vendor_ID.compare("0x8086") == 0 && device_ID.compare("0x1583") == 0) //INTEL XL710
    {
        nic_interface = new INTEL_XL710(m_ethernet_name.c_str());
    }
    if (vendor_ID.compare("0x8086") == 0 && device_ID.compare("0x1572") == 0) //INTEL XL710
    {
        nic_interface = new INTEL_XL710(m_ethernet_name.c_str());
    }
    if (vendor_ID.compare("0x1425") == 0 && device_ID.compare("0x6407") == 0) //Chelsio T62100
    {
       //nic_interface = new CHELSIS_T62100(m_ethernet_name.c_str());
       return returnErrorMsg((ERRORCODE)(ERROR_ETHERNET_DDAPI_ENABLE_EXTERNAL_LOOPBACK | ERROR_FUNCTION_NOT_SUPPORTED));
    }
    if (vendor_ID.compare("0x1425") == 0 && device_ID.compare("0x640d") == 0) //Chelsio T62100-CR
    {
        //nic_interface = new CHELSIS_T62100(m_ethernet_name.c_str());
        return returnErrorMsg((ERRORCODE)(ERROR_ETHERNET_DDAPI_ENABLE_EXTERNAL_LOOPBACK | ERROR_FUNCTION_NOT_SUPPORTED));
    }
    if (vendor_ID.compare("0x1425") == 0 && device_ID.compare("0x6401") == 0) //Chelsio T6225-CR
    {
        //nic_interface = new CHELSIS_T62100(m_ethernet_name.c_str());
        return returnErrorMsg((ERRORCODE)(ERROR_ETHERNET_DDAPI_ENABLE_EXTERNAL_LOOPBACK | ERROR_FUNCTION_NOT_SUPPORTED));
    }

    if (nic_interface != NULL)
    {
        //Start setup speed
        m_errorCode = nic_interface->GetMAC(ethernet_name, mac_address);

        if (m_errorCode != NO_ERROR_MSG)
        {
           return returnErrorMsg(ERROR_ETHERNET_DDAPI_SENDIOCTL, "Setup speed fail", m_errorCode);
        }
    }
    else
    {
        return returnErrorMsg( (ERRORCODE)(ERROR_ETHERNET_DDAPI_SENDIOCTL | ERROR_ETHERNET_DDAPI_IOCTL_FAIL), "No define ethernet", NULL);
    }

    //Check MAC stress length
    if (mac_address->length() != 17)
    {
        return returnErrorMsg( (ERRORCODE)(ERROR_ETHERNET_DDAPI_SENDIOCTL | ERROR_ETHERNET_DDAPI_IOCTL_FAIL), "MAC length is incorrect", NULL);
    }

    delete nic_interface;

    return NO_ERROR_MSG;
}



ErrorMsg* EthernetDDAPI::GetEthernetNamesList(list<struct ethernet_data>*  ethernet_name_list)
{
    DIR     *dir;
    struct  dirent *ent;
    string  fileName, fileBDF, valueBDF, line, vendorID, deviceID, linePCI, controllerName;
    fstream file, PCIIDFile;
    int     find_vendorID = 0;
    struct  ethernet_data ethData;
    int     sockfd;

    struct ifreq           ifr;
    struct ethtool_drvinfo drvinf;

    if ((dir = opendir("/sys/class/net/")) != NULL)
    {
        /* print all the files and directories within directory */
        while ((ent = readdir (dir)) != NULL)
        {
             fileName = ent->d_name;
             //Filter ".", ".." and "lo"
             if (fileName != "." && fileName != ".." && fileName != "lo" && fileName != "sit0")
             {
                 //Combination the ethernet BDF "uevent" file path
                 fileBDF = "/sys/class/net/" + fileName + "/device/uevent";

                 file.open(fileBDF.c_str(), ios::in);
                 if (file.is_open())
                 {
                     while ( getline(file,line) )
                     {
                         //Find "PCI_SLOT_NAME" and get the value
                         if (line.find("PCI_SLOT_NAME", 0) == 0)
                         {
                             valueBDF = line.substr(line.find("=") + 1, line.length() - line.find("=") + 1);
                             //cout << valueBDF << '\n';
                         }

                         //Find "PCI_ID" and get PCI vendorID, deviceID.
                         if (line.find("PCI_ID", 0) == 0)
                         {
                             vendorID = line.substr(line.find("=") + 1, 4);
                             deviceID = line.substr(line.find(":") + 1, 4);

                             //Transform ID to lowcase
                             transform(vendorID.begin(), vendorID.end(), vendorID.begin(), ::tolower);
                             transform(deviceID.begin(), deviceID.end(), deviceID.begin(), ::tolower);

                             //Open pci.ids for search
                             //PCIIDFile.open("/usr/share/misc/pci.ids", ios::in);
                             PCIIDFile.open("/usr/share/pci.ids", ios::in);

                             if (PCIIDFile.is_open())
                             {
                                 while ( getline(PCIIDFile, linePCI) )
                                 {
                                     //Find vender ID first
                                     if (linePCI.find(vendorID, 0) == 0 && linePCI.find("\t", 0) != 0 )
                                     {
                                         controllerName = linePCI.substr(linePCI.find(vendorID) + 4, linePCI.length() - linePCI.find(vendorID) + 4);
                                         find_vendorID = 1;
                                         //cout << "Find: " << find_vendorID << linePCI << endl;
                                     }

                                     //Then find device ID
                                     if (linePCI.find("\t", 0) == 0 && linePCI.find("\t\t", 0) != 0  && find_vendorID == 1 && linePCI.find(deviceID) == 1)
                                     {
                                         controllerName += linePCI.substr(linePCI.find(deviceID) + 4, linePCI.length() - linePCI.find(deviceID) + 4);
                                         //cout << "Find: " << linePCI.find(deviceID.c_str()) << endl;
                                     }
                                 }
                                 find_vendorID = 0;
                                 PCIIDFile.close();
                             }
                         }
                     }
                 }

                 //Get Ethernet IP address
                 sockfd = socket(AF_INET, SOCK_DGRAM, 0);
                 memset(&ifr, 0, sizeof(ifr));
                 //To get an IPv4 IP address
                 ifr.ifr_addr.sa_family = AF_INET;
                 strncpy(ifr.ifr_name, fileName.c_str(), IFNAMSIZ-1);
                 ioctl(sockfd, SIOCGIFADDR, &ifr);
                 ethData.ip = inet_ntoa(((struct sockaddr_in *)&ifr.ifr_addr)->sin_addr);

                 //Get Ethernet driver name
                 drvinf.cmd = ETHTOOL_GDRVINFO;
                 ifr.ifr_data = (char *) &drvinf;
                 ioctl(sockfd, SIOCETHTOOL, &ifr);
                 ethData.driver = drvinf.driver;

                 close(sockfd);

                 //print the final result
                 //cout << setw(15) << setiosflags(ios::left) << fileName << setw(10) << valueBDF << controllerName << endl;
                 ethData.name = fileName;
                 ethData.BDF = valueBDF;
                 ethData.controller = controllerName;

                 ethernet_name_list->push_back(ethData);

                 valueBDF = "";
                 vendorID = "";
                 deviceID = "";
                 controllerName = "";

                 file.close();
             }
        }
        closedir (dir);
    }

    else
    {
         /* could not open directory */
         return returnErrorMsg( (ERRORCODE)(ERROR_ETHERNET_DDAPI_GET_NAME_LIST | ERROR_OPEN_FILE), strerror(errno), NULL);
    }

    return NO_ERROR_MSG;
}



CHAR * EthernetDDAPI::LTrim(CHAR *string,
                            CHAR junk)
{
    char *original = string;
    char *p = original;
    int  trimmed = 0;

    do
    {
        if (*original != junk || trimmed)
        {
            trimmed = 1;
            *p++ = *original;
        }
    }
    while (*original++ != '\0');

    return string;
}



ErrorMsg* EthernetDDAPI::GetEthernetIP(CHAR  *ethernet_name,
                                       UCHAR *ip)
{
    struct ifreq ifr;
    size_t if_name_len = strlen(ethernet_name);
    int fd;

    errno = 0;

    if (if_name_len == 0)
    {
        return returnErrorMsg( (ERRORCODE)(ERROR_ETHERNET_DDAPI_GET_IP | ERROR_ETHERNET_DDAPI_NO_SPECIFY_NIC), NULL, NULL);
    }

    if (if_name_len < sizeof(ifr.ifr_name))
    {
        memcpy(ifr.ifr_name,ethernet_name,if_name_len);
        ifr.ifr_name[if_name_len] = 0;
    } 
    else 
    {
         return returnErrorMsg( (ERRORCODE)(ERROR_ETHERNET_DDAPI_GET_IP | ERROR_ETHERNET_DDAPI_NIC_NAME_TOO_LONG), NULL, NULL);
    }


    fd = socket(AF_INET,SOCK_DGRAM,0);
    if (fd < 0)
    {
        close(fd);
        return returnErrorMsg( (ERRORCODE)(ERROR_ETHERNET_DDAPI_GET_IP | ERROR_ETHERNET_DDAPI_SOCKET_OPEN_FAIL), NULL, NULL);
    }


    if (ioctl(fd,SIOCGIFADDR,&ifr)==-1 && errno != EADDRNOTAVAIL)
    {
        close(fd);
        return returnErrorMsg( (ERRORCODE)(ERROR_ETHERNET_DDAPI_GET_IP | ERROR_ETHERNET_DDAPI_IOCTL_FAIL), strerror(errno), NULL);
    }

    struct sockaddr_in* ipaddr = (struct sockaddr_in*)&ifr.ifr_addr;
    if (errno == EADDRNOTAVAIL)
    {
       ip[0] = 0;
       ip[1] = 0;
       ip[2] = 0;
       ip[3] = 0;
    }
    else
    {
       ip[0] = (ipaddr->sin_addr.s_addr & 0xff000000) >> 24;
       ip[1] = (ipaddr->sin_addr.s_addr & 0x00ff0000) >> 16;
       ip[2] = (ipaddr->sin_addr.s_addr & 0x0000ff00) >> 8;
       ip[3] =  ipaddr->sin_addr.s_addr & 0x000000ff;
    }

    //printf("IP : %s\n",inet_ntoa(ipaddr->sin_addr));
    close(fd);

    return NO_ERROR_MSG;
}



ErrorMsg* EthernetDDAPI::GetEthernetMask(CHAR  *ethernet_name,
                                         UCHAR *mask)
{
    struct ifreq ifr;
    size_t if_name_len = strlen(ethernet_name);
    int fd;

    errno = 0;

    if (if_name_len == 0)
    {
        return returnErrorMsg( (ERRORCODE)(ERROR_ETHERNET_DDAPI_GET_MASK | ERROR_ETHERNET_DDAPI_NO_SPECIFY_NIC), NULL, NULL);
    }


    if (if_name_len<sizeof(ifr.ifr_name))
    {
        memcpy(ifr.ifr_name,ethernet_name,if_name_len);
        ifr.ifr_name[if_name_len]=0;
    }
    else
    {
        return returnErrorMsg( (ERRORCODE)(ERROR_ETHERNET_DDAPI_GET_MASK | ERROR_ETHERNET_DDAPI_NIC_NAME_TOO_LONG), NULL, NULL);
    }


    fd = socket(AF_INET,SOCK_DGRAM,0);
    if (fd < 0)
    {
        close(fd);
        return returnErrorMsg( (ERRORCODE)(ERROR_ETHERNET_DDAPI_GET_MASK | ERROR_ETHERNET_DDAPI_SOCKET_OPEN_FAIL), NULL, NULL);
    }


    if (ioctl(fd,SIOCGIFNETMASK,&ifr) == -1 && errno != EADDRNOTAVAIL)
    {
        close(fd);
        return returnErrorMsg( (ERRORCODE)(ERROR_ETHERNET_DDAPI_GET_MASK | ERROR_ETHERNET_DDAPI_IOCTL_FAIL), strerror(errno), NULL);
    }

    struct sockaddr_in* ipaddr = (struct sockaddr_in*)&ifr.ifr_addr;
    if (errno == EADDRNOTAVAIL)
    {
       mask[0] = 0;
       mask[1] = 0;
       mask[2] = 0;
       mask[3] = 0;
    }
    else
    {
       mask[0] = (ipaddr->sin_addr.s_addr & 0xff000000) >> 24;
       mask[1] = (ipaddr->sin_addr.s_addr & 0x00ff0000) >> 16;
       mask[2] = (ipaddr->sin_addr.s_addr & 0x0000ff00) >> 8;
       mask[3] =  ipaddr->sin_addr.s_addr & 0x000000ff;
    }

    //printf("IP : %s\n",inet_ntoa(ipaddr->sin_addr));
    close(fd);

    return NO_ERROR_MSG;
}



ErrorMsg* EthernetDDAPI::GetEthernetMTU(CHAR  *ethernet_name,
                                        INT32 *mtu)
{
    struct ifreq ifr;
    size_t if_name_len = strlen(ethernet_name);
    int fd;

    errno = 0;

    if (if_name_len == 0)
    {
        return returnErrorMsg( (ERRORCODE)(ERROR_ETHERNET_DDAPI_GET_MTU | ERROR_ETHERNET_DDAPI_NO_SPECIFY_NIC), NULL, NULL);
    }

    if (if_name_len<sizeof(ifr.ifr_name))
    {
        memcpy(ifr.ifr_name,ethernet_name,if_name_len);
        ifr.ifr_name[if_name_len] = 0;
    }
    else
    {
        return returnErrorMsg( (ERRORCODE)(ERROR_ETHERNET_DDAPI_GET_MTU | ERROR_ETHERNET_DDAPI_NIC_NAME_TOO_LONG), NULL, NULL);
    }


    fd = socket(AF_INET,SOCK_DGRAM,0);
    if (fd < 0)
    {
        close(fd);
        return returnErrorMsg( (ERRORCODE)(ERROR_ETHERNET_DDAPI_GET_MTU | ERROR_ETHERNET_DDAPI_SOCKET_OPEN_FAIL), NULL, NULL);
    }


    if (ioctl(fd,SIOCGIFMTU,&ifr)==-1)
    {
        close(fd);
        return returnErrorMsg( (ERRORCODE)(ERROR_ETHERNET_DDAPI_GET_MTU | ERROR_ETHERNET_DDAPI_IOCTL_FAIL), strerror(errno), NULL);
    }

    *mtu = ifr.ifr_mtu;
    close(fd);

    return NO_ERROR_MSG;
}



ErrorMsg* EthernetDDAPI::GetEthernetEEPROM(CHAR  *ethernet_name,
                                           UCHAR *data,
                                           INT32 data_counter)
{
    INT32 err, fd, i;
    struct ethtool_eeprom *eeprom;
    struct ifreq ifr;
    size_t if_name_len = strlen(ethernet_name);

    errno = 0;

    if (if_name_len == 0)
    {
        return returnErrorMsg( (ERRORCODE)(ERROR_ETHERNET_DDAPI_GET_EEPROM | ERROR_ETHERNET_DDAPI_NO_SPECIFY_NIC), NULL, NULL);
    }

    fd = socket(AF_INET,SOCK_DGRAM,0);
    if (fd < 0)
    {
        return returnErrorMsg( (ERRORCODE)(ERROR_ETHERNET_DDAPI_GET_EEPROM | ERROR_ETHERNET_DDAPI_SOCKET_OPEN_FAIL), NULL, NULL);
    }

    strncpy(ifr.ifr_name, ethernet_name, sizeof(ifr.ifr_name));

    eeprom = (ethtool_eeprom *)calloc(1, sizeof(*eeprom) + data_counter);
    if (!eeprom)
    {
        free(eeprom);
        close(fd);
        return returnErrorMsg( (ERRORCODE)(ERROR_ETHERNET_DDAPI_GET_EEPROM | ERROR_ETHERNET_DDAPI_ALLOCAT_EEPROM_FAIL), NULL, NULL);
    }

    eeprom->cmd = ETHTOOL_GEEPROM;
    eeprom->len = data_counter;
    eeprom->offset = 0;

    ifr.ifr_data = (__caddr_t)eeprom;

    err = ioctl(fd, SIOCETHTOOL, &ifr);
    if (err < 0)
    {
        //perror("Cannot do get EEPROM");
        free(eeprom);
        close(fd);
        return returnErrorMsg( (ERRORCODE)(ERROR_ETHERNET_DDAPI_GET_EEPROM | ERROR_ETHERNET_DDAPI_IOCTL_FAIL), strerror(errno), NULL);
    }

    for (i = 0; i < eeprom->len; i++)
    {
        data[i] = eeprom->data[i];
    }

    free(eeprom);
    close(fd);

    return NO_ERROR_MSG;
}



ErrorMsg* EthernetDDAPI::GetEthernetEEPROM(string ethernet_name,
                                           INT32  StartAddress,
                                           INT32  Length,
                                           INT32  *data)
{
    string vendor_ID;
    string device_ID;
    EthernetBaseClass *nic_interface = NULL;

    m_ethernet_name = ethernet_name;

    //Get ethernet PCI configuration vendor ID and device ID
    m_errorCode = get_vendor_ID(&vendor_ID);
    if (m_errorCode != NO_ERROR_MSG)
    {
       return returnErrorMsg(ERROR_ETHERNET_DDAPI_SENDIOCTL, "Get vender ID fail", m_errorCode);
    }

    m_errorCode = get_device_ID(&device_ID);
    if (m_errorCode != NO_ERROR_MSG)
    {
       return returnErrorMsg(ERROR_ETHERNET_DDAPI_SENDIOCTL, "Get device ID fail", m_errorCode);
    }
    //cout << "vendor_ID: " <<vendor_ID << "  device_ID:" << device_ID << endl;

    //Select which type of ethernet to create object
    if (vendor_ID.compare("0x8086") == 0 && (device_ID.compare("0x1533") == 0 || device_ID.compare("0x1537") == 0)) //INTEL I210
    {
        nic_interface = new INTEL_I210(m_ethernet_name.c_str());
    }
    if (vendor_ID.compare("0x8086") == 0 && device_ID.compare("0x15ab") == 0) //INTEL X552(KR)
    {
        nic_interface = new INTEL_X552(m_ethernet_name.c_str());
    }
    if (vendor_ID.compare("0x8086") == 0 && device_ID.compare("0x1583") == 0) //INTEL XL710
    {
        nic_interface = new INTEL_XL710(m_ethernet_name.c_str());
    }
    if (vendor_ID.compare("0x8086") == 0 && device_ID.compare("0x1572") == 0) //INTEL XL710
    {
        nic_interface = new INTEL_XL710(m_ethernet_name.c_str());
    }
    if (vendor_ID.compare("0x1425") == 0 && device_ID.compare("0x6407") == 0) //Chelsio T62100
    {
       nic_interface = new CHELSIS_T62100(m_ethernet_name.c_str());
    }
    if (vendor_ID.compare("0x1425") == 0 && device_ID.compare("0x640d") == 0) //Chelsio T62100-CR
    {
       nic_interface = new CHELSIS_T62100(m_ethernet_name.c_str());
    }
    if (vendor_ID.compare("0x1425") == 0 && device_ID.compare("0x6401") == 0) //Chelsio T6225-CR
    {
       nic_interface = new CHELSIS_T62100(m_ethernet_name.c_str());
    }

    if (nic_interface != NULL)
    {
        //Start get EEPROM data
        m_errorCode = nic_interface->GetEEPROM(ethernet_name, StartAddress, Length, data);

        if (m_errorCode != NO_ERROR_MSG)
        {
           return returnErrorMsg(ERROR_ETHERNET_DDAPI_SENDIOCTL, "Get EEPROM fail", m_errorCode);
        }
    }
    else
    {
        return returnErrorMsg( (ERRORCODE)(ERROR_ETHERNET_DDAPI_SENDIOCTL | ERROR_ETHERNET_DDAPI_IOCTL_FAIL), "No define ethernet", NULL);
    }

    delete nic_interface;

    return NO_ERROR_MSG;
}



ErrorMsg* EthernetDDAPI::GetEthernetEEPROMSize(CHAR  *ethernet_name,
                                               INT32 *data_counter)
{
    INT32 err,fd;
    struct ethtool_drvinfo drvinfo;
    struct ifreq ifr;
    size_t if_name_len = strlen(ethernet_name);

    errno = 0;

    if (if_name_len == 0)
    {
        return returnErrorMsg( (ERRORCODE)(ERROR_ETHERNET_DDAPI_GET_EEPROM_SIZE | ERROR_ETHERNET_DDAPI_NO_SPECIFY_NIC), NULL, NULL);
    }

    fd = socket(AF_INET,SOCK_DGRAM,0);
    if (fd < 0)
    {
        return returnErrorMsg( (ERRORCODE)(ERROR_ETHERNET_DDAPI_GET_EEPROM_SIZE | ERROR_ETHERNET_DDAPI_SOCKET_OPEN_FAIL), NULL, NULL);
    }

    strncpy(ifr.ifr_name, ethernet_name, sizeof(ifr.ifr_name));


    drvinfo.cmd = ETHTOOL_GDRVINFO;
    ifr.ifr_data = (__caddr_t)&drvinfo;

    err = ioctl(fd, SIOCETHTOOL, &ifr);
    if (err < 0)
    {
        return returnErrorMsg( (ERRORCODE)(ERROR_ETHERNET_DDAPI_GET_EEPROM_SIZE | ERROR_ETHERNET_DDAPI_IOCTL_FAIL), strerror(errno), NULL);
    }

    //printf("EEPROM len:%d\n",drvinfo.eedump_len);
    *data_counter = drvinfo.eedump_len;
    close(fd);

    return NO_ERROR_MSG;
}



ErrorMsg* EthernetDDAPI::SendIoctl(string   ethernet_name,
                                   TestType index,
                                   INT32    *test_result)
{
    struct
    {
        struct ethtool_sset_info hdr;
        int buf[1];
    } sset_info;

    struct ethtool_drvinfo drvinfo;
    int len, fd, error_return, drvinfo_offset;
    // struct ethtool_gstrings *strings;
    struct ifreq ifr;
    struct ethtool_test *test;
    string vendor_ID;
    string device_ID;
    EthernetBaseClass *nic_interface = NULL;

    errno = 0;

    if (ethernet_name.length() == 0)
    {
        return returnErrorMsg( (ERRORCODE)(ERROR_ETHERNET_DDAPI_SENDIOCTL | ERROR_ETHERNET_DDAPI_NO_SPECIFY_NIC), NULL, NULL);
    }

    drvinfo_offset = offsetof(struct ethtool_drvinfo, testinfo_len);

    fd = socket(AF_INET,SOCK_DGRAM,0);
    if (fd < 0)
    {
        //printf("SendIoctl: Socket fail!!\n");
        //close(fd);
        return returnErrorMsg( (ERRORCODE)(ERROR_ETHERNET_DDAPI_SENDIOCTL | ERROR_ETHERNET_DDAPI_SOCKET_OPEN_FAIL), NULL, NULL);
    }

    //Covert well know name to system real name
    //ConverWellKnowNameToRealName(ethernet_name, &ethernet_name);
    strncpy(ifr.ifr_name, ethernet_name.c_str(), sizeof(ifr.ifr_name));

    sset_info.hdr.cmd = ETHTOOL_GSSET_INFO;
    sset_info.hdr.reserved = 0;
    sset_info.hdr.sset_mask = 1ULL << 0;

    ifr.ifr_data = (__caddr_t)&sset_info;

    error_return = ioctl(fd, SIOCETHTOOL, &ifr);
    if (error_return == 0)
    {
        len = sset_info.hdr.sset_mask ? sset_info.hdr.data[0] : 0;
    }
    else if (errno == EOPNOTSUPP && drvinfo_offset != 0)
    {
        /* Fallback for old kernel versions */
        drvinfo.cmd = ETHTOOL_GDRVINFO;
        ifr.ifr_data = (__caddr_t)&drvinfo;
        if (ioctl(fd, SIOCETHTOOL, &ifr))
        {
            close(fd);
            return returnErrorMsg( (ERRORCODE)(ERROR_ETHERNET_DDAPI_SENDIOCTL | ERROR_ETHERNET_DDAPI_IOCTL_NOT_SUPPORT), strerror(errno), NULL);
        }
        len = *(int *)((char *)&drvinfo + drvinfo_offset);
    }
    else
    {
        //perror("Cannot do ioctl");
        close(fd);
        return returnErrorMsg( (ERRORCODE)(ERROR_ETHERNET_DDAPI_SENDIOCTL | ERROR_ETHERNET_DDAPI_IOCTL_FAIL), strerror(errno), NULL);
    }

    test = (struct ethtool_test *)calloc(1, sizeof(*test) + len * sizeof(unsigned long long));
    memset(test->data, 0, len * sizeof(unsigned long long));

    test->cmd = ETHTOOL_TEST;
    test->len = len;
    test->flags = 1;

    ifr.ifr_data = (__caddr_t)test;

    error_return = ioctl(fd, SIOCETHTOOL, &ifr);
    if (error_return < 0)
    {
        //perror("Cannot do nic test");
        close(fd);
        free (test);
        return returnErrorMsg( (ERRORCODE)(ERROR_ETHERNET_DDAPI_SENDIOCTL | ERROR_ETHERNET_DDAPI_IOCTL_ETHTOOL_FAIL), strerror(errno), NULL);
    }

    m_ethernet_name = ethernet_name;

    //Get ethernet PCI configuration vendor ID and device ID
    m_errorCode = get_vendor_ID(&vendor_ID);
    if (m_errorCode != NO_ERROR_MSG)
    {
       return returnErrorMsg(ERROR_ETHERNET_DDAPI_SENDIOCTL, "Get vender ID fail", m_errorCode);
    }

    m_errorCode = get_device_ID(&device_ID);
    if (m_errorCode != NO_ERROR_MSG)
    {
       return returnErrorMsg(ERROR_ETHERNET_DDAPI_SENDIOCTL, "Get device ID fail", m_errorCode);
    }
    //cout << "vendor_ID: " <<vendor_ID << "  device_ID:" << device_ID << endl;

    //Select which type of ethernet to create object
    if (vendor_ID.compare("0x8086") == 0 && (device_ID.compare("0x1533") == 0 || device_ID.compare("0x1537") == 0)) //INTEL I210
    {
        nic_interface = new INTEL_I210(m_ethernet_name.c_str());
    }
    if (vendor_ID.compare("0x8086") == 0 && device_ID.compare("0x15ab") == 0) //INTEL X552(KR)
    {
        nic_interface = new INTEL_X552(m_ethernet_name.c_str());
    }
    if (vendor_ID.compare("0x8086") == 0 && device_ID.compare("0x1583") == 0) //INTEL XL710
    {
        nic_interface = new INTEL_XL710(m_ethernet_name.c_str());
    }
    if (vendor_ID.compare("0x8086") == 0 && device_ID.compare("0x1572") == 0) //INTEL XL710
    {
        nic_interface = new INTEL_XL710(m_ethernet_name.c_str());
    }


    if (nic_interface != NULL)
    {
        //Start self get test result
        m_errorCode = nic_interface->Get_self_test_result(test, index, test_result);
    }
    else
    {
        close(fd);
        free(test);
        return returnErrorMsg( (ERRORCODE)(ERROR_ETHERNET_DDAPI_SENDIOCTL | ERROR_ETHERNET_DDAPI_IOCTL_FAIL), "No define ethernet", NULL);
    }

    delete nic_interface;

    //printf("Len:%d\n",len);
    //for(i=0; i<len ;i++)
    //   printf("%d: %d\n",i, test->data[i]);
    close(fd);
    free(test);

    return m_errorCode;
}


ErrorMsg* EthernetDDAPI::SetEthernetEEPROM(CHAR  *ethernet_name,
                                           INT32 offset,
                                           UCHAR data,
                                           INT32 Length)
{
    INT32  fd, err;
    struct ethtool_eeprom *eeprom;
    struct ifreq ifr;
    INT32  magic;
    size_t if_name_len = strlen(ethernet_name);

    errno = 0;

    if (if_name_len == 0)
    {
        return returnErrorMsg( (ERRORCODE)(ERROR_ETHERNET_DDAPI_SET_EEPROM | ERROR_ETHERNET_DDAPI_NO_SPECIFY_NIC), NULL, NULL);
    }


    fd = socket(AF_INET,SOCK_DGRAM,0);
    if (fd < 0)
    {
        return returnErrorMsg( (ERRORCODE)(ERROR_ETHERNET_DDAPI_SET_EEPROM | ERROR_ETHERNET_DDAPI_SOCKET_OPEN_FAIL), NULL, NULL);
    }

    strncpy(ifr.ifr_name, ethernet_name, sizeof(ifr.ifr_name));

    eeprom = (ethtool_eeprom *)calloc(1, sizeof(*eeprom) + 1);
    if (!eeprom)
    {
        close(fd);
        return returnErrorMsg( (ERRORCODE)(ERROR_ETHERNET_DDAPI_SET_EEPROM | ERROR_ETHERNET_DDAPI_ALLOCAT_EEPROM_FAIL), NULL, NULL);
    }

    m_errorCode = GetEthernetMagic(ethernet_name, &magic);

    if (m_errorCode != NO_ERROR_MSG)
    {
        return returnErrorMsg( (ERRORCODE)(ERROR_ETHERNET_DDAPI_SET_EEPROM), NULL, m_errorCode);
    }
    //printf("Magic: 0x%x\n",magic);

    eeprom->cmd = ETHTOOL_SEEPROM;
    eeprom->len = Length;
    eeprom->offset = offset;
    eeprom->magic = magic; //Intel device_id:vend_id
    eeprom->data[0] = data;

    ifr.ifr_data = (__caddr_t)eeprom;

    err = ioctl(fd, SIOCETHTOOL, &ifr);
    if (err < 0)
    {
        close(fd);
        free(eeprom);
        return returnErrorMsg( (ERRORCODE)(ERROR_ETHERNET_DDAPI_SET_EEPROM | ERROR_ETHERNET_DDAPI_IOCTL_FAIL), strerror(errno), NULL);
    }

    close(fd);
    free(eeprom);

    return NO_ERROR_MSG;
}



ErrorMsg* EthernetDDAPI::SetEthernetEEPROM(string ethernet_name,
                                           INT32  StartAddress,
                                           INT32  *data,
                                           INT32  Length)
{
    string vendor_ID;
    string device_ID;
    EthernetBaseClass *nic_interface = NULL;

    m_ethernet_name = ethernet_name;

    //Get ethernet PCI configuration vendor ID and device ID
    m_errorCode = get_vendor_ID(&vendor_ID);
    if (m_errorCode != NO_ERROR_MSG)
    {
       return returnErrorMsg(ERROR_ETHERNET_DDAPI_SENDIOCTL, "Get vender ID fail", m_errorCode);
    }

    m_errorCode = get_device_ID(&device_ID);
    if (m_errorCode != NO_ERROR_MSG)
    {
       return returnErrorMsg(ERROR_ETHERNET_DDAPI_SENDIOCTL, "Get device ID fail", m_errorCode);
    }
    //cout << "vendor_ID: " <<vendor_ID << "  device_ID:" << device_ID << endl;

    //Select which type of ethernet to create object
    if (vendor_ID.compare("0x8086") == 0 && (device_ID.compare("0x1533") == 0 || device_ID.compare("0x1537") == 0)) //INTEL I210
    {
        nic_interface = new INTEL_I210(m_ethernet_name.c_str());
    }
    if (vendor_ID.compare("0x8086") == 0 && device_ID.compare("0x15ab") == 0) //INTEL X552(KR)
    {
        nic_interface = new INTEL_X552(m_ethernet_name.c_str());
    }
    if (vendor_ID.compare("0x8086") == 0 && device_ID.compare("0x1583") == 0) //INTEL XL710
    {
        nic_interface = new INTEL_XL710(m_ethernet_name.c_str());
    }
    if (vendor_ID.compare("0x8086") == 0 && device_ID.compare("0x1572") == 0) //INTEL XL710
    {
        nic_interface = new INTEL_XL710(m_ethernet_name.c_str());
    }
    if (vendor_ID.compare("0x1425") == 0 && device_ID.compare("0x6407") == 0) //Chelsio T62100
    {
       //nic_interface = new CHELSIS_T62100(m_ethernet_name.c_str());
       SetEthernetEEPROM((CHAR*)ethernet_name.c_str(), StartAddress, *data, Length);
    }
    if (vendor_ID.compare("0x1425") == 0 && device_ID.compare("0x640d") == 0) //Chelsio T62100-CR
    {
       //nic_interface = new CHELSIS_T62100(m_ethernet_name.c_str());
       SetEthernetEEPROM((CHAR*)ethernet_name.c_str(), StartAddress, *data, Length);
    }
    if (vendor_ID.compare("0x1425") == 0 && device_ID.compare("0x6401") == 0) //Chelsio T6225-CR
    {
       //nic_interface = new CHELSIS_T62100(m_ethernet_name.c_str());
       SetEthernetEEPROM((CHAR*)ethernet_name.c_str(), StartAddress, *data, Length);
    }

    if (nic_interface != NULL)
    {
        //Start get EEPROM data
        m_errorCode = nic_interface->SetEEPROM(ethernet_name, StartAddress, data, Length);

        if (m_errorCode != NO_ERROR_MSG)
        {
           return returnErrorMsg(ERROR_ETHERNET_DDAPI_SENDIOCTL, "Set EEPROM fail", m_errorCode);
        }
    }
    else
    {
        return returnErrorMsg( (ERRORCODE)(ERROR_ETHERNET_DDAPI_SENDIOCTL | ERROR_ETHERNET_DDAPI_IOCTL_FAIL), "No define ethernet", NULL);
    }

    delete nic_interface;

    return NO_ERROR_MSG;
}



ErrorMsg* EthernetDDAPI::GetEthernetMagic(CHAR  *ethernet_name,
                                          INT32 *magic)
{
    INT32  err,fd,device_id,vendor_id;
    struct ethtool_drvinfo drvinfo;
    struct ifreq ifr;
    CHAR   pci_path[36];
    CHAR   tmp[10];
    FILE   *pFile;
    size_t if_name_len = strlen(ethernet_name);

    errno = 0;

    if (if_name_len == 0)
    {
        return returnErrorMsg( (ERRORCODE)(ERROR_ETHERNET_DDAPI_GET_MAIC | ERROR_ETHERNET_DDAPI_NO_SPECIFY_NIC), NULL, NULL);
    }

    fd = socket(AF_INET,SOCK_DGRAM,0);
    if (fd < 0)
    {
        return returnErrorMsg( (ERRORCODE)(ERROR_ETHERNET_DDAPI_GET_MAIC | ERROR_ETHERNET_DDAPI_SOCKET_OPEN_FAIL), NULL, NULL);
    }

    strncpy(ifr.ifr_name, ethernet_name, sizeof(ifr.ifr_name));

    drvinfo.cmd = ETHTOOL_GDRVINFO;
    ifr.ifr_data = (__caddr_t)&drvinfo;

    err = ioctl(fd, SIOCETHTOOL, &ifr);
    if (err < 0)
    {
        //perror("Cannot get driver information");
        close(fd);
        return returnErrorMsg( (ERRORCODE)(ERROR_ETHERNET_DDAPI_GET_MAIC | ERROR_ETHERNET_DDAPI_IOCTL_FAIL), strerror(errno), NULL);
    }

    //Clean path string and get device ID.
    memset(pci_path, '\0', sizeof(pci_path));
    strcat(pci_path, "/sys/bus/pci/devices/");
    strcat(pci_path, drvinfo.bus_info);
    strcat(pci_path, "/device");

    //Prepare to get the nic device id
    pFile = fopen (pci_path,"r");
    if (pFile == NULL)
    {
         //printf("GetEthernetMagic device id: Error opening file\n");
         close(fd);
         return returnErrorMsg( (ERRORCODE)(ERROR_ETHERNET_DDAPI_GET_MAIC | ERROR_OPEN_FILE), strerror(errno), NULL);
    }

    fgets(tmp, 50, pFile);
    sscanf(tmp, "%x",&device_id);
    fclose(pFile);

    //Clean path string and get vendor ID.
    memset(pci_path, '\0', sizeof(pci_path));
    strcat(pci_path, "/sys/bus/pci/devices/");
    strcat(pci_path, drvinfo.bus_info);
    strcat(pci_path, "/vendor");

    //Prepare to get the nic vendor id
    pFile=fopen (pci_path,"r");
    if (pFile==NULL)
    {
         //printf("GetEthernetMagic vendor id: Error opening file\n");
         close(fd);
         return returnErrorMsg( (ERRORCODE)(ERROR_ETHERNET_DDAPI_GET_MAIC | ERROR_OPEN_FILE), strerror(errno), NULL);
    }

    fgets(tmp, 50, pFile);
    sscanf(tmp, "%x",&vendor_id);

    *magic = (device_id << 16) | vendor_id;

    fclose(pFile);
    close(fd);

    return NO_ERROR_MSG;
}



ErrorMsg* EthernetDDAPI::SetEthernetMAC(string  ethernet_name,
                                        string  *mac_address)
{
    string vendor_ID;
    string device_ID;
    EthernetBaseClass *nic_interface = NULL;

    m_ethernet_name = ethernet_name;

    //Get ethernet PCI configuration vendor ID and device ID
    m_errorCode = get_vendor_ID(&vendor_ID);
    if (m_errorCode != NO_ERROR_MSG)
    {
       return returnErrorMsg(ERROR_ETHERNET_DDAPI_SENDIOCTL, "Get vender ID fail", m_errorCode);
    }

    m_errorCode = get_device_ID(&device_ID);
    if (m_errorCode != NO_ERROR_MSG)
    {
       return returnErrorMsg(ERROR_ETHERNET_DDAPI_SENDIOCTL, "Get device ID fail", m_errorCode);
    }
    //cout << "vendor_ID: " <<vendor_ID << "  device_ID:" << device_ID << endl;

    //Select which type of ethernet to create object
    if (vendor_ID.compare("0x8086") == 0 && (device_ID.compare("0x1533") == 0 || device_ID.compare("0x1537") == 0)) //INTEL I210
    {
        nic_interface = new INTEL_I210(m_ethernet_name.c_str());
    }
    if (vendor_ID.compare("0x8086") == 0 && device_ID.compare("0x15ab") == 0) //INTEL X552(KR)
    {
        nic_interface = new INTEL_X552(m_ethernet_name.c_str());
    }
    if (vendor_ID.compare("0x8086") == 0 && device_ID.compare("0x1583") == 0) //INTEL XL710
    {
        nic_interface = new INTEL_XL710(m_ethernet_name.c_str());
    }
    if (vendor_ID.compare("0x8086") == 0 && device_ID.compare("0x1572") == 0) //INTEL XL710
    {
        nic_interface = new INTEL_XL710(m_ethernet_name.c_str());
    }
    if (vendor_ID.compare("0x1425") == 0 && device_ID.compare("0x6407") == 0) //Chelsio T62100
    {
       //nic_interface = new CHELSIS_T62100(m_ethernet_name.c_str());
       return returnErrorMsg((ERRORCODE)(ERROR_ETHERNET_DDAPI_ENABLE_EXTERNAL_LOOPBACK | ERROR_FUNCTION_NOT_SUPPORTED));
    }
    if (vendor_ID.compare("0x1425") == 0 && device_ID.compare("0x640d") == 0) //Chelsio T62100-CR
    {
        //nic_interface = new CHELSIS_T62100(m_ethernet_name.c_str());
        return returnErrorMsg((ERRORCODE)(ERROR_ETHERNET_DDAPI_ENABLE_EXTERNAL_LOOPBACK | ERROR_FUNCTION_NOT_SUPPORTED));
    }
    if (vendor_ID.compare("0x1425") == 0 && device_ID.compare("0x6401") == 0) //Chelsio T6225-CR
    {
        //nic_interface = new CHELSIS_T62100(m_ethernet_name.c_str());
        return returnErrorMsg((ERRORCODE)(ERROR_ETHERNET_DDAPI_ENABLE_EXTERNAL_LOOPBACK | ERROR_FUNCTION_NOT_SUPPORTED));
    }

    if (nic_interface != NULL)
    {
        //Start setup speed
        m_errorCode = nic_interface->SetMAC(ethernet_name, mac_address);

        if (m_errorCode != NO_ERROR_MSG)
        {
           return returnErrorMsg(ERROR_ETHERNET_DDAPI_SET_MAC, NULL, m_errorCode);
        }
    }
    else
    {
        return returnErrorMsg( (ERRORCODE)(ERROR_ETHERNET_DDAPI_SET_MAC), "No define ethernet", NULL);
    }

    delete nic_interface;

    return NO_ERROR_MSG;
}



ErrorMsg* EthernetDDAPI::Enable_external_loopback(string ethernet_name)
{
    string vendor_ID;
    string device_ID;
    EthernetBaseClass *nic_interface;
    char              command_string[50];

    m_errorCode = get_vendor_ID(&vendor_ID);
    if (m_errorCode != NO_ERROR_MSG)
    {
       return returnErrorMsg(ERROR_ETHERNET_DDAPI_ENABLE_EXTERNAL_LOOPBACK, NULL, m_errorCode);
    }

    m_errorCode = get_device_ID(&device_ID);
    if (m_errorCode != NO_ERROR_MSG)
    {
       return returnErrorMsg(ERROR_ETHERNET_DDAPI_ENABLE_EXTERNAL_LOOPBACK, NULL, m_errorCode);
    }
    //cout << vendor_ID << device_ID << endl;

    if (vendor_ID.compare("0x8086") == 0 && (device_ID.compare("0x1533") == 0 || device_ID.compare("0x1537") == 0)) //INTEL I210
       nic_interface = new INTEL_I210(m_ethernet_name.c_str());
    else if (vendor_ID.compare("0x1425") == 0 && device_ID.compare("0x6407") == 0) //Chelsio T62100
       //nic_interface = new CHELSIS_T62100(m_ethernet_name.c_str());
       return returnErrorMsg((ERRORCODE)(ERROR_ETHERNET_DDAPI_ENABLE_EXTERNAL_LOOPBACK | ERROR_FUNCTION_NOT_SUPPORTED));
    else if (vendor_ID.compare("0x1425") == 0 && device_ID.compare("0x640d") == 0) //T62100-CR
       //nic_interface = new CHELSIS_T62100(m_ethernet_name.c_str());
       return returnErrorMsg((ERRORCODE)(ERROR_ETHERNET_DDAPI_ENABLE_EXTERNAL_LOOPBACK | ERROR_FUNCTION_NOT_SUPPORTED));
    else if (vendor_ID.compare("0x1425") == 0 && device_ID.compare("0x6401") == 0) //T6225-CR
       //nic_interface = new CHELSIS_T62100(m_ethernet_name.c_str());
       return returnErrorMsg((ERRORCODE)(ERROR_ETHERNET_DDAPI_ENABLE_EXTERNAL_LOOPBACK | ERROR_FUNCTION_NOT_SUPPORTED));
    else
       //nic_interface = new INTEL_I210(m_ethernet_name.c_str());
       nic_interface = new EthernetBaseClass(m_ethernet_name.c_str());

    if (nic_interface != NULL)
    {
        //Turn on the ethernet
        sprintf(command_string, "ifconfig %s up\n", ethernet_name.c_str());
        system(command_string);

        nic_interface->Set_external_loopback();
    }

    delete nic_interface;

    return NO_ERROR_MSG;
}



ErrorMsg* EthernetDDAPI::Disable_external_loopback()
{
    string vendor_ID;
    string device_ID;
    EthernetBaseClass *nic_interface;

    m_errorCode = get_vendor_ID(&vendor_ID);
    if (m_errorCode != NO_ERROR_MSG)
    {
       return returnErrorMsg(ERROR_ETHERNET_DDAPI_DISABLE_EXTERNAL_LOOPBACK, NULL, m_errorCode);
    }

    m_errorCode = get_device_ID(&device_ID);
    if (m_errorCode != NO_ERROR_MSG)
    {
       return returnErrorMsg(ERROR_ETHERNET_DDAPI_DISABLE_EXTERNAL_LOOPBACK, NULL, m_errorCode);
    }
    //cout << vendor_ID << device_ID << endl;

    if (vendor_ID.compare("0x8086") == 0 && (device_ID.compare("0x1533") == 0 || device_ID.compare("0x1537") == 0)) //INTEL I210
       nic_interface = new INTEL_I210(m_ethernet_name.c_str());
    else
       //nic_interface = new INTEL_I210(m_ethernet_name.c_str());
       nic_interface = new EthernetBaseClass(m_ethernet_name.c_str());

    if (nic_interface != NULL)
    {
       nic_interface->Remove_external_loopback();
    }

    delete nic_interface;

    return NO_ERROR_MSG;
}




ErrorMsg* EthernetDDAPI::get_device_ID(string *vendor_ID)
{
    string node_path;
    ifstream infile;

    node_path = "/sys/class/net/" + m_ethernet_name + "/device/device";

    infile.open(node_path.c_str());
    if (infile.is_open())
    {
       getline(infile,*vendor_ID); // Saves the line in STRING.
       infile.close();
    }
    else
    {
       return returnErrorMsg((ERRORCODE)(ERROR_ETHERNET_DDAPI_GET_DEVICE_ID | ERROR_ETHERNET_DDAPI_OPEN_NODE_FAIL), node_path.c_str(), NULL);
    }

    return NO_ERROR_MSG;
}



ErrorMsg* EthernetDDAPI::get_vendor_ID(string *device_ID)
{
    string node_path;
    ifstream infile;

    node_path = "/sys/class/net/" + m_ethernet_name + "/device/vendor";

    infile.open(node_path.c_str());
    if (infile.is_open())
    {
       getline(infile,*device_ID); // Saves the line in STRING.
       infile.close();
    }
    else
    {
       return returnErrorMsg((ERRORCODE)(ERROR_ETHERNET_DDAPI_GET_DEVICE_ID | ERROR_ETHERNET_DDAPI_OPEN_NODE_FAIL), node_path.c_str(), NULL);
    }

    return NO_ERROR_MSG;
}



ErrorMsg* EthernetDDAPI::SetupSpeed(string ethernet_name, int speed)
{
    string vendor_ID;
    string device_ID;
    EthernetBaseClass *nic_interface = NULL;
    char              command_string[50];

    m_ethernet_name = ethernet_name;

    //Get ethernet PCI configuration vendor ID and device ID
    m_errorCode = get_vendor_ID(&vendor_ID);
    if (m_errorCode != NO_ERROR_MSG)
    {
       return returnErrorMsg(ERROR_ETHERNET_DDAPI_SENDIOCTL, "Get vender ID fail", m_errorCode);
    }

    m_errorCode = get_device_ID(&device_ID);
    if (m_errorCode != NO_ERROR_MSG)
    {
       return returnErrorMsg(ERROR_ETHERNET_DDAPI_SENDIOCTL, "Get device ID fail", m_errorCode);
    }
    //cout << "vendor_ID: " <<vendor_ID << "  device_ID:" << device_ID << endl;

    //Select which type of ethernet to create object
    if (vendor_ID.compare("0x8086") == 0 && (device_ID.compare("0x1533") == 0 || device_ID.compare("0x1537") == 0)) //INTEL I210
    {
        nic_interface = new INTEL_I210(m_ethernet_name.c_str());
    }
    if (vendor_ID.compare("0x8086") == 0 && device_ID.compare("0x15ab") == 0) //INTEL X552(KR)
    {
        nic_interface = new INTEL_X552(m_ethernet_name.c_str());
    }
    if (vendor_ID.compare("0x8086") == 0 && device_ID.compare("0x1583") == 0) //INTEL XL710
    {
        nic_interface = new INTEL_XL710(m_ethernet_name.c_str());
    }
    if (vendor_ID.compare("0x8086") == 0 && device_ID.compare("0x1572") == 0) //INTEL XL710
    {
        nic_interface = new INTEL_XL710(m_ethernet_name.c_str());
    }
    if (vendor_ID.compare("0x1425") == 0 && device_ID.compare("0x6407") == 0) //Chelsio T62100
    {
        nic_interface = new CHELSIS_T62100(m_ethernet_name.c_str());
    }
    if (vendor_ID.compare("0x1425") == 0 && device_ID.compare("0x640d") == 0) //Chelsio T62100-CR
    {
        nic_interface = new CHELSIS_T62100(m_ethernet_name.c_str());
    }
    if (vendor_ID.compare("0x1425") == 0 && device_ID.compare("0x6401") == 0) //Chelsio T6225-CR
    {
        nic_interface = new CHELSIS_T62100(m_ethernet_name.c_str());
    }

    if (nic_interface != NULL)
    {
        //Turn on the ethernet
        sprintf(command_string, "ifconfig %s up\n", ethernet_name.c_str());
        system(command_string);

        //Check ethernet driver support speed or not
        m_errorCode = CheckAdvertisedSpeed(ethernet_name, speed);
        if (m_errorCode != NO_ERROR_MSG)
        {
           return returnErrorMsg(ERROR_ETHERNET_DDAPI_SENDIOCTL, NULL, m_errorCode);
        }

        //Start setup speed
        m_errorCode = nic_interface->SetupSpeed(ethernet_name, speed);
        if (m_errorCode != NO_ERROR_MSG)
        {
           return returnErrorMsg(ERROR_ETHERNET_DDAPI_SENDIOCTL, "Setup speed fail", m_errorCode);
        }
    }
    else
    {
        return returnErrorMsg( (ERRORCODE)(ERROR_ETHERNET_DDAPI_SENDIOCTL | ERROR_ETHERNET_DDAPI_IOCTL_FAIL), "No define ethernet", NULL);
    }

    delete nic_interface;

    return NO_ERROR_MSG;
}



ErrorMsg* EthernetDDAPI::Execute_external_loopback(string nic_1, string nic_2, float *throughput)
{
//   char nic_1[10] = "e0a", nic_2[10] = "e0a";
   int send_socket, receive_socket; /*socket descriptor */
   int result = 0, i;
   unsigned char src_mac[6];  /*our MAC address */
   unsigned char dest_mac[6]; /*other host MAC address */
   unsigned char ethernet_frame[10000];
   struct ethhdr *ethernet_header = (struct ethhdr *) ethernet_frame; /*Another pointer to ethernet header */
   unsigned char *data = ethernet_frame + 14; /*Userdata in ethernet frame */
   int interface_index_send = 0;
   struct sockaddr_ll sll_send;
 
   unsigned char receive_buffer[18000];
   struct ethhdr *receive_header = (struct ethhdr *) receive_buffer;
   int interface_index_receive = 0;
   long long total_receive = 0, total_tx = 0, total_rx = 0;
   struct sockaddr_ll sll_receive;
   int mtu = 1500;
   float speed = 0;
 
   struct timeval begin;
   struct timeval end;
   struct timeval time_result;

   struct ifreq ifstruct;
 
   struct timeval timeout_begin;
   struct timeval timeout_end;
   struct timeval timeout_result;
 
   memset(receive_buffer, 0, sizeof(receive_buffer));
   memset(&sll_send, 0, sizeof(sll_send));
   memset(&sll_receive, 0, sizeof(sll_receive));

   errno = 0;

   //send_socket = socket(AF_PACKET, SOCK_RAW, htons(ETH_P_ALL));
   send_socket = socket(AF_PACKET, SOCK_RAW, htons (ETH_P_IP));
   if (send_socket == -1)
   {
      //perror("Send socket()");
      return returnErrorMsg( (ERRORCODE)(ERROR_ETHERNET_DDAPI_EXECUTE_EXTERNAL_LOOPBACK | ERROR_ETHERNET_DDAPI_SEND_SOCKET_OPEN_FAIL), strerror(errno), NULL);
   }
   /*else
   {
      printf("Send socket create ok\n");
   }*/
 
   receive_socket = socket(AF_PACKET, SOCK_RAW, htons (ETH_P_IP));
   if (receive_socket == -1)
   {
     // perror("Receive socket()");
      close(send_socket);
      return returnErrorMsg( (ERRORCODE)(ERROR_ETHERNET_DDAPI_EXECUTE_EXTERNAL_LOOPBACK | ERROR_ETHERNET_DDAPI_RECEIVE_SOCKET_OPEN_FAIL), strerror(errno), NULL);
   }
   /*else
   {
      printf ("Receive socket create ok\n");
   }*/
 
   strcpy(ifstruct.ifr_name, nic_1.c_str());
   result = ioctl(send_socket, SIOCGIFINDEX, &ifstruct);
   interface_index_send = ifstruct.ifr_ifindex;
   if (result == -1)
   {
      //perror("SIOCGIFINDEX");
      close(send_socket);
      close(receive_socket);
      return returnErrorMsg( (ERRORCODE)(ERROR_ETHERNET_DDAPI_EXECUTE_EXTERNAL_LOOPBACK | ERROR_ETHERNET_DDAPI_IOCTL_SIOCGIFINDEX), strerror(errno), NULL);
   }
   //printf ("Send interface ID:%d\n", ifstruct.ifr_ifindex);
 
   /*retrieve corresponding MAC */
   if (ioctl(send_socket, SIOCGIFHWADDR, &ifstruct) == -1)
   {
       //perror("SIOCGIFHWADDR");
       close(send_socket);
       close(receive_socket);
       return returnErrorMsg( (ERRORCODE)(ERROR_ETHERNET_DDAPI_EXECUTE_EXTERNAL_LOOPBACK | ERROR_ETHERNET_DDAPI_IOCTL_SIOCGIFHWADDR), strerror(errno), NULL);
   }
 
   //printf("Send MAC: ");
   for (i = 0; i < 6; i++)
   {
       src_mac[i] = ifstruct.ifr_hwaddr.sa_data[i];
   /*    if (i == 5)
 	        printf ("%02x\n", src_mac[i]);
       else
 	        printf ("%02x:", src_mac[i]);*/
   }
 
   /*retrieve corresponding MTU */
   if (ioctl(send_socket, SIOCGIFMTU, &ifstruct) == -1)
   {
      //perror("SIOCGIFMTU");
      close(send_socket);
      return returnErrorMsg( (ERRORCODE)(ERROR_ETHERNET_DDAPI_EXECUTE_EXTERNAL_LOOPBACK | ERROR_ETHERNET_DDAPI_IOCTL_SIOCGIFMTU), strerror(errno), NULL);
   }
   mtu = ifstruct.ifr_mtu;
   //printf("MTU: %d\n",mtu);
 
 
   strcpy (ifstruct.ifr_name, nic_2.c_str());
   result = ioctl(receive_socket, SIOCGIFINDEX, &ifstruct);
   interface_index_receive = ifstruct.ifr_ifindex;
   if (result == -1)
   {
       //perror("SIOCGIFINDEX");
       close(send_socket);
       close(receive_socket);
       return returnErrorMsg( (ERRORCODE)(ERROR_ETHERNET_DDAPI_EXECUTE_EXTERNAL_LOOPBACK | ERROR_ETHERNET_DDAPI_IOCTL_SIOCGIFINDEX), strerror(errno), NULL);
   }
   //printf ("Receive interface ID:%d\n", ifstruct.ifr_ifindex);
 
   /*retrieve corresponding MAC */
   if (ioctl(receive_socket, SIOCGIFHWADDR, &ifstruct) == -1)
   {
       //perror("SIOCGIFHWADDR");
       close(send_socket);
       close(receive_socket);
       return returnErrorMsg( (ERRORCODE)(ERROR_ETHERNET_DDAPI_EXECUTE_EXTERNAL_LOOPBACK | ERROR_ETHERNET_DDAPI_IOCTL_SIOCGIFHWADDR), strerror(errno), NULL);
   }
 
   //printf ("Receive MAC: ");
   for (i = 0; i < 6; i++)
   {
      dest_mac[i] = ifstruct.ifr_hwaddr.sa_data[i];
   /*   if (i == 5)
     	   printf ("%02x\n", dest_mac[i]);
      else
     	   printf ("%02x:", dest_mac[i]);*/
   }
 
 
   //Bind the send nic
   sll_send.sll_family = AF_PACKET;
   sll_send.sll_ifindex = interface_index_send;
   //sll_send.sll_protocol = htons(ETH_P_IP);
   sll_send.sll_protocol = htons (ETH_P_ALL);
 
   if (bind(send_socket, (struct sockaddr *)&sll_send, sizeof(sll_send)) == -1)
   {
       //perror("Send bind()");
       close(send_socket);
       close(receive_socket);
       return returnErrorMsg( (ERRORCODE)(ERROR_ETHERNET_DDAPI_EXECUTE_EXTERNAL_LOOPBACK | ERROR_ETHERNET_DDAPI_SOCKET_BIND), strerror(errno), NULL);
   }

   //Bind the receive nic
   sll_receive.sll_family = AF_PACKET;
   sll_receive.sll_ifindex = interface_index_receive;
   sll_receive.sll_protocol = htons (ETH_P_ALL);
   if (bind(receive_socket, (struct sockaddr *)&sll_receive, sizeof(sll_receive)) == -1)
   {
       //perror("Receive bind()");
       close(send_socket);
       close(receive_socket);
       return returnErrorMsg( (ERRORCODE)(ERROR_ETHERNET_DDAPI_EXECUTE_EXTERNAL_LOOPBACK | ERROR_ETHERNET_DDAPI_SOCKET_BIND), strerror(errno), NULL);
   }

   /*prepare ethernet_frame */
   memcpy(ethernet_frame, dest_mac, ETH_ALEN);
   memcpy(ethernet_frame + ETH_ALEN, src_mac, ETH_ALEN);
   //ethernet_header->h_proto = 0x0800;		//packet type:IP v4
   ethernet_header->h_proto = mtu;
 
   /*fill the data */
   for (i = 0; i < mtu; i++)
     data[i] = 0xAA;
 
   /*clear the timers: */
   timerclear(&begin);
   timerclear(&end);

   timerclear(&timeout_begin);
   timerclear(&timeout_end);
 
   /*get time before sending */
   gettimeofday(&begin, NULL);

   gettimeofday(&timeout_begin, NULL);

   long receive_start, receive_end;
   long transmit_start, transmit_end;
   get_receive_packet_number(nic_1, nic_2, &transmit_start, &receive_start);

   //Start the speed test procedure
   while (1)
   {
      timersub(&timeout_end, &timeout_begin, &timeout_result);
      if (timeout_result.tv_sec > 3)
      {
         //Waiting for data transfer complete
         sleep(1);
         break;
      }

      result = sendto(send_socket, ethernet_frame, ETH_HLEN + mtu, 0, (struct sockaddr *) &sll_send, sizeof (sll_send));
      if (result == -1)
      {
         //perror("sendto()");
         return returnErrorMsg( (ERRORCODE)(ERROR_ETHERNET_DDAPI_EXECUTE_EXTERNAL_LOOPBACK | ERROR_ETHERNET_DDAPI_SENDTO), strerror(errno), NULL);
      }
      if(result == 0)
         printf("Send \"0\"\n");
      total_tx = total_tx + result;
      /*Wait for incoming packet... */
      // result = recvfrom(receive_socket, receive_buffer, sizeof(receive_buffer), 0, NULL, NULL);
      result = recvfrom(receive_socket, receive_buffer, sizeof(receive_buffer), MSG_DONTWAIT, NULL, NULL);
      gettimeofday(&end, NULL);
      timersub(&end, &begin, &time_result);
      if (time_result.tv_sec > 5)
      {
         //printf("Recv timeout!\n");
         return returnErrorMsg( (ERRORCODE)(ERROR_ETHERNET_DDAPI_EXECUTE_EXTERNAL_LOOPBACK | ERROR_ETHERNET_DDAPI_RECV_TIMEOUT), NULL, NULL);
      }
      if (result == -1 && errno != EAGAIN)
      {
         //perror("recvfrom()");
         return returnErrorMsg( (ERRORCODE)(ERROR_ETHERNET_DDAPI_EXECUTE_EXTERNAL_LOOPBACK | ERROR_ETHERNET_DDAPI_RECVFROM), strerror(errno), NULL);
      }

      //Only the packet we created need to be checked, other drop it.
      if (memcmp(receive_header->h_source, src_mac, ETH_ALEN) == 0 &&
          memcmp(receive_header->h_dest, dest_mac, ETH_ALEN) == 0 && result != -1)
      {
         //Check the recv user data is the same send user data
         if (memcmp(ethernet_frame, receive_buffer, ETH_HLEN + mtu) != 0)
         {
            /*printf("Compare error!!!\n");
            for (i = 0; i < result; i++)
               printf ("%02x ", receive_buffer[i]);
            printf ("\n");*/
            break;
         }
         total_rx = total_rx + result;
         total_receive = total_receive + result;
         timersub (&end, &begin, &time_result);
         if (time_result.tv_sec > 0)
         //if (time_result.tv_usec > 999000)
         {
            speed = total_receive * 8 / 1024.0 / 1024.0 / (time_result.tv_sec + (time_result.tv_usec / 1000000.0));

            //Show the measured speed
            /*printf ("Total_recv=%lld byte ", total_receive);
            printf ("Time:%f sec. ", time_result.tv_sec + (time_result.tv_usec / 1000000.0));
            //printf ("Speed:%.2f Mbit/s       \r", total_receive * 8 / 1024.0 / 1024.0 / (time_result.tv_sec + (time_result.tv_usec / 1000000.0)));
            if (speed > 1024)
            {
               printf("Speed:%.2f Gbit/s       \r", speed /1024);
            }
            else
            {
               printf ("Speed:%.2f Mbit/s       \r", speed);
            }
            fflush(stdout);*/
 
            total_receive = 0;
            //clear the timers:
            timerclear (&begin);
            timerclear (&end);
            gettimeofday (&begin, NULL);
         }
      }
      gettimeofday(&timeout_end, NULL);
   }

   get_receive_packet_number(nic_1, nic_2, &transmit_end, &receive_end);

   //cout << "tx start: " << transmit_start << "  rx start: " << receive_start << endl;
   //cout << "tx end:   " << transmit_end << "  rx end:   " << receive_end << endl;
   //cout << "compare: tx:" << transmit_end - transmit_start << "  rx:" << receive_end - receive_start << "  tx-rx:" << (transmit_end - transmit_start)-(receive_end - receive_start) << endl;

   //cout << "Send:" << total_tx << endl;
   //cout << "recv:" << total_rx << endl;

   if (receive_end - receive_start < 20)
      speed = 0;

   close (send_socket);
   close (receive_socket);
   *throughput = speed;

   return NO_ERROR_MSG;
}



ErrorMsg* EthernetDDAPI::get_device_bus_type(string *bus)
{
   char   linkname[100];
   char   *pch;
   string bus_type_path = "/sys/class/net/" + m_ethernet_name + "/device/subsystem";
   string bus_type_result;

   //cout << bus_type_path << endl;
   if (readlink(bus_type_path.c_str(), linkname, 100) == -1)
   {
      return returnErrorMsg( (ERRORCODE)(ERROR_ETHERNET_DDAPI_GET_DEVICE_BUS_TYPE | ERROR_ETHERNET_DDAPI_READ_LINK_FAIL), strerror(errno), NULL);
   }

   pch = strtok(linkname, "/");
   while (pch != NULL)
   {
      //printf("%s\n",pch);
      bus_type_result = pch;
      pch = strtok(NULL, "/");
   }
   //printf("Final:%s\n",str.c_str());

   *bus = bus_type_result;

   return NO_ERROR_MSG;
}



ErrorMsg* EthernetDDAPI::get_receive_packet_number(string nic1_name, string nic2_name, long *transmit, long *receive_end)
{
   ifstream infile;
   string   node_path = "/sys/class/net/" + nic2_name + "/statistics/" + "rx_packets";
   string   value;

   infile.open(node_path.c_str());
   if (infile.is_open())
   {
      getline(infile, value); // Saves the line in STRING.
      infile.close();
      sscanf(value.c_str(), "%ld", receive_end);
   }
   else
   {
      //return returnErrorMsg((ERRORCODE)(ERROR_ETHERNET_DAPI_GET_NODE_VALUE | ERROR_ETHERNET_DAPI_NODE_OPEN_FAIL), node_path.c_str(), NULL);
   }


   node_path = "/sys/class/net/" + nic1_name + "/statistics/" + "tx_packets";
   infile.open(node_path.c_str());
   if (infile.is_open())
   {
      getline(infile, value); // Saves the line in STRING.
      infile.close();
      sscanf(value.c_str(), "%ld", transmit);
   }
   else
   {
      //return returnErrorMsg((ERRORCODE)(ERROR_ETHERNET_DAPI_GET_NODE_VALUE | ERROR_ETHERNET_DAPI_NODE_OPEN_FAIL), node_path.c_str(), NULL);
   }

   return NO_ERROR_MSG;
}



ErrorMsg* EthernetDDAPI::ConverWellKnowNameToRealName(string name, string *real_name)
{
    FILE *pFile;
    char line[500], teg[20];
    int  teg_number, string_length, counter = 0;
    char *pch;


    pFile = fopen ("Ethernet_profile.ini","r");
    if (pFile == NULL)
    {
       return returnErrorMsg( (ERRORCODE)(ERROR_ETHERNET_DDAPI_GET_NAME_LIST | ERROR_OPEN_FILE), strerror(errno), NULL);
    }

    //Read each line of Ethernet_profile.ini
    while (fgets(line, 500, pFile) != NULL)
    {
        if (line[0] == '#')
            continue;

        teg_number = sscanf(line, " [%[^]]", teg); //paser the teg: Ethernet
        if ( (strlen(line) == 1 && line[0] == '\n') || (teg_number != 0))
        {
            continue;
        }

        //Remove the '\n' at the end of line
        string_length = strlen(line);
        if (line[string_length-1] == '\n')
        {
            line[string_length-1] = '\0';
        }

        //Start parser by teg
        if ( strcmp(teg,"Ethernet") == 0 )
        {
            //Start parse the line
            pch = strtok(line, " ");

            while (pch != NULL)
            {
                counter++;

                //The first item(Well known name)
                if (counter == 1)
                {
                    if (name.compare(pch) == 0)
                    {
                        pch = strtok (NULL, " ");
                        real_name->assign(pch);
                        break;
                    }
                }

                pch = strtok (NULL, " ");
            }
            counter = 0;
        }
    }

    fclose(pFile);

    return NO_ERROR_MSG;
}



ErrorMsg* EthernetDDAPI::Program(string ethernet_name, string filePath)
{
    string vendor_ID;
    string device_ID;
    EthernetBaseClass *nic_interface = NULL;

    m_ethernet_name = ethernet_name;

    //Get ethernet PCI configuration vendor ID and device ID
    m_errorCode = get_vendor_ID(&vendor_ID);
    if (m_errorCode != NO_ERROR_MSG)
    {
       return returnErrorMsg(ERROR_ETHERNET_DDAPI_SENDIOCTL, "Get vender ID fail", m_errorCode);
    }

    m_errorCode = get_device_ID(&device_ID);
    if (m_errorCode != NO_ERROR_MSG)
    {
       return returnErrorMsg(ERROR_ETHERNET_DDAPI_SENDIOCTL, "Get device ID fail", m_errorCode);
    }
    //cout << "vendor_ID: " <<vendor_ID << "  device_ID:" << device_ID << endl;

    //Select which type of ethernet to create object
    if (vendor_ID.compare("0x8086") == 0) //INTEL chip
    {
        nic_interface = new INTEL_I210(m_ethernet_name.c_str());
    }
    if (vendor_ID.compare("0x1425") == 0 && device_ID.compare("0x6407") == 0) //Chelsio T62100
    {
        nic_interface = new CHELSIS_T62100(m_ethernet_name.c_str());
    }
    if (vendor_ID.compare("0x1425") == 0 && device_ID.compare("0x640d") == 0) //Chelsio T62100-CR
    {
        nic_interface = new CHELSIS_T62100(m_ethernet_name.c_str());
    }
    if (vendor_ID.compare("0x1425") == 0 && device_ID.compare("0x6401") == 0) //Chelsio T6225-CR
    {
        nic_interface = new CHELSIS_T62100(m_ethernet_name.c_str());
    }

    if (nic_interface != NULL)
    {
        //Start program eeprom
        m_errorCode = nic_interface->Program(ethernet_name, filePath);

        if (m_errorCode != NO_ERROR_MSG)
        {
           return returnErrorMsg(ERROR_ETHERNET_DDAPI_SENDIOCTL, "Program EEPROM fail", m_errorCode);
        }
    }
    else
    {
        return returnErrorMsg( (ERRORCODE)(ERROR_ETHERNET_DDAPI_SENDIOCTL | ERROR_ETHERNET_DDAPI_IOCTL_FAIL), "No define ethernet", NULL);
    }

    delete nic_interface;

    return NO_ERROR_MSG;
}



ErrorMsg* EthernetDDAPI::ReadSFP(string ethernet_name, INT32 DeviceAddress, INT32 StartAddress, INT32 Length, INT32 *data)
{
    string vendor_ID;
    string device_ID;
    EthernetBaseClass *nic_interface = NULL;

    m_ethernet_name = ethernet_name;

    //Get ethernet PCI configuration vendor ID and device ID
    m_errorCode = get_vendor_ID(&vendor_ID);
    if (m_errorCode != NO_ERROR_MSG)
    {
       return returnErrorMsg(ERROR_ETHERNET_DDAPI_SENDIOCTL, "Get vender ID fail", m_errorCode);
    }

    m_errorCode = get_device_ID(&device_ID);
    if (m_errorCode != NO_ERROR_MSG)
    {
       return returnErrorMsg(ERROR_ETHERNET_DDAPI_SENDIOCTL, "Get device ID fail", m_errorCode);
    }
    //cout << "vendor_ID: " <<vendor_ID << "  device_ID:" << device_ID << endl;

    //Select which type of ethernet to create object
    if (vendor_ID.compare("0x8086") == 0 && (device_ID.compare("0x1533") == 0 || device_ID.compare("0x1537") == 0)) //INTEL I210
    {
        return returnErrorMsg(ERROR_ETHERNET_DDAPI_SENDIOCTL, "Intel I210 not support", NULL);
    }
    if (vendor_ID.compare("0x8086") == 0 && device_ID.compare("0x15ab") == 0) //INTEL X552(KR)
    {
        return returnErrorMsg(ERROR_ETHERNET_DDAPI_SENDIOCTL, "Intel X552(KR) not support", NULL);
    }
    if (vendor_ID.compare("0x8086") == 0 && device_ID.compare("0x1583") == 0) //INTEL XL710
    {
        nic_interface = new INTEL_XL710(m_ethernet_name.c_str());
    }
    if (vendor_ID.compare("0x8086") == 0 && device_ID.compare("0x1572") == 0) //INTEL XL710
    {
        nic_interface = new INTEL_XL710(m_ethernet_name.c_str());
    }
    if (vendor_ID.compare("0x1425") == 0 && device_ID.compare("0x6407") == 0) //Chelsio T62100
    {
        //nic_interface = new CHELSIS_T62100(m_ethernet_name.c_str());
        return returnErrorMsg((ERRORCODE)(ERROR_ETHERNET_DDAPI_ENABLE_EXTERNAL_LOOPBACK | ERROR_FUNCTION_NOT_SUPPORTED));
    }
    if (vendor_ID.compare("0x1425") == 0 && device_ID.compare("0x640d") == 0) //Chelsio T62100-CR
    {
        //nic_interface = new CHELSIS_T62100(m_ethernet_name.c_str());
        return returnErrorMsg((ERRORCODE)(ERROR_ETHERNET_DDAPI_ENABLE_EXTERNAL_LOOPBACK | ERROR_FUNCTION_NOT_SUPPORTED));
    }
    if (vendor_ID.compare("0x1425") == 0 && device_ID.compare("0x6401") == 0) //Chelsio T6225-CR
    {
        //nic_interface = new CHELSIS_T62100(m_ethernet_name.c_str());
        return returnErrorMsg((ERRORCODE)(ERROR_ETHERNET_DDAPI_ENABLE_EXTERNAL_LOOPBACK | ERROR_FUNCTION_NOT_SUPPORTED));
    }


    if (nic_interface != NULL)
    {
        //Start get SFP data
        m_errorCode = nic_interface->ReadSFP(ethernet_name, DeviceAddress, StartAddress, Length, data);

        if (m_errorCode != NO_ERROR_MSG)
        {
           return returnErrorMsg(ERROR_ETHERNET_DDAPI_SENDIOCTL, "Read SFP fail", m_errorCode);
        }
    }
    else
    {
        return returnErrorMsg( (ERRORCODE)(ERROR_ETHERNET_DDAPI_SENDIOCTL | ERROR_ETHERNET_DDAPI_IOCTL_FAIL), "No define ethernet", NULL);
    }

    delete nic_interface;

    return NO_ERROR_MSG;
}



ErrorMsg* EthernetDDAPI::WriteSFP(string ethernet_name, INT32 DeviceAddress, INT32 StartAddress, INT32 data)
{
    string vendor_ID;
    string device_ID;
    EthernetBaseClass *nic_interface = NULL;

    m_ethernet_name = ethernet_name;

    //Get ethernet PCI configuration vendor ID and device ID
    m_errorCode = get_vendor_ID(&vendor_ID);
    if (m_errorCode != NO_ERROR_MSG)
    {
       return returnErrorMsg(ERROR_ETHERNET_DDAPI_SENDIOCTL, "Get vender ID fail", m_errorCode);
    }

    m_errorCode = get_device_ID(&device_ID);
    if (m_errorCode != NO_ERROR_MSG)
    {
       return returnErrorMsg(ERROR_ETHERNET_DDAPI_SENDIOCTL, "Get device ID fail", m_errorCode);
    }
    //cout << "vendor_ID: " <<vendor_ID << "  device_ID:" << device_ID << endl;

    //Select which type of ethernet to create object
    if (vendor_ID.compare("0x8086") == 0 && (device_ID.compare("0x1533") == 0 || device_ID.compare("0x1537") == 0)) //INTEL I210
    {
        return returnErrorMsg(ERROR_ETHERNET_DDAPI_SENDIOCTL, "Intel I210 not support", NULL);
    }
    if (vendor_ID.compare("0x8086") == 0 && device_ID.compare("0x15ab") == 0) //INTEL X552(KR)
    {
        return returnErrorMsg(ERROR_ETHERNET_DDAPI_SENDIOCTL, "Intel X552(KR) not support", NULL);
    }
    if (vendor_ID.compare("0x8086") == 0 && device_ID.compare("0x1583") == 0) //INTEL XL710
    {
        nic_interface = new INTEL_XL710(m_ethernet_name.c_str());
    }
    if (vendor_ID.compare("0x8086") == 0 && device_ID.compare("0x1572") == 0) //INTEL XL710
    {
        nic_interface = new INTEL_XL710(m_ethernet_name.c_str());
    }
    if (vendor_ID.compare("0x1425") == 0 && device_ID.compare("0x6407") == 0) //Chelsio T62100
    {
        //nic_interface = new CHELSIS_T62100(m_ethernet_name.c_str());
        return returnErrorMsg((ERRORCODE)(ERROR_ETHERNET_DDAPI_ENABLE_EXTERNAL_LOOPBACK | ERROR_FUNCTION_NOT_SUPPORTED));
    }
    if (vendor_ID.compare("0x1425") == 0 && device_ID.compare("0x640d") == 0) //Chelsio T62100-CR
    {
        //nic_interface = new CHELSIS_T62100(m_ethernet_name.c_str());
        return returnErrorMsg((ERRORCODE)(ERROR_ETHERNET_DDAPI_ENABLE_EXTERNAL_LOOPBACK | ERROR_FUNCTION_NOT_SUPPORTED));
    }
    if (vendor_ID.compare("0x1425") == 0 && device_ID.compare("0x6401") == 0) //Chelsio T6225-CR
    {
        //nic_interface = new CHELSIS_T62100(m_ethernet_name.c_str());
        return returnErrorMsg((ERRORCODE)(ERROR_ETHERNET_DDAPI_ENABLE_EXTERNAL_LOOPBACK | ERROR_FUNCTION_NOT_SUPPORTED));
    }

    if (nic_interface != NULL)
    {
        //Start get SFP data
        m_errorCode = nic_interface->WriteSFP(ethernet_name, DeviceAddress, StartAddress, data);

        if (m_errorCode != NO_ERROR_MSG)
        {
           return returnErrorMsg(ERROR_ETHERNET_DDAPI_SENDIOCTL, "Write SFP fail", m_errorCode);
        }
    }
    else
    {
        return returnErrorMsg( (ERRORCODE)(ERROR_ETHERNET_DDAPI_SENDIOCTL | ERROR_ETHERNET_DDAPI_IOCTL_FAIL), "No define ethernet", NULL);
    }

    delete nic_interface;

    return NO_ERROR_MSG;
}



ErrorMsg* EthernetDDAPI::getLEDList(list<string> *out_pLEDList)
{
    list<struct ethernet_data> net_list;
    EthernetBaseClass          *nic_interface = NULL;
    string                     vendor_ID, device_ID;
    FILE                       *pFile = NULL;
    CHAR                       line[500], nic_LED_name[20];
    list<string>               LED_black_list;

    //Open the Ethernet_LED_profile.ini for parser
    #ifdef DEBUG
        pFile = fopen ("Ethernet_LED_profile.ini","r");
    #else
        pFile = fopen ("/etc/lemur/Ethernet_LED_profile.ini","r");
    #endif

    if (pFile != NULL)
    {
        //Read each line of Ethernet_LED_profile.ini
        while (fgets(line, 500, pFile) != NULL)
        {
            if (line[0] == '#')
                continue;

            if ( strlen(line) == 1 && line[0] == '\n' )
                continue;

            sscanf(line, "%s ", nic_LED_name);
            //convert a string to all lower case
            for(int i = 0; nic_LED_name[i]; i++)
            {
                nic_LED_name[i] = tolower(nic_LED_name[i]);
            }

            //push LED string in balcklist
            LED_black_list.push_back(nic_LED_name);
        }
        fclose(pFile);
    }

    //Get all system ethernet list
    GetEthernetNamesList(&net_list);

    //Retrieve LED name by each ethernet name
    while (!net_list.empty())
    {
        m_ethernet_name = net_list.front().name;
        //Get ethernet PCI vender and device ID
        get_vendor_ID(&vendor_ID);
        get_device_ID(&device_ID);

        //cout << ' ' << net_list.front().name << ":" << vendor_ID << " " << device_ID << endl;
        if (vendor_ID.compare("0x8086") == 0 && (device_ID.compare("0x1533") == 0 || device_ID.compare("0x1537") == 0)) //INTEL I210
        {
            nic_interface = new INTEL_I210(m_ethernet_name.c_str());
        }
        if (vendor_ID.compare("0x8086") == 0 && device_ID.compare("0x1583") == 0) //INTEL XL710
        {
            nic_interface = new INTEL_XL710(m_ethernet_name.c_str());
        }
        if (vendor_ID.compare("0x8086") == 0 && device_ID.compare("0x1572") == 0) //INTEL XL710
        {
            nic_interface = new INTEL_XL710(m_ethernet_name.c_str());
        }

        if (nic_interface != NULL)
        {
            nic_interface->getLEDName(m_ethernet_name, out_pLEDList);
            delete nic_interface;
            nic_interface = NULL;
        }
        net_list.pop_front();
    }

    //Remove specify LED form blacklist
    while( !LED_black_list.empty() )
    {
        out_pLEDList->remove(LED_black_list.front());
        LED_black_list.pop_front();
    }

    //For debug
    /*while (!out_pLEDList->empty())
    {
        cout << " " << out_pLEDList->front().c_str();
        out_pLEDList->pop_front();
    }
    printf("\n");*/

    return NO_ERROR_MSG;
}



ErrorMsg* EthernetDDAPI::setLEDStatus(string in_strLEDName, string in_strStatus)
{
    CHAR              nic_name[30], nic_led[30];
    EthernetBaseClass *nic_interface = NULL;
    string            vendor_ID, device_ID;

    sscanf(in_strLEDName.c_str(), "%[^_]_%s", nic_name, nic_led);
    //printf("%s   %s\n", nic_name, nic_led);

    m_ethernet_name = nic_name;
    //Get ethernet PCI vender and device ID
    get_vendor_ID(&vendor_ID);
    get_device_ID(&device_ID);
    //cout << m_ethernet_name.c_str() <<":" << vendor_ID << " " << device_ID << endl;

    if (vendor_ID.compare("0x8086") == 0 && (device_ID.compare("0x1533") == 0 || device_ID.compare("0x1537") == 0)) //INTEL I210
    {
        nic_interface = new INTEL_I210(m_ethernet_name.c_str());
    }
    if (vendor_ID.compare("0x8086") == 0 && device_ID.compare("0x1583") == 0) //INTEL XL710
    {
        nic_interface = new INTEL_XL710(m_ethernet_name.c_str());
    }
    if (vendor_ID.compare("0x8086") == 0 && device_ID.compare("0x1572") == 0) //INTEL XL710
    {
        nic_interface = new INTEL_XL710(m_ethernet_name.c_str());
    }

    //Turn on/off LED
    if (nic_interface != NULL)
    {
        m_errorCode = nic_interface->setLEDStatus(nic_name, nic_led, in_strStatus);
        delete nic_interface;
        nic_interface = NULL;

        if (m_errorCode != NO_ERROR_MSG)
        {
            return returnErrorMsg( (ERRORCODE)(ERROR_ETHERNET_DDAPI_GET_NAME_LIST), NULL, m_errorCode);
        }
    }

    return NO_ERROR_MSG;
}



ErrorMsg* EthernetDDAPI::getLEDStatus(string in_strLEDName, string *out_strpStatus)
{
    CHAR              nic_name[30], nic_led[30];
    EthernetBaseClass *nic_interface = NULL;
    string            vendor_ID, device_ID;

    sscanf(in_strLEDName.c_str(), "%[^_]_%s", nic_name, nic_led);
    //printf("%s   %s\n", nic_name, nic_led);

    m_ethernet_name = nic_name;
    //Get ethernet PCI vender and device ID
    get_vendor_ID(&vendor_ID);
    get_device_ID(&device_ID);
    //cout << m_ethernet_name.c_str() <<":" << vendor_ID << " " << device_ID << endl;

    if (vendor_ID.compare("0x8086") == 0 && (device_ID.compare("0x1533") == 0 || device_ID.compare("0x1537") == 0)) //INTEL I210
    {
        nic_interface = new INTEL_I210(m_ethernet_name.c_str());
    }
    if (vendor_ID.compare("0x8086") == 0 && device_ID.compare("0x1583") == 0) //INTEL XL710
    {
        nic_interface = new INTEL_XL710(m_ethernet_name.c_str());
    }
    if (vendor_ID.compare("0x8086") == 0 && device_ID.compare("0x1572") == 0) //INTEL XL710
    {
        nic_interface = new INTEL_XL710(m_ethernet_name.c_str());
    }

    //Get LED status
    if (nic_interface != NULL)
    {
        m_errorCode = nic_interface->getLEDStatus(nic_name, nic_led, out_strpStatus);
        delete nic_interface;
        nic_interface = NULL;

        if (m_errorCode != NO_ERROR_MSG)
        {
            return returnErrorMsg( (ERRORCODE)(ERROR_ETHERNET_DDAPI_GET_NAME_LIST), NULL, m_errorCode);
        }
    }

    return NO_ERROR_MSG;
}



ErrorMsg* EthernetDDAPI::getStatistics(string nic, string *statistics)
{
    char command_string[100], newline[256], item[100];
    char buffer[200];
    int  value = 0, flag = 0;
    FILE *fd;

    //sprintf(command_string, "ethtool -S %s | grep error | awk 'BEGIN {FS=\":\"} {print $1 $2}'", nic.c_str());
    sprintf(command_string, "ethtool -S %s | grep error", nic.c_str());
    fd = popen(command_string, "r");
    while((fgets(newline, 256, fd)) != NULL)
    {
        //sscanf(newline, "%[^:]:%d", item, &value);
        sscanf(newline, "     %[^0-9:]: %d", item, &value);

        if (value != 0)
        {
            flag = 1;
        }

        sprintf(buffer, "%s: %d\n", item, value);
        statistics->append(buffer);
    }
    pclose(fd);

    if (flag != 0)
    {
        return returnErrorMsg( (ERRORCODE)(ERROR_ETHERNET_DDAPI_GET_STATISTICS), "Statistics not zero", NULL);
    }

    return NO_ERROR_MSG;
}



ErrorMsg* EthernetDDAPI::clearPCIError(string nic)
{
    int       domain, bus, device, func;
    PFA       PCI_BDF_NIC;
    PCIDDAPI  PCI_driver("PCI_driver");

    //Get nic PCI BDF
    m_errorCode = GetPCIBusDeviceFunction(nic, &domain, &bus, &device, &func);
    if (m_errorCode != NO_ERROR_MSG)
    {
        return returnErrorMsg(ERROR_ETHERNET_DDAPI_CLEAR_AER, NULL, m_errorCode);
    }

    PCI_BDF_NIC.u16Domain  = domain;
    PCI_BDF_NIC.u16Bus     = bus;
    PCI_BDF_NIC.u16Device  = device;
    PCI_BDF_NIC.u8Function = func;

    //Clean nic ARE
    PCI_driver.clearErrors(PCI_BDF_NIC);
    //m_errorCode = PCI_driver.clearErrors(PCI_BDF_NIC);
    //if (m_errorCode != NO_ERROR_MSG)
    //{
    //    return returnErrorMsg(ERROR_ETHERNET_DDAPI_CLEAR_AER, NULL, m_errorCode);
    //}

    return NO_ERROR_MSG;
}



ErrorMsg* EthernetDDAPI::clearStatistics(string nic)
{
    list<struct ethernet_data>            eth_list;
    list<struct ethernet_data>::iterator  it;
    list<string>                          driver_list;
    list<string>::iterator                it_driver;
    int                                   driver_flag = 0;
    char                                  command[256];

    //Get ethernet list
    m_errorCode = GetEthernetNamesList(&eth_list);
    if (m_errorCode != NO_ERROR_MSG)
    {
        return returnErrorMsg(ERROR_ETHERNET_DDAPI_CLEAR_STATISTICS, NULL, m_errorCode);
    }

    //Search driver and add to list
    for (it = eth_list.begin(); it != eth_list.end(); ++it)
    {
        if (it->name != nic)
        {
            continue;
        }

        driver_flag = 0;
        for (it_driver = driver_list.begin(); it_driver != driver_list.end(); ++it_driver)
        {
            if ((it->driver) == *it_driver)
            {
                driver_flag = 1;
                break;
            }
        }
        if (driver_flag == 0)
        {
            driver_list.push_back(it->driver);
        }
    }


    //Reinstall driver
    for (it_driver = driver_list.begin(); it_driver != driver_list.end(); ++it_driver)
    {
        sprintf(command, "modprobe -r %s\n", it_driver->c_str());
        system(command);
        sprintf(command, "modprobe %s\n", it_driver->c_str());
        system(command);
    }

    sleep(1);
    //Setting all IP address
    for (it = eth_list.begin(); it != eth_list.end(); ++it)
    {
        if (it->ip != "0.0.0.0" )
        {
            sprintf(command, "ifconfig %s %s\n", it->name.c_str(), it->ip.c_str());
            system(command);
        }
    }

    return NO_ERROR_MSG;
}


