#include "INTEL_I210.hpp"



INTEL_I210::INTEL_I210(STRING nic_name):EthernetBaseClass(nic_name)
{
   Append_Node("/device/resource0");
   //printf("I210: %s\n",Get_Node().c_str());
   m_errorCode = NO_ERROR_MSG;
}



INTEL_I210::~INTEL_I210()
{

}



ErrorMsg* INTEL_I210::Set_external_loopback()
{
   //printf("INTEL I210 Set_external_loopback\n");

   unsigned int *addr;
   int fd;

   fd = open(Get_Node().c_str(), O_RDWR | O_SYNC);
   if (fd < 0)
   {
       //printf("%s file open error!\n",Get_Node().c_str());
       return returnErrorMsg( (ERRORCODE)(ERROR_ETHERNET_DDAPI_SET_EXTERNAL_LOOPBACK | ERROR_ETHERNET_DDAPI_OPEN_NODE_FAIL), NULL, NULL);
   }

   addr = (unsigned int *)mmap(NULL, 0x1000, PROT_READ|PROT_WRITE, MAP_SHARED, fd, 0x00);
   MDIO_point = addr + (MDIO_CONTROL_REGISTER / 4);

   //printf("0x%04x\n",read_PHY_reg(0, 0));
   //1G setup
   write_PHY_reg(18, 6, 0x0008);
   write_PHY_reg(0, 0, 0x9340);

   //100/10M setup
   //write_PHY_reg(18, 6, 0x0000);
   //write_PHY_reg(0, 0, 0x1140);

   munmap(addr, 0x1000);
   close(fd);

   return NO_ERROR_MSG;
}



ErrorMsg* INTEL_I210::Remove_external_loopback()
{
   //printf("INTEL I210 Remove_external_loopback\n");

   unsigned int *addr;
   int fd;

   fd = open(Get_Node().c_str(), O_RDWR | O_SYNC);
   if (fd < 0)
   {
       //printf("%s file open error!\n",Get_Node().c_str());
       return returnErrorMsg( (ERRORCODE)(ERROR_ETHERNET_DDAPI_SET_EXTERNAL_LOOPBACK | ERROR_ETHERNET_DDAPI_OPEN_NODE_FAIL), NULL, NULL);
   }

   addr = (unsigned int *)mmap(NULL, 0x1000, PROT_READ|PROT_WRITE, MAP_SHARED, fd, 0x00);
   MDIO_point = addr + (MDIO_CONTROL_REGISTER / 4);

   //printf("0x%04x\n",read_PHY_reg(0, 0));

   //100/10M setup
   write_PHY_reg(18, 6, 0x0000);
   write_PHY_reg(0, 0, 0x1140);

   munmap(addr, 0x1000);
   close(fd);

   return NO_ERROR_MSG;
}



int INTEL_I210::read_PHY_reg(int reg, int page)
{
   //write page(Register 22)
   if (reg != 22)
   {
      *MDIO_point =  MDIO_WRITE | ((0x16 & PHY_REG_NUM) << 16) | (page & PHY_DATA_LEN);
      usleep(1);
   }

   *MDIO_point =  MDIO_READ | ((reg & PHY_REG_NUM) << 16);
   usleep(1);
   return *MDIO_point & 0x0FFFF;
}



int INTEL_I210::write_PHY_reg(int reg, int page , int data)
{
   //write page(Register 22)
   if (reg != 22)
   {
      *MDIO_point =  MDIO_WRITE | ((0x16 & PHY_REG_NUM) << 16) | (page & PHY_DATA_LEN);
      usleep(1);
   }

   *MDIO_point =  MDIO_WRITE | ((reg & PHY_REG_NUM) << 16) | (data & PHY_DATA_LEN);
   usleep(1);

   return 0;
}



ErrorMsg* INTEL_I210::Get_self_test_result(struct ethtool_test *test, TestType index, INT32 *test_result)
{
    //printf("test->data[0]:%d  test->data[1]:%d  test->data[2]:%d  test->data[3]:%d  test->data[4]:%d\n"
    //       ,test->data[0], test->data[1], test->data[2], test->data[3], test->data[4]);

    switch(index)
    {
        case Register:
            *test_result = test->data[0];
            if (*test_result == 1)
                return returnErrorMsg( (ERRORCODE)(ERROR_ETHERNET_DDAPI_SENDIOCTL | ERROR_ETHERNET_DDAPI_REGISTER_TEST_FAIL), NULL, NULL);
            break;
        case EepromValidation:
            *test_result = test->data[1];
            if (*test_result == 1)
                return returnErrorMsg( (ERRORCODE)(ERROR_ETHERNET_DDAPI_SENDIOCTL | ERROR_ETHERNET_DDAPI_EEPROM_VALIDATION_TEST_FAIL), NULL, NULL);
            break;
        case Interrupt:
            *test_result = test->data[2];
            if (*test_result == 1)
                return returnErrorMsg( (ERRORCODE)(ERROR_ETHERNET_DDAPI_SENDIOCTL | ERROR_ETHERNET_DDAPI_INTERRUPT_TEST_FAIL), NULL, NULL);
            break;
        case InternalLoopback:
            *test_result = test->data[3];
            if (*test_result == 1)
                return returnErrorMsg( (ERRORCODE)(ERROR_ETHERNET_DDAPI_SENDIOCTL | ERROR_ETHERNET_DDAPI_INTERNAL_LOOPBACK_TEST_FAIL), NULL, NULL);
            break;
        case LinkStatus:
            *test_result = test->data[4];
            if (*test_result == 1)
                return returnErrorMsg( (ERRORCODE)(ERROR_ETHERNET_DDAPI_SENDIOCTL | ERROR_ETHERNET_DDAPI_MDI_REGISTER_TEST_FAIL), NULL, NULL);
            break;
        default:
            return returnErrorMsg( (ERRORCODE)(ERROR_ETHERNET_DDAPI_SENDIOCTL | ERROR_ETHERNET_DDAPI_SENDIOCTL_UNDEFINED), NULL, NULL);
    }

    return NO_ERROR_MSG;
}



ErrorMsg* INTEL_I210::SetupSpeed(string nic, INT32 speed)
{
    int    duplex = 0x01;
    int    err = 0, fd;
    string link_up;
    int    sleep_count = 0;

    struct ifreq ifr;
    struct ethtool_cmd ecmd;

    fd = socket(AF_INET,SOCK_DGRAM,0);
    if (fd < 0)
    {
       //perror("socket");
       close(fd);
       return returnErrorMsg((ERRORCODE)(ERROR_ETHERNET_DAPI_SETUP_SPEED | ERROR_ETHERNET_DAPI_SOCKET_OPEN_FAIL), strerror(errno), NULL);
    }

    strncpy(ifr.ifr_name, nic.c_str(), sizeof(ifr.ifr_name));

    ecmd.cmd = ETHTOOL_GSET;
    ifr.ifr_data = (__caddr_t)&ecmd;
    err = ioctl(fd, SIOCETHTOOL, &ifr);
    if (err < 0)
    {
       //perror("Cannot get current device settings");
       close(fd);
       return returnErrorMsg((ERRORCODE)(ERROR_ETHERNET_DAPI_SETUP_SPEED | ERROR_ETHERNET_DAPI_GET_ETHERNET_DRIVER_SET_FAIL), strerror(errno), NULL);
    }

    //If port is FIBER and speed is 1000 then return NO_ERROR_MSG or return error.
    if (ecmd.port == PORT_FIBRE)
    {
        if(speed == 1000)
        {
            close(fd);
            return NO_ERROR_MSG;
        }
        else
        {
            return returnErrorMsg((ERRORCODE)(ERROR_ETHERNET_DAPI_SETUP_SPEED | ERROR_ETHERNET_DAPI_SET_ETHERNET_DRIVER_SET_FAIL), "Speed not supported", NULL);
        }
    }

    //Speed setup
    ecmd.speed = (__u16)speed;
    ecmd.speed_hi = (__u16)(speed >> 16);

    //Duplex setup
    ecmd.duplex = duplex;  //0x01 for FULL duplex, 0x00 Half

    //Autonegotiation setup. 1G MUST enable.
    if (speed >= 1000)
       ecmd.autoneg = 0x01; //disable autonegotiation.  If this is set to enable, the forced link modes are completely ignored.
    else
       ecmd.autoneg = 0x00;

    /* Try to perform the update. */
    ecmd.cmd = ETHTOOL_SSET;
    err = ioctl(fd, SIOCETHTOOL, &ifr);
    if (err < 0)
    {
       //perror("Cannot set new settings");
       close(fd);
       return returnErrorMsg((ERRORCODE)(ERROR_ETHERNET_DAPI_SETUP_SPEED | ERROR_ETHERNET_DAPI_SET_ETHERNET_DRIVER_SET_FAIL), strerror(errno), NULL);
    }

    GetLinkStatus(nic, &link_up);

    while (link_up.compare("Down") == 0)
    {
        //printf("Wait link UP!!!\n");
        sleep(1);
        sleep_count++;
        if (sleep_count > 6)
        {
            //Reset I210
            ecmd.cmd = ETHTOOL_SSET;
            ecmd.autoneg = 0x01;
            ecmd.advertising = ecmd.supported;
            ecmd.duplex = duplex;

            ioctl(fd, SIOCETHTOOL, &ifr);
            close(fd);
            return returnErrorMsg((ERRORCODE)(ERROR_ETHERNET_DAPI_SETUP_SPEED | ERROR_ETHERNET_DAPI_SETUP_SPEED_TIMEOUT), nic.c_str(), NULL);
        }
        GetLinkStatus(nic, &link_up);
    }

    close(fd);

    return NO_ERROR_MSG;
}



ErrorMsg* INTEL_I210::GetMAC(string ethernet_name, string *mac_address)
{
    INT32  err, fd, i;
    struct ethtool_eeprom *eeprom;
    struct ifreq ifr;
    char   string_temp[3];

    errno = 0;

    if (ethernet_name.length() == 0)
    {
        return returnErrorMsg( (ERRORCODE)(ERROR_ETHERNET_DDAPI_GET_EEPROM | ERROR_ETHERNET_DDAPI_NO_SPECIFY_NIC), NULL, NULL);
    }

    fd = socket(AF_INET, SOCK_DGRAM, 0);
    if (fd < 0)
    {
        return returnErrorMsg( (ERRORCODE)(ERROR_ETHERNET_DDAPI_GET_EEPROM | ERROR_ETHERNET_DDAPI_SOCKET_OPEN_FAIL), NULL, NULL);
    }

    strncpy(ifr.ifr_name, ethernet_name.c_str(), sizeof(ifr.ifr_name));

    eeprom = (ethtool_eeprom *)calloc(1, sizeof(*eeprom) + 6);
    if (!eeprom)
    {
        free(eeprom);
        close(fd);
        return returnErrorMsg( (ERRORCODE)(ERROR_ETHERNET_DDAPI_GET_EEPROM | ERROR_ETHERNET_DDAPI_ALLOCAT_EEPROM_FAIL), NULL, NULL);
    }

    eeprom->cmd = ETHTOOL_GEEPROM;
    eeprom->len = 6;
    eeprom->offset = 0x00; //MAC store address in EEPROM

    ifr.ifr_data = (__caddr_t)eeprom;

    err = ioctl(fd, SIOCETHTOOL, &ifr);
    if (err < 0)
    {
        //perror("Cannot do get EEPROM");
        free(eeprom);
        close(fd);
        return returnErrorMsg( (ERRORCODE)(ERROR_ETHERNET_DDAPI_GET_EEPROM | ERROR_ETHERNET_DDAPI_IOCTL_FAIL), strerror(errno), NULL);
    }

    for (i = 0; i < eeprom->len; i++)
    {
        sprintf(string_temp, "%02X",eeprom->data[i]);

        *mac_address += string_temp;
        if ( i != 5 )
        {
            *mac_address += ":";
        }
    }

    free(eeprom);
    close(fd);

    return NO_ERROR_MSG;
}



ErrorMsg* INTEL_I210::SetMAC(string ethernet_name, string *mac_address)
{
    INT32  fd;
    struct ethtool_eeprom *eeprom;
    struct ifreq ifr;
    INT32  MACnumber;
    string vendor_ID;

    errno = 0;

    if (ethernet_name.length() == 0)
    {
        return returnErrorMsg( (ERRORCODE)(ERROR_ETHERNET_DDAPI_GET_EEPROM | ERROR_ETHERNET_DDAPI_NO_SPECIFY_NIC), NULL, NULL);
    }

    strncpy(ifr.ifr_name, ethernet_name.c_str(), sizeof(ifr.ifr_name));

    // Open control socket.
    fd = socket(AF_INET, SOCK_DGRAM, 0);
    if (fd < 0)
    {
        return returnErrorMsg( (ERRORCODE)(ERROR_ETHERNET_DDAPI_GET_EEPROM | ERROR_ETHERNET_DDAPI_SOCKET_OPEN_FAIL), NULL, NULL);
    }

    eeprom = (ethtool_eeprom *)calloc(1, sizeof(*eeprom) + 6);
    if (!eeprom)
    {
        free(eeprom);
        close(fd);
        return returnErrorMsg( (ERRORCODE)(ERROR_ETHERNET_DDAPI_GET_EEPROM | ERROR_ETHERNET_DDAPI_ALLOCAT_EEPROM_FAIL), NULL, NULL);
    }

    //Get I210 PCI vender ID
    GetPCIDeviceID(ethernet_name, &vendor_ID);
    //printf("vendor_ID: %s\n", vendor_ID.c_str());

    eeprom->cmd = ETHTOOL_SEEPROM;
    eeprom->len = 6;
    eeprom->offset = 0;
    if (vendor_ID == "0x1533")
        eeprom->magic = 0x15338086;
    else if (vendor_ID == "0x1537")
        eeprom->magic = 0x15378086;

    sscanf((mac_address[0].c_str()), "%x", &MACnumber);
    eeprom->data[0] = MACnumber;
    sscanf((mac_address[1].c_str()), "%x", &MACnumber);
    eeprom->data[1] = MACnumber;
    sscanf((mac_address[2].c_str()), "%x", &MACnumber);
    eeprom->data[2] = MACnumber;
    sscanf((mac_address[3].c_str()), "%x", &MACnumber);
    eeprom->data[3] = MACnumber;
    sscanf((mac_address[4].c_str()), "%x", &MACnumber);
    eeprom->data[4] = MACnumber;
    sscanf((mac_address[5].c_str()), "%x", &MACnumber);
    eeprom->data[5] = MACnumber;

    ifr.ifr_data = (__caddr_t)eeprom;

    if (ioctl(fd, SIOCETHTOOL, &ifr) < 0)
    {
        free(eeprom);
        close(fd);
        return returnErrorMsg( (ERRORCODE)(ERROR_ETHERNET_DDAPI_GET_EEPROM | ERROR_ETHERNET_DDAPI_IOCTL_FAIL), strerror(errno), NULL);
    }

    close(fd);

    return NO_ERROR_MSG;
}



ErrorMsg* INTEL_I210::GetEEPROM(string ethernet_name, INT32 StartAddress, INT32 Length, INT32 *data)
{
    struct ethtool_eeprom *eeprom;
    struct ifreq          ifr;
           INT32          err, fd, i;

    errno = 0;

    if (ethernet_name.length() == 0)
    {
        return returnErrorMsg( (ERRORCODE)(ERROR_ETHERNET_DDAPI_GET_EEPROM | ERROR_ETHERNET_DDAPI_NO_SPECIFY_NIC), NULL, NULL);
    }

    fd = socket(AF_INET, SOCK_DGRAM, 0);
    if (fd < 0)
    {
        return returnErrorMsg( (ERRORCODE)(ERROR_ETHERNET_DDAPI_GET_EEPROM | ERROR_ETHERNET_DDAPI_SOCKET_OPEN_FAIL), NULL, NULL);
    }

    strncpy(ifr.ifr_name, ethernet_name.c_str(), sizeof(ifr.ifr_name));

    eeprom = (ethtool_eeprom *)calloc(1, sizeof(*eeprom) + Length);
    if (!eeprom)
    {
        free(eeprom);
        close(fd);
        return returnErrorMsg( (ERRORCODE)(ERROR_ETHERNET_DDAPI_GET_EEPROM | ERROR_ETHERNET_DDAPI_ALLOCAT_EEPROM_FAIL), NULL, NULL);
    }

    eeprom->cmd    = ETHTOOL_GEEPROM;
    eeprom->len    = Length;
    eeprom->offset = StartAddress;

    ifr.ifr_data = (__caddr_t)eeprom;

    err = ioctl(fd, SIOCETHTOOL, &ifr);
    if (err < 0)
    {
        //perror("Cannot do get EEPROM");
        free(eeprom);
        close(fd);
        return returnErrorMsg( (ERRORCODE)(ERROR_ETHERNET_DDAPI_GET_EEPROM | ERROR_ETHERNET_DDAPI_IOCTL_FAIL), strerror(errno), NULL);
    }

    for (i = 0; i < eeprom->len; i++)
    {
        data[i] = eeprom->data[i];
    }

    free(eeprom);
    close(fd);

    return NO_ERROR_MSG;
}



ErrorMsg* INTEL_I210::SetEEPROM(string ethernet_name, INT32 StartAddress, INT32 *data, INT32 Length)
{
    struct ethtool_eeprom *eeprom;
    struct ifreq          ifr;
           INT32          err, fd, i;

    errno = 0;

    if (ethernet_name.length() == 0)
    {
        return returnErrorMsg( (ERRORCODE)(ERROR_ETHERNET_DDAPI_GET_EEPROM | ERROR_ETHERNET_DDAPI_NO_SPECIFY_NIC), NULL, NULL);
    }

    fd = socket(AF_INET, SOCK_DGRAM, 0);
    if (fd < 0)
    {
        return returnErrorMsg( (ERRORCODE)(ERROR_ETHERNET_DDAPI_GET_EEPROM | ERROR_ETHERNET_DDAPI_SOCKET_OPEN_FAIL), NULL, NULL);
    }

    strncpy(ifr.ifr_name, ethernet_name.c_str(), sizeof(ifr.ifr_name));

    eeprom = (ethtool_eeprom *)calloc(1, sizeof(*eeprom) + Length);
    if (!eeprom)
    {
        free(eeprom);
        close(fd);
        return returnErrorMsg( (ERRORCODE)(ERROR_ETHERNET_DDAPI_GET_EEPROM | ERROR_ETHERNET_DDAPI_ALLOCAT_EEPROM_FAIL), NULL, NULL);
    }

    eeprom->cmd     = ETHTOOL_SEEPROM;
    eeprom->len     = Length;
    eeprom->offset  = StartAddress;
    eeprom->magic   = 0x15338086; //Intel device_id:vend_id
    for (i = 0; i < Length ; i++)
    {
        eeprom->data[i] = data[i];
    }

    ifr.ifr_data = (__caddr_t)eeprom;

    err = ioctl(fd, SIOCETHTOOL, &ifr);
    if (err < 0)
    {
        //perror("Cannot do get EEPROM");
        free(eeprom);
        close(fd);
        return returnErrorMsg( (ERRORCODE)(ERROR_ETHERNET_DDAPI_GET_EEPROM | ERROR_ETHERNET_DDAPI_IOCTL_FAIL), strerror(errno), NULL);
    }

    free(eeprom);
    close(fd);

    return NO_ERROR_MSG;
}



ErrorMsg* INTEL_I210::Program(string ethernet_name, string filePath)
{
    int     domain, bus, device, func, result = 0;
    string  command = "eeupdate64e";
    char    parameter[512];

    //Get I210 PCI BDF
    m_errorCode = GetPCIBusDeviceFunction(ethernet_name, &domain, &bus, &device, &func);
    if (m_errorCode != NO_ERROR_MSG)
    {
        return returnErrorMsg((ERRORCODE)(ERROR_ETHERNET_DAPI_SETUP_SPEED), "Get PCI BDF fail", m_errorCode);
    }
    //printf("bus:%x  device:%x  func:%x\n",bus, device, func);

    sprintf(parameter, " /bus=%d /dev=%d /fun=%d /D %s >> /dev/null",bus, device, func, filePath.c_str());


    command.append(parameter);
    //printf("%s\n", command.c_str());
    result = system(command.c_str());
    if (result != 0)
    {
        return returnErrorMsg((ERRORCODE)(ERROR_ETHERNET_DAPI_SETUP_SPEED), "I210 program EEPROM fail", m_errorCode);
    }

    return NO_ERROR_MSG;
}



ErrorMsg* INTEL_I210::ReadSFP(string ethernet_name, INT32 DeviceAddress, INT32 StartAddress, INT32 Length, INT32 *data)
{
    return returnErrorMsg(ERROR_ETHERNET_DDAPI_SENDIOCTL, "Intel I210 not support", NULL);
}



ErrorMsg* INTEL_I210::getLEDName(string nic, list<string> *LED)
{

    LED->push_back(nic + "_led0");
    LED->push_back(nic + "_led1");
    LED->push_back(nic + "_led2");

    return NO_ERROR_MSG;
}



ErrorMsg* INTEL_I210::setLEDStatus(string ethernet_name, string in_strLEDName, string in_strStatus)
{
    unsigned int *addr, *LED_point;
    int fd;


    fd = open(Get_Node().c_str(), O_RDWR | O_SYNC);
    if (fd < 0)
    {
        //printf("%s file open error!\n",Get_Node().c_str());
        return returnErrorMsg( (ERRORCODE)(ERROR_ETHERNET_DDAPI_SET_EXTERNAL_LOOPBACK | ERROR_ETHERNET_DDAPI_OPEN_NODE_FAIL), NULL, NULL);
    }

    addr = (unsigned int *)mmap(NULL, 0x1000, PROT_READ|PROT_WRITE, MAP_SHARED, fd, 0x00);
    LED_point = addr + (LED_CONTROL_REGISTER / 4);

    if (in_strStatus == "on")
    {
        if (in_strLEDName == "led0")
        {
            *LED_point = (*LED_point & 0xFFFF70) | 0x00000E;
        }
        else if (in_strLEDName == "led1")
        {
            *LED_point = (*LED_point & 0xFF70FF) | 0x000E00;
        }
        else if (in_strLEDName == "led2")
        {
            *LED_point = (*LED_point & 0x70FFFF) | 0x0E0000;
        }
        else
        {
            munmap(addr, 0x1000);
            close(fd);
            return returnErrorMsg( (ERRORCODE)(ERROR_ETHERNET_DDAPI_OPEN_NODE_FAIL), "Undefine LED index" , NULL);
        }
    }
    else if (in_strStatus == "off")
    {
        if (in_strLEDName == "led0")
        {
            *LED_point = *LED_point | 0x00000F;
        }
        else if (in_strLEDName == "led1")
        {
            *LED_point = *LED_point | 0x000F00;
        }
        else if (in_strLEDName == "led2")
        {
            *LED_point = *LED_point | 0x0F0000;
        }
        else
        {
            munmap(addr, 0x1000);
            close(fd);
            return returnErrorMsg( (ERRORCODE)(ERROR_ETHERNET_DDAPI_OPEN_NODE_FAIL), "Undefine LED index" , NULL);
        }
    }
    else
    {
         munmap(addr, 0x1000);
         close(fd);
         return returnErrorMsg( (ERRORCODE)(ERROR_ETHERNET_DDAPI_OPEN_NODE_FAIL), "Undefine LED status" , NULL);
    }

    munmap(addr, 0x1000);
    close(fd);

    return NO_ERROR_MSG;
}



ErrorMsg* INTEL_I210::getLEDStatus(string ethernet_name, string in_strLEDName, string *out_strpStatus)
{
    unsigned int *addr, *LED_point;
    int fd;


    fd = open(Get_Node().c_str(), O_RDWR | O_SYNC);
    if (fd < 0)
    {
        //printf("%s file open error!\n",Get_Node().c_str());
        return returnErrorMsg( (ERRORCODE)(ERROR_ETHERNET_DDAPI_SET_EXTERNAL_LOOPBACK | ERROR_ETHERNET_DDAPI_OPEN_NODE_FAIL), NULL, NULL);
    }

    addr = (unsigned int *)mmap(NULL, 0x1000, PROT_READ|PROT_WRITE, MAP_SHARED, fd, 0x00);
    LED_point = addr + (LED_CONTROL_REGISTER / 4);

    if (in_strLEDName == "led0")
    {
        if ((*LED_point & 0x0F) == 0x0E)
        {
            *out_strpStatus = "on";
        }
        else if ((*LED_point & 0x0F) == 0x0F)
        {
            *out_strpStatus = "off";
        }
        else
        {
            munmap(addr, 0x1000);
            close(fd);
            return returnErrorMsg( (ERRORCODE)(ERROR_ETHERNET_DDAPI_OPEN_NODE_FAIL), "Undefine LED status" , NULL);
        }
    }
    else if (in_strLEDName == "led1")
    {
        if ((*LED_point & 0x0F00) == 0x0E00)
        {
            *out_strpStatus = "on";
        }
        else if ((*LED_point & 0x0F00) == 0x0F00)
        {
            *out_strpStatus = "off";
        }
        else
        {
            munmap(addr, 0x1000);
            close(fd);
            return returnErrorMsg( (ERRORCODE)(ERROR_ETHERNET_DDAPI_OPEN_NODE_FAIL), "Undefine LED status" , NULL);
        }
    }
    else if (in_strLEDName == "led2")
    {
        if ((*LED_point & 0x0F0000) == 0x0E0000)
        {
            *out_strpStatus = "on";
        }
        else if ((*LED_point & 0x0F0000) == 0x0F0000)
        {
            *out_strpStatus = "off";
        }
        else
        {
            munmap(addr, 0x1000);
            close(fd);
            return returnErrorMsg( (ERRORCODE)(ERROR_ETHERNET_DDAPI_OPEN_NODE_FAIL), "Undefine LED status" , NULL);
        }
    }
    else
    {
        munmap(addr, 0x1000);
        close(fd);
        return returnErrorMsg( (ERRORCODE)(ERROR_ETHERNET_DDAPI_OPEN_NODE_FAIL), "Undefine LED index" , NULL);
    }


    munmap(addr, 0x1000);
    close(fd);

    return NO_ERROR_MSG;
}


