/// @file       FPGADDAPI.hpp
/// @brief      FPGADDAPI class
/// @details    Provide low layer diagnostic and utility APIs for FPGA module
/// @author     KenHu
/// @copyright  Copyright 2015-2016 Hon Hai Technology Group. All Rights Reserved.

#ifndef FPGA_DDAPI_HPP
#define FPGA_DDAPI_HPP

#include "BaseClassDDAPI.hpp"
#include "MemoryDDAPI.hpp"

#define FPGA_DEVICE_NAME    "fpga0"

#define READ_WRITE  0
#define READ_ONLY   1
#define WRITE_ONLY  2

typedef struct {
    UINT32  u32Offset;
    UINT8   u8Width;
    INT32   s32Attribute;
} FPGARegInfo;

#define FPGA_SPAD_REG_OFFSET_START  0x8000
#define FPGA_SPAD_REG_TOTAL_SIZE    0x4000

/// @class      FPGADDAPI FPGADDAPI.hpp "FPGADDAPI.hpp"
/// @brief      Provide low layer diagnostic and utility APIs for FPGA module
/// @param      m_u32BaseAddress is a UINT32 value for FPGA's memory mapped base address
/// @param      m_pMemoryDDAPI is a pointer to Memory DDAPI class
class FPGADDAPI : public BaseClassDDAPI
{
    public:
                            FPGADDAPI();
        virtual             ~FPGADDAPI(void);

        ErrorMsg*           getDeviceList(list<string> *out_pList);
        ErrorMsg*           getSPADInfo(string in_strDeviceName,
                                        UINT32 *out_u32pStartOffset,
                                        UINT32 *out_u32pTotalSize);
        UINT32              getRegSize(UINT32 in_u32Offset);

        ErrorMsg*           read(string in_strDeviceName,
                                 UINT32 in_u32Offset,
                                 UINT8  *out_u8pValue);
        ErrorMsg*           readSPAD(string in_strDeviceName,
                                     UINT32 in_u32Offset,
                                     UINT8  *out_u8Value);

        ErrorMsg*           write(string in_strDeviceName,
                                  UINT32 in_u32Offset,
                                  UINT8  in_u8Value);
        ErrorMsg*           writeSPAD(string in_strDeviceName,
                                      UINT32 in_u32Offset,
                                      UINT8  in_u8Value);
    protected:

    private:
        UINT32              getBaseAddress(void);
        FPGARegInfo*        getRegInfo(UINT32 in_u32Offset);

        UINT32              m_u32BaseAddress;
        MemoryDDAPI*        m_pMemoryDDAPI;
};

#endif
