#include "GPIODDAPI.hpp"


GPIODDAPI::GPIODDAPI(STRING className) : BaseClassDDAPI(className)
{
    //printf("%s\n", __func__);
    FILE  *pFile;
    char  line[500], teg[20];
    int   teg_number, string_length;
    GPIO_TYPE   tempTag = GPIO_MAX_TYPE_NUMBER;


    list<string> TempGPIODeviceBusPortList;
    TempGPIODeviceBusPortList.clear();

    for(UINT32 index = 0; index <  MAX_NUMBER_OF_GPIO_DRIVERS; index++)
        m_GPIODriverList[index] = (CHAR)NULL;


    pFile=fopen (PROFILE_PATH_NAME,"r");
    if (pFile == NULL)
    {
        //printf("Error opening file\n");
        addErrorMsg((ERRORCODE)(ERROR_GPIO_DDAPI_CONSTRUCTOR | ERROR_GPIO_DDAPI_OPEN_PROFILE_FAIL), NULL);
    	//addErrorMsg(ERROR_GPIO_DDAPI_CONSTRUCTOR, NULL);

        return;
    }

    //Read each line of I2C_profile.ini
    while (fgets(line, 500, pFile) != NULL)
    {
        if (line[0] == '#')
            continue;

        teg_number = sscanf(line, "[%[^]]", teg); //paser the teg:BUS, Topology, Protocol
        if ( (strlen(line) == 1 && line[0] == '\n') || (teg_number != 0))
        {
            continue;
        }
        //printf("Teg(%d):%s line:%s",teg_number, teg, line);

        //Remove the '\n' at the end of line
        string_length = strlen(line);
        if (line[string_length-1] == '\n')
        {
            line[string_length-1] = '\0';
        }
        //Start parser by teg
        if ( strcmp(teg,"FPGA") == 0 )
        {
            //Initial_I2C_BUS(line);
            //printf("!!!!\n");
            m_GPIODriverList[GPIO_FPGA_DRIVER] = new GPIO_FPGADriver("GPIO_FPGA");
            tempTag = GPIO_FPGA_DRIVER;
            //printf("%d\n", GPIO_FPGA_DRIVER);
            memset(teg, 0, sizeof(teg));
        }

        if ( strcmp(teg,"PCH") == 0 )
        {
            //Initial_I2C_BUS(line);
            //printf("!!!!\n");
            m_GPIODriverList[GPIO_PCH_DRIVER] = new GPIO_PCHDriver("GPIO_PCH");
            tempTag = GPIO_PCH_DRIVER;
            //printf("%d\n", GPIO_MIDPLANE_RIVER);

            memset(teg, 0, sizeof(teg));
        }

        switch(tempTag)
        {
            case GPIO_FPGA_DRIVER:
                m_GPIODriverList[GPIO_FPGA_DRIVER]->readINIFile(line);
                break;
            case GPIO_PCH_DRIVER:
            	m_GPIODriverList[GPIO_PCH_DRIVER]->readINIFile(line);
                break;
            case GPIO_MAX_TYPE_NUMBER:
                break;
            default:
                break;
        }

        //
/*
        if ( strcmp(teg,"Topology") == 0 )
        {
            Initial_I2C_device(line, I2C_device_count);
            I2C_device_count++;
        }

        if ( strcmp(teg,"Protocol") == 0 )
        {
            Initial_I2C_Protocol(line);
        }
*/
    }
    fclose(pFile);


    //memset(GPIOConfig, "", sizeof(GPIOConfig));
/*
    for(UINT32 index = 0; index < MAX_GPIO_DEVICE_NO; index++)
    {
        GPIOConfig[index].LogicalID          = GPIO_ID_UNDEFINED;
        GPIOConfig[index].WellKnownName[0]   = (CHAR)NULL;
        GPIOConfig[index].DeviceNode[0]      = (CHAR)NULL;
        GPIOConfig[index].BusPortAddress[0]  = (CHAR)NULL;
    }
    for(UINT32 index = 0; index < MAX_DATA_ELEMENT_NUMBER; index++)
    {
        GPIOElementMap[index].DateElement  = MAX_DATA_ELEMENT_NUMBER;
        GPIOElementMap[index].FileName[0]  = (CHAR)NULL;
    }
    */

/*
    for(UINT32 index = 0; index < MAX_DATA_ELEMENT_NUMBER; index++)
    {
        GPIO_DATA_ELEMENT eCurrent = GPIO_DATA_ELEMENT(index);
        GPIOElementMap[eCurrent].DateElement = GPIO_DATA_ALL_INFO;
        switch(eCurrent)
        {
            case GPIO_DATA_ALL_INFO:
                GPIOElementMap[eCurrent].FileName[0] = (CHAR)NULL;
                break;
            case GPIO_DATA_NODE_INFO:
                GPIOElementMap[eCurrent].FileName[0]= (CHAR)NULL;
                break;
            case GPIO_DATA_SN_INFO:
                sprintf(GPIOElementMap[eCurrent].FileName,"%s", "serial");
                break;
            case GPIO_DATA_VENDOR_INFO:
                sprintf(GPIOElementMap[eCurrent].FileName,"%s", "manufacturer");
                break;
            case GPIO_DATA_CAPACITY_INFO:
                sprintf(GPIOElementMap[eCurrent].FileName,"%s", "size");
                break;
            case GPIO_DATA_LINK_INFO:
                sprintf(GPIOElementMap[eCurrent].FileName,"%s", "version");
                break;
            case GPIO_DATA_SPEED_INFO:
                sprintf(GPIOElementMap[eCurrent].FileName,"%s", "speed");
            case MAX_DATA_ELEMENT_NUMBER:
                break;
            default:
                break;

        }
    }
    */

}

GPIODDAPI::~GPIODDAPI()
{
    for(UINT32 Index = 0; Index <  MAX_NUMBER_OF_GPIO_DRIVERS; Index++)
    {
        if(m_GPIODriverList[Index] != NULL )
            delete m_GPIODriverList[Index];
    }

}

ErrorMsg *GPIODDAPI::getGPIOList(list<string> *pGPIOList)
{
    //printf("%s\n", __func__);

    ErrorMsg*   pErrorMsg = NO_ERROR_MSG;
    INT32 intIndex = 0;


    while ((intIndex < MAX_NUMBER_OF_GPIO_DRIVERS))
    {
    	//printf("!!!!\n");
    	if((m_GPIODriverList[intIndex] != 0))
    	{
    		pErrorMsg = m_GPIODriverList[intIndex]->getGPIOList(pGPIOList);

    		if (pErrorMsg != NO_ERROR_MSG)
    		{
    			// return Error
    			return returnErrorMsg(ERROR_GPIO_DDAPI_GET_GPIO_LIST, NULL, pErrorMsg);
    		}
    	}
        intIndex++;
    }
    return NO_ERROR_MSG;
}



ErrorMsg* GPIODDAPI::getLogicalIdByName(STRING WellKnownName, UINT32 *LogicalID)
{
    printf("%s\n", __func__);

    //ErrorMsg* pErrorMsg = NULL;

    CHAR tempName[(GPIO_DEVICE_NAME_LENGTH + 1)];

    // we need upper-case names, but don't change the callers input string
    strcpy(tempName, WellKnownName);
    ConvertStringToUpperCase(tempName);

    /*
    for (UINT32 GPIOIndex = 0; GPIOIndex < MAX_GPIO_DEVICE_NO; GPIOIndex++ )
    {
        if (GPIOConfig[GPIOIndex].WellKnownName == (CHAR)NULL)
            continue;
        if (strcmp(GPIOConfig[GPIOIndex].WellKnownName, tempName) == 0)
        {
            *LogicalID = GPIOConfig[GPIOIndex].LogicalID;
            return NO_ERROR_MSG;
            break;
        }
    }
    */

    //return returnErrorMsg((ERRORCODE)(ERRIR_GPIO_DDAPI_GET_LOGICAL_ID_BY_NAME | ERROR_GPIO_DDAPI_DEVICE_NAME_NOT_FOUND), NULL, pErrorMsg);
    return NO_ERROR_MSG;
}



ErrorMsg*   GPIODDAPI::GetWellKnownNameByIndex(INT32 index, CHAR* gpioName, INT32 stringLength)
{
    printf("%s\n", __PRETTY_FUNCTION__);


    return NO_ERROR_MSG;
}



ErrorMsg*   GPIODDAPI::readBit(string gpioName, BOOL* readValue)
{
    //printf("%s\n", __PRETTY_FUNCTION__);

    ErrorMsg*   pErrorMsg = NO_ERROR_MSG;

    UINT32 Index = 0;

    // display a debug message if requested
/*
    sprintf(m_tempString, "ReadBit function for %s\n", gpioName);
    displayDebugString(DEBUG_LEVEL_5, m_tempString);
*/
    //m_errorCode = ERROR_GPIO_DRIVER_UNKNOWN_NAME_REQUEST;
    while (Index < MAX_NUMBER_OF_GPIO_DRIVERS)
    {
    	if(m_GPIODriverList[Index] != 0)
    	{
			if( (pErrorMsg=m_GPIODriverList[Index]->isMyGPIO(gpioName)) == NO_ERROR_MSG)
			{
				pErrorMsg = m_GPIODriverList[Index]->readBit(gpioName, readValue);
				if(pErrorMsg != NO_ERROR_MSG)
				{
					return returnErrorMsg(ERROR_GPIO_DDAPI_GPIO_READ_BIT, NULL, pErrorMsg);
				}
				break;
			}
			else
			{
				return returnErrorMsg(ERROR_GPIO_DDAPI_GPIO_READ_BIT, NULL, pErrorMsg);
			}
    	}
        Index++;
    }

    return pErrorMsg;
}



ErrorMsg*   GPIODDAPI::writeBit (string  gpioName, BOOL Status)
{
    //printf("%s\n", __PRETTY_FUNCTION__);

    ErrorMsg*   ErrorMsg = NO_ERROR_MSG;

    UINT32 Index = 0;

    while (Index < MAX_NUMBER_OF_GPIO_DRIVERS)
    {
    	if (m_GPIODriverList[Index] != 0)
    	{
			if ((ErrorMsg=m_GPIODriverList[Index]->isMyGPIO(gpioName)) == NO_ERROR_MSG)
			{
				// display a debug message if requested
	//            sprintf(m_tempString, "ReadBit \"%s\" GPIO in %s\n", gpioName, m_gpioDriverList[index]->getClassName());
	//            displayDebugString(DEBUG_LEVEL_5, m_tempString);

				ErrorMsg = m_GPIODriverList[Index]->writeBit(gpioName, Status);
				if( ErrorMsg != NO_ERROR_MSG)
				{
					return returnErrorMsg(ERROR_GPIO_DDAPI_GPIO_SET_BIT, NULL, ErrorMsg);
				}
				break;
			}
			else
			{
				return returnErrorMsg(ERROR_GPIO_DDAPI_GPIO_SET_BIT, NULL, ErrorMsg);
			}
    	}
        Index++;
    }

    return ErrorMsg;
}






