#ifndef I2C_DDAPI_H
#define I2C_DDAPI_H

#include "BaseClassDDAPI.hpp"
#include "I2CDevice.hpp"
#include "I2CSwitch.hpp"
#include "I2CBus.hpp"
#include <list>
#include "PCIDDAPI.hpp"
#include "I2CTemperatureBaseClass.hpp"
#include "MAX6657.hpp"
#include "DIMMTemperature.hpp"

#define DEVICE_NUMBER  200
#define MASTER_NUMBER  20
#define ROUTE_NUMBER   20

struct I2C_RoutingTable
    {
     UINT32      switchLogicalID;
     UINT32      logicalOutputBusNumber;
    };
typedef struct I2C_RoutingTable I2C_ROUTING_TABLE_DATA;



enum TEMPERATURE_OPERATION_ID
{
    CURRENT_TEMPERATURE,
    THRESHOLD_HIGH,
    THRESHOLD_LOW,
};



class I2CDDAPI : public BaseClassDDAPI
{
public:
                     I2CDDAPI(STRING className);
                    ~I2CDDAPI();

         ErrorMsg  * read(string well_know_name, int StartRegister, int EndRegister, int *data, int *count);
         ErrorMsg  * write(string well_know_name, int StartRegister, int *data, int count);
         ErrorMsg  * write(string well_know_name, int StartRegister, int *data, int count, int *numberRealWrite);
         ErrorMsg  * read(int bus, int address, int StartRegister, int EndRegister, int *data, int *count);
         ErrorMsg  * write(int bus, int address, int StartRegister, int *data, int count);
         ErrorMsg  * write(int bus, int address, int StartRegister, int *data, int count, int *numberRealWrite);
         I2CDevice * get_device_by_well_known_name(string well_know_name);
         I2CDevice * get_device_by_bus_address(int bus, int address);
         ErrorMsg  * get_i2c_devices_list(list<string> *list);
         ErrorMsg  * get_i2c_devices_list(I2CDevice *list[]);
         ErrorMsg  * mux_test(string well_know_name);
         ErrorMsg  * get_i2c_device_path(string well_know_name, string *route_path);
         ErrorMsg  * stuckPrevention(int busID);
         ErrorMsg  * getTemperatureDataByID(string well_know_name, TEMPERATURE_OPERATION_ID ID, string *result);
         ErrorMsg  * setTemperatureDataByID(string well_know_name, TEMPERATURE_OPERATION_ID ID, string data);
         ErrorMsg  * DisplayI2cDevicesListByType(I2CDevice *i2c_list[], I2C_DEVICE_TYPE type);

protected:


private:
         int         Is_physical_bus(int bus);
         int         Find_physical_master(int logical_bus);
         int         Route_path(int logical_bus);
         ErrorMsg *  Initial_I2C_BUS(char *line);
         ErrorMsg *  Initial_I2C_device(char *line, int device_count);
         ErrorMsg *  Initial_I2C_Protocol(char *line);

         ErrorMsg    *m_errorCode;
         I2CDevice   *m_I2C_device_table[DEVICE_NUMBER];
         I2CBus      *m_I2C_bus_table[MASTER_NUMBER];
         int         m_mux_counter;
         I2C_ROUTING_TABLE_DATA   m_route_table[ROUTE_NUMBER];
};

#endif

