#include <sys/time.h>
#include <sys/mman.h>
#include <csignal>

#include "NTBDDAPI.hpp"

NTBDDAPI::NTBDDAPI(STRING className) : BaseClassDDAPI(className)
{
    m_pPCIDriverAPI = new PCIDDAPI("PCIDDAPI");
}

NTBDDAPI::~NTBDDAPI()
{
    if(m_pPCIDriverAPI != NULL)
    {
        delete m_pPCIDriverAPI;
        m_pPCIDriverAPI = NULL; 
    }
}

#if 0
bool
NTBDDAPI::isNTB(UINT16  in_u16VendorID,
                UINT16  in_u16DeviceID)
{
   int s32Index = 0;

   while(ntbDeviceTable[s32Index].u16VendorID != 0xFFFF)
   {
       if(ntbDeviceTable[s32Index].u16VendorID == in_u16VendorID && 
          ntbDeviceTable[s32Index].u16DeviceID == in_u16DeviceID)
       {
           return TRUE;
       }
       s32Index++;
   }

    return FALSE;   
}
#endif

/*
void 
NTBDDAPI::signalHandler(int signum)
{
    free(m_u64pBufferSend);
    free(m_u64pBufferRev);
    free(m_u64pBufferInit);
}
*/

ErrorMsg*
NTBDDAPI::searchDevice(vector<NTB_INFO>*  out_pNtbList)
{
    char        ntbName[8];
    bool        isNTB;
    UINT16      u16DeviceCount = 0;
    NTB_INFO    ntbInfo;

    vector<PCIInformantion>             pciList;
    vector<PCIInformantion>::iterator   it;

    m_pPCIDriverAPI->getDeviceList(&pciList, NOT_SORT_PCI_DEVICES); //not sort
    for(it = pciList.begin(); it != pciList.end(); ++it)
    {
        isNTB = m_pPCIDriverAPI->isNtbDevice(it->u16VendorID, it->u16DeviceID);
        if(isNTB)
        {
            ntbInfo.u16LogicalID = u16DeviceCount;
            
            sprintf(ntbName, "NTB%d", u16DeviceCount);
            ntbInfo.ntbName = ntbName;

            ntbInfo.pciInfo = &(*it);
            
            out_pNtbList->push_back(ntbInfo);

            u16DeviceCount++;
        }
    }

    if(u16DeviceCount == 0)
    {
        m_pErrorMsg = returnErrorMsg(ERROR_NTB_DDAPI_SEARCH_DEVICE, "no device is found", NULL);
    }
    return m_pErrorMsg;
}

//in_u32Size: user-defined buffer size
//if 0, use default size which is caught from /tmp/ntbInfo 
ErrorMsg*
NTBDDAPI::getWindowInfo(UINT32  in_u32Size)
{
    FILE*   pFile;
    char*   tempPtr;
    char    line[NTB_MAX_LINE];
    void*   UserAddrRead;
    void*   UserAddrWrite;
    UINT32  u32Size;
    UINT32  u32RemoteSize;
    UINT64  u64ReadWindow;
    UINT64  u64WriteWindow;
    UINT64  u64PageSize;
    UINT64  u64PageMask;

    pFile = fopen(NTB_INFO_FILE, "r");
    if(!pFile)
    {
        return returnErrorMsg(ERROR_NTB_DDAPI_GET_WINDOW_INFO, "can't find ntbInfo file", NULL);
    }

    while(fgets(line, sizeof(line), pFile) != NULL)
    {
        tempPtr = strtok(line, "=");
        if(strcmp(tempPtr, "ntbWriteWindow") == 0)
        {
            tempPtr = strtok(NULL, "=");
            u64WriteWindow = strtol(tempPtr, NULL, 16);
        }
        if(strcmp(tempPtr, "ntbWindowSize") == 0)
        {
            tempPtr = strtok(NULL, "=");
            u32Size = strtol(tempPtr, NULL, 10);
        }
        if(strcmp(tempPtr, "ntbRemoteBuffSize") == 0)
        {
            tempPtr = strtok(NULL, "=");
            u32RemoteSize = strtol(tempPtr, NULL, 10);
        }
        if(strcmp(tempPtr, "ntbReadWindow") == 0)
        {
            tempPtr = strtok(NULL, "=");
            u64ReadWindow = strtol(tempPtr, NULL, 16);
        }
    }
    fclose(pFile);

    //mark for test on sandhawk !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    //check window size 
    if(u32Size != u32RemoteSize) 
    {
        return returnErrorMsg(ERROR_NTB_DDAPI_GET_WINDOW_INFO, "window size mismatch", NULL);
    }

    if(in_u32Size != 0)
    {
        if(in_u32Size > u32Size)
        {
            return returnErrorMsg(ERROR_NTB_DDAPI_GET_WINDOW_INFO, "invalid buffer size", NULL);
        }
        u32Size = in_u32Size;
    }

    m_fd = open(NTB_DRIVER_NAME, O_RDWR|O_SYNC);
    //m_fd = open("/dev/mem", O_RDWR|O_SYNC);
    if(m_fd < 0)
    {
        return returnErrorMsg(ERROR_NTB_DDAPI_GET_WINDOW_INFO, "fail to open driver", NULL);
    }

    u64PageSize = 1 << 20;
    u64PageMask = u64PageSize - 1;

    UserAddrWrite = mmap64(0, 
                         u32Size, 
                         PROT_READ|PROT_WRITE, 
                         MAP_SHARED, 
                         m_fd, 
                         (u64WriteWindow & ~u64PageMask));
    if(UserAddrWrite == MAP_FAILED)
    {
        close(m_fd);
        return returnErrorMsg(ERROR_NTB_DDAPI_GET_WINDOW_INFO, "write window mmap fail", NULL); 
    }
    
    UserAddrRead = mmap64(0, 
                        u32Size,
                        PROT_READ|PROT_WRITE,
                        MAP_SHARED,
                        m_fd,
                        (u64ReadWindow & ~u64PageMask));
    if(UserAddrRead == MAP_FAILED)
    {
        close(m_fd);
        munmap(UserAddrWrite, u32Size);
        return returnErrorMsg(ERROR_NTB_DDAPI_GET_WINDOW_INFO, "read window mmap fail", NULL); 
    }

    m_windowInfo.u32Size            = u32Size;
    m_windowInfo.u32SizeDW          = u32Size/8;
    m_windowInfo.u64pUserAddrRead   = (UINT64*)((UINT64)UserAddrRead + (u64ReadWindow & u64PageMask)); 
    m_windowInfo.u64pUserAddrWrite  = (UINT64*)((UINT64)UserAddrWrite + (u64WriteWindow & u64PageMask));

    return NO_ERROR_MSG;
}


ErrorMsg*
NTBDDAPI::stress(UINT64     in_u64PatternSend,
                 UINT64     in_u64PatternRev,
                 UINT64     in_u64Loop,
                 bool       in_bCheckData)
{
    ErrorMsg*   pErrorMsg;
    UINT32      u32SizeDW;
    UINT64      u64Loop;
    UINT64*     u64pWriteAddr;
    UINT64*     u64pReadAddr;
    UINT64      u64Data;

    pErrorMsg = setForIoTransfer(in_u64PatternSend, 0x646f6e65);
    if(pErrorMsg != NO_ERROR_MSG)
    {
        return returnErrorMsg(ERROR_NTB_DDAPI_STRESS, NULL, pErrorMsg);
    }

    u64Loop         = in_u64Loop;
    u32SizeDW       = m_windowInfo.u32SizeDW;
    u64pWriteAddr   = m_windowInfo.u64pUserAddrWrite;
    u64pReadAddr    = m_windowInfo.u64pUserAddrRead;

    //signal(SIGTERM, (sighandler_t)((void*)signalHandler));
    while(u64Loop > 0)
    {
        //printf("--------------------------------\n");//check 
        //send data 
        memcpy(u64pWriteAddr, m_u64pBufferSend, sizeof(UINT64)*u32SizeDW);

        if(in_bCheckData)
        {
            //receive and compare data 
            pErrorMsg = receiveData(m_u64pBufferSend, u64pReadAddr, u32SizeDW);
            if(pErrorMsg != NO_ERROR_MSG)
            {
                u64Data = *u64pReadAddr;
                printf("u64Data: 0x%llx\n", u64Data);//check 
                if(u64Data == 0x646f6e65)
                {
                    pErrorMsg = NO_ERROR_MSG;
                }
                break;
            }
            if(u64Loop == 1)
            {
                *u64pWriteAddr = 0x646f6e65; //done
            }
            else 
            {
                //reset read buffer
                memcpy(u64pReadAddr, m_u64pBufferInit, sizeof(UINT64)*u32SizeDW);
            }
        }
        
        u64Loop--;
    }

    //release resource
    releaseDataBlocks();
    releaseWindowResource();

    if(pErrorMsg != NO_ERROR_MSG)
    {
        return returnErrorMsg(ERROR_NTB_DDAPI_STRESS, NULL, pErrorMsg);
    }
    return NO_ERROR_MSG;
}


ErrorMsg*
NTBDDAPI::testPerformance(UINT64    in_u64PatternSend,
                          UINT64    in_u64BlockSize,
                          UINT64    in_u64Loop)
{
    ErrorMsg*   pErrorMsg;
    UINT32      u32SizeDW;
    UINT32      u32Index;
    UINT64      u64Loop;
    UINT64*     u64pWriteAddr;
    UINT64*     u64pReadAddr;
    UINT64      u64DataSize;
    UINT64      u64CheckData;
    double      dfCostTime;
    double      dfSpeed;

    struct timeval StartTime;
    struct timeval EndTime;
    struct timeval spendTime;

    //get window resources  
    pErrorMsg = getWindowInfo(in_u64BlockSize); 
    if(pErrorMsg != NO_ERROR_MSG)
    {
        return returnErrorMsg(ERROR_NTB_DDAPI_PERFORMANCE, NULL, pErrorMsg);
    }
    
    u64Loop         = in_u64Loop;
    u32SizeDW       = m_windowInfo.u32SizeDW;
    u64pWriteAddr   = m_windowInfo.u64pUserAddrWrite;
    u64pReadAddr    = m_windowInfo.u64pUserAddrRead;
    
    //create write buffer 
    m_u64pBufferSend = (UINT64*)malloc(sizeof(UINT64)*u32SizeDW);
    for(u32Index = 0; u32Index < u32SizeDW; u32Index++)
    {
        m_u64pBufferSend[u32Index] = in_u64PatternSend;
    }

    u64CheckData = 0x646f6e65;

    memcpy(u64pWriteAddr, m_u64pBufferSend, sizeof(UINT64)*u32SizeDW);
    pErrorMsg = receiveData(&u64CheckData, u64pReadAddr, 1);
    if(pErrorMsg != NO_ERROR_MSG)
    {
        //release resource
        free(m_u64pBufferSend);
        releaseWindowResource();
        return returnErrorMsg(ERROR_NTB_DDAPI_PERFORMANCE, NULL, pErrorMsg);
    }

    gettimeofday(&StartTime, NULL);

    //signal(SIGTERM, (sighandler_t)((void*)signalHandler));
    while(u64Loop > 0)
    {
        //send data 
        memcpy(u64pWriteAddr, m_u64pBufferSend, sizeof(UINT64)*u32SizeDW);
        u64Loop--;
    }

    gettimeofday(&EndTime, NULL);
    spendTime.tv_sec = EndTime.tv_sec - StartTime.tv_sec;
    spendTime.tv_usec = EndTime.tv_usec - StartTime.tv_usec;
    dfCostTime = (double)(spendTime.tv_sec)+ ((double)(spendTime.tv_usec)/1000000);
    //printf("costTime: %.8f s\n", dfCostTime);

    u64DataSize = m_windowInfo.u32Size * in_u64Loop;
    //printf("data size: 0x%llx\n", u64DataSize);
    dfSpeed = ((u64DataSize/dfCostTime)/SIZE_1G)*8;
    printf("speed: %.8f Gb/s\n", dfSpeed);

    //release resource
    free(m_u64pBufferSend);
    releaseWindowResource();

    if(pErrorMsg != NO_ERROR_MSG)
    {
        return returnErrorMsg(ERROR_NTB_DDAPI_PERFORMANCE, NULL, pErrorMsg);
    }
    return NO_ERROR_MSG;
}

ErrorMsg* 
NTBDDAPI::receiver(UINT64   in_u64PatternRev,
                   UINT64   in_u64Loop)
{
    ErrorMsg*   pErrorMsg;
    UINT32      u32SizeDW;
    UINT32      u32Index;
    UINT64      u64Loop;
    UINT64*     u64pWriteAddr;
    UINT64*     u64pReadAddr;

    //get window resources  
    pErrorMsg = getWindowInfo(0); 
    if(pErrorMsg != NO_ERROR_MSG)
    {
        return returnErrorMsg(ERROR_NTB_DDAPI_RECEIVER, NULL, pErrorMsg);
    }
    
    u64Loop         = in_u64Loop;
    u32SizeDW       = m_windowInfo.u32SizeDW;
    u64pWriteAddr   = m_windowInfo.u64pUserAddrWrite;
    u64pReadAddr    = m_windowInfo.u64pUserAddrRead;
    
    //create expected read block 
    m_u64pBufferRev = (UINT64*)malloc(sizeof(UINT64)*u32SizeDW);
    for(u32Index = 0; u32Index < u32SizeDW; u32Index++)
    {
        m_u64pBufferRev[u32Index] = in_u64PatternRev;
    }
    
    while(u64Loop > 0)
    {
        pErrorMsg = receiveData(m_u64pBufferRev, u64pReadAddr, u32SizeDW);
        if(pErrorMsg != NO_ERROR_MSG)
        {
            break;
        }
        *u64pWriteAddr = 0x646f6e65;
        u64Loop--;
    }

    //release resource
    free(m_u64pBufferRev);
    releaseWindowResource();

    if(pErrorMsg != NO_ERROR_MSG)
    {
        return returnErrorMsg(ERROR_NTB_DDAPI_RECEIVER, NULL, pErrorMsg);
    }
    return NO_ERROR_MSG;
}

ErrorMsg*
NTBDDAPI::server(string     in_name,
                 UINT64     in_u64PatternSend,
                 UINT64     in_u64PatternRev,
                 UINT32     in_u32Loop)
{
    ErrorMsg*   pErrorMsg;
    UINT32      u32SizeDW;
    UINT32      u32Loop;
    UINT64*     u64pWriteAddr;
    UINT64*     u64pReadAddr;
    UINT64      u64PatternSend;
    UINT64      u64PatternRev;
    UINT32      u32Index;

    pErrorMsg = setForIoTransfer(in_u64PatternSend, in_u64PatternRev);
    if(pErrorMsg != NO_ERROR_MSG)
    {
        return returnErrorMsg(ERROR_NTB_DDAPI_SERVER, NULL, pErrorMsg);
    }

    u32Loop         = in_u32Loop;
    u32SizeDW       = m_windowInfo.u32SizeDW;
    u64pWriteAddr   = m_windowInfo.u64pUserAddrWrite;
    u64pReadAddr    = m_windowInfo.u64pUserAddrRead;

    u64PatternSend = in_u64PatternSend;
    u64PatternRev = in_u64PatternRev;

    //signal(SIGTERM, (sighandler_t)((void*)signalHandler));
    while(u32Loop > 0)
    {
        //send data 
        memcpy(u64pWriteAddr, m_u64pBufferSend, sizeof(UINT64)*u32SizeDW);
  
        //receive and compare data 
        pErrorMsg = receiveData(m_u64pBufferRev, u64pReadAddr, u32SizeDW);
        if(pErrorMsg != NO_ERROR_MSG)
        {
            break;
        }
        
        u32Loop--;
        
        //create sending block 
        u64PatternSend = ~u64PatternSend;
        for(u32Index = 0; u32Index < u32SizeDW; u32Index++)
        {
            m_u64pBufferSend[u32Index] = u64PatternSend;
        }
        //create receive block 
        u64PatternRev = ~u64PatternRev;
        for(u32Index = 0; u32Index < u32SizeDW; u32Index++)
        {
            m_u64pBufferRev[u32Index] = u64PatternRev;
        }
    }

    //release resource
    releaseDataBlocks();
    releaseWindowResource();

    if(pErrorMsg != NO_ERROR_MSG)
    {
        return returnErrorMsg(ERROR_NTB_DDAPI_SERVER, NULL, pErrorMsg);
    }
    return NO_ERROR_MSG;
}


ErrorMsg*
NTBDDAPI::client(string     in_name,
                 UINT64     in_u64PatternRev,
                 UINT64     in_u64PatternSend,
                 UINT32     in_u32Loop)
{
    ErrorMsg*   pErrorMsg;
    UINT64*     u64pWriteAddr;
    UINT64*     u64pReadAddr;
    UINT64      u64PatternSend;
    UINT64      u64PatternRev;
    UINT32      u32SizeDW;
    UINT32      u32Loop;
    UINT32      u32Index;

    pErrorMsg = setForIoTransfer(in_u64PatternSend, in_u64PatternRev);
    if(pErrorMsg != NO_ERROR_MSG)
    {
        return returnErrorMsg(ERROR_NTB_DDAPI_CLIENT, NULL, pErrorMsg);
    }
    
    u32Loop         = in_u32Loop;
    u32SizeDW       = m_windowInfo.u32SizeDW;
    u64pReadAddr    = m_windowInfo.u64pUserAddrRead;
    u64pWriteAddr   = m_windowInfo.u64pUserAddrWrite;
    
    u64PatternSend = in_u64PatternSend;
    u64PatternRev = in_u64PatternRev;

    while(u32Loop > 0)
    {
        pErrorMsg = receiveData(m_u64pBufferRev, u64pReadAddr, u32SizeDW);
        if(pErrorMsg != NO_ERROR_MSG)
        {
            break;
        }
            
        //send data 
        memcpy(u64pWriteAddr, m_u64pBufferSend, sizeof(UINT64)*u32SizeDW);
        
        u32Loop--;
        
        //create sending block 
        u64PatternSend = ~u64PatternSend;
        for(u32Index = 0; u32Index < u32SizeDW; u32Index++) 
        {
            m_u64pBufferSend[u32Index] = u64PatternSend;
        }
        //create receive block
        u64PatternRev = ~u64PatternRev;
        for(u32Index = 0; u32Index < u32SizeDW; u32Index++)
        {
            m_u64pBufferRev[u32Index] = u64PatternRev;
        }

    }
    
    //release resource
    releaseDataBlocks();
    releaseWindowResource();

    if(pErrorMsg != NO_ERROR_MSG)
    {
        return returnErrorMsg(ERROR_NTB_DDAPI_CLIENT, NULL, pErrorMsg);
    }
    return NO_ERROR_MSG;
}



ErrorMsg*
NTBDDAPI::createDataBlocks(UINT64   in_u64PatternSend,
                           UINT64   in_u64PatternRev)
{
    UINT32  u32Index;
    UINT32  u32SizeDW;

    if(m_windowInfo.u32SizeDW == 0)
    {
        return returnErrorMsg(ERROR_NTB_DDAPI_CREATE_BLOCK, "invalid buffer size", NULL);
    }

    u32SizeDW = m_windowInfo.u32SizeDW;

    m_u64pBufferSend = (UINT64*)malloc(sizeof(UINT64)*u32SizeDW);
    m_u64pBufferInit = (UINT64*)malloc(sizeof(UINT64)*u32SizeDW); 
    m_u64pBufferRev = (UINT64*)malloc(sizeof(UINT64)*u32SizeDW);

    //create init block
    for(u32Index = 0; u32Index < u32SizeDW; u32Index++)
    {
        m_u64pBufferInit[u32Index] = 0x0;
    }
    //create sending block  
    for(u32Index = 0; u32Index < u32SizeDW; u32Index++)
    {
        m_u64pBufferSend[u32Index] = in_u64PatternSend;
    }
    //create receive block
    for(u32Index = 0; u32Index < u32SizeDW; u32Index++)
    {
        m_u64pBufferRev[u32Index] = in_u64PatternRev;
    }

    return NO_ERROR_MSG;
}



ErrorMsg*
NTBDDAPI::releaseDataBlocks()
{
    UINT32  u32Index;
    UINT32  u32SizeDW;

    if(m_windowInfo.u32SizeDW == 0)
    {
        return returnErrorMsg(ERROR_NTB_DDAPI_RLS_BLOCK, "invalid buffer size", NULL); 
    }

    u32SizeDW = m_windowInfo.u32SizeDW;

    //create init block
    for(u32Index = 0; u32Index < u32SizeDW; u32Index++)
    {
        m_u64pBufferInit[u32Index] = 0x0;
    }
    //create sending block  
    for(u32Index = 0; u32Index < u32SizeDW; u32Index++)
    {
        m_u64pBufferSend[u32Index] = 0x0;
    }
    //create receive block
    for(u32Index = 0; u32Index < u32SizeDW; u32Index++)
    {
        m_u64pBufferRev[u32Index] = 0x0;
    }
    
    //release resource 
    free(m_u64pBufferSend);
    free(m_u64pBufferRev);
    free(m_u64pBufferInit);

    return NO_ERROR_MSG;
}


void
NTBDDAPI::releaseWindowResource()
{
    munmap(m_windowInfo.u64pUserAddrWrite, m_windowInfo.u32Size);
    munmap(m_windowInfo.u64pUserAddrRead, m_windowInfo.u32Size);
    
    m_windowInfo.u32Size = 0;
    m_windowInfo.u32SizeDW = 0;
    
    close(m_fd);
}


//receive and compare data 
ErrorMsg*
NTBDDAPI::receiveData(UINT64*   in_u64pBuffer,
                      UINT64*   in_u64pAddr,
                      UINT32    in_u32SizeDW)
{
    int     ret;
    UINT32  u32Index;

    //wait receiver's response 
    for(u32Index = 10; u32Index > 0; u32Index--)
    {
        printf("Receiving data ...\n"); 
        usleep(1000000);
        ret = memcmp(in_u64pBuffer, in_u64pAddr, sizeof(UINT64)*in_u32SizeDW);
        if(ret == 0)
        {
            printf("data is caught !!!\n");
            break;
        }
    }
    if(u32Index == 0)
    {
        return returnErrorMsg(ERROR_NTB_DDAPI_RECEIVE_DATA, "timeout to wait receiver", NULL);
    }
    return NO_ERROR_MSG;
}


ErrorMsg*
NTBDDAPI::setForIoTransfer(UINT64     in_u64PatternSend,
                           UINT64     in_u64PatternRev)
{
    ErrorMsg*   pErrorMsg;
    
    //get window resources  
    pErrorMsg = getWindowInfo(0); 
    if(pErrorMsg != NO_ERROR_MSG)
    {
        return returnErrorMsg(ERROR_NTB_DDAPI_IO_SETTING, NULL, pErrorMsg);
    }

    //create m_u64pBufferInit, m_u64pBufferSend, and m_u64pBufferRev 
    pErrorMsg = createDataBlocks(in_u64PatternSend, in_u64PatternRev);
    if(pErrorMsg != NO_ERROR_MSG)
    {
        return returnErrorMsg(ERROR_NTB_DDAPI_IO_SETTING, NULL, pErrorMsg);
    }
    return pErrorMsg;
}




ErrorMsg*
NTBDDAPI::initialWindow()
{
    ErrorMsg*   pErrorMsg;
    UINT32      u32SizeDW;
    UINT32      u32Index;
    UINT64*     u64pWriteAddr;
    UINT64*     u64pReadAddr;
    UINT64*     u64pBuffer;

    //get window resources  
    pErrorMsg = getWindowInfo(0); 
    if(pErrorMsg != NO_ERROR_MSG)
    {
        return returnErrorMsg(ERROR_NTB_DDAPI_IO_SETTING, NULL, pErrorMsg);
    }
    
    u32SizeDW       = m_windowInfo.u32SizeDW;
    u64pReadAddr    = m_windowInfo.u64pUserAddrRead;
    u64pWriteAddr   = m_windowInfo.u64pUserAddrWrite;

    u64pBuffer = (UINT64*)malloc(sizeof(UINT64)*u32SizeDW);
    for(u32Index = 0; u32Index < u32SizeDW; u32Index++)
    {
        u64pBuffer[u32Index] = 0x0;
    }
    memcpy(u64pReadAddr, u64pBuffer, sizeof(UINT64)*u32SizeDW);
    memcpy(u64pWriteAddr, u64pBuffer, sizeof(UINT64)*u32SizeDW);
   

    //release resource 
    free(u64pBuffer);
    releaseWindowResource();

    return NO_ERROR_MSG;
}
