/// @file       PCIBus.hpp
/// @brief      PCIBus class
/// @details    Used to store information of a PCI bus
/// @author     KenHu
/// @copyright  Copyright 2017 Hon Hai Technology Group. All Rights Reserved.

#ifndef PCIBUS_HPP
#define PCIBUS_HPP

#include <list>
#include "common.hpp"
#include "PCIDev.hpp"

using namespace std;

class PCIDev;

/// @class      PCIBus PCIBus.hpp "PCIBus.hpp"
/// @brief      Used to store information of a PCI bus
/// @param      m_u16Domain is UINT16 value for domain
/// @param      m_u8Bus is UINT8 value for bus
/// @param      m_devList is list of PCIDev for devices connected to this bus
/// @param      m_pUpperDev is pointer to PCIDev for upper layer device of this bus
class PCIBus
{
    public:
        PCIBus(UINT16 in_u16Domain, UINT8 in_u8Bus)
        {
            m_u16Domain = in_u16Domain;
            m_u8Bus = in_u8Bus;
            m_devList.clear();
            m_pUpperDev = NULL;
        }
        ~PCIBus()
        {

        }

        UINT16          m_u16Domain;
        UINT8           m_u8Bus;
        list<PCIDev *>  m_devList;
        PCIDev*         m_pUpperDev;
};

#endif
