#include "RTCDDAPI.hpp"
#include <unistd.h>
#include <sys/io.h>
#include <fcntl.h>
#include <linux/rtc.h>
#include <sys/ioctl.h>
#include <errno.h>

RTCDDAPI :: RTCDDAPI(STRING className) : BaseClassDDAPI(className)
{
    // Initializations
    m_rtcDevNode = -1;

    // 8254 timer
    if (ioperm(PIT_SYSTEM_COUNTER_IOPORT, 4, 1))
    {
        addErrorMsg((ERRORCODE)(ERROR_RTC_DDAPI_CONSTRUCTOR | ERROR_RTC_DDAPI_PORT_PERMISSION), NULL);
        return;
    }
}

RTCDDAPI :: ~RTCDDAPI()
{
    // 8254 timer
    ioperm(PIT_SYSTEM_COUNTER_IOPORT, 4, 0);

}

// about date and time
ErrorMsg * RTCDDAPI :: showDateTime(void)
{
    UINT32      year, month, date, hour, minute, second;
    INT32       retVal;
    struct      rtc_time rtc_tm;
    INT32       tmpYear;
    INT32       leapYear;
    INT32       tmpMonth;
    INT32       tmpDay;
    UINT32      i;
    UINT32      smallYear, bigYear;
    STRING      strDay[] = {"SUN", "MON", "TUE", "WED", "THU", "FRI", "SAT"};
        
    if ((m_rtcDevNode = open(RTC_DEV, O_RDONLY)) < 0)
    {
        return returnErrorMsg((ERRORCODE)(ERROR_RTC_DDAPI_SHOW_DATE_TIME | ERROR_RTC_DDAPI_RTC_DEVICE_OPEN_FAIL), NULL, NULL);
    }

	// Read the RTC time/date
    if(ioctl(m_rtcDevNode, RTC_RD_TIME, &rtc_tm))
    {
        close(m_rtcDevNode);

        sprintf(m_tmpString, "RTC IOCTL Error: RTC_RD_TIME  %s\n", strerror(errno));
        return returnErrorMsg((ERRORCODE)(ERROR_RTC_DDAPI_SHOW_DATE_TIME | ERROR_RTC_DDAPI_RTC_IOCTL_ACCESS_FAIL), m_tmpString, NULL);
    }
        
    year = rtc_tm.tm_year + 1900;
    month = rtc_tm.tm_mon + 1;
    date = rtc_tm.tm_mday;
    hour = rtc_tm.tm_hour;
    minute = rtc_tm.tm_min;
    second = rtc_tm.tm_sec;

    tmpYear = year - 2000;
    leapYear = 0;
    bigYear = year;
    smallYear = 2000;
    if (year < 2000)
    {
        bigYear = 2000;
        smallYear = year;
    }
    for (i = smallYear; i <= bigYear; i++)
    {
        if (i % 4 == 0)
        {
            if ((i % 100 == 0) && (i % 400 != 0))   //Not leap year
            {
            }
            else
            {
                if (i == year)
                {
                    if (month > 2)
                        leapYear++;
                }
                else
                    leapYear++;
            }
        }
    }
    
    switch (month)
    {
        case 1:
        case 10:
            tmpMonth = 5;
            break;
        case 2:
        case 3:
        case 11:
            tmpMonth = 1;
            break;
        case 4:
        case 7:
            tmpMonth = 4;
            break;
        case 5:
            tmpMonth = 6;
            break;
        case 6:
            tmpMonth = 2;
            break;
        case 8:
            tmpMonth = 0;
            break;
        case 9:
        case 12:
            tmpMonth = 3;
            break;
    }
    tmpDay = (tmpYear + leapYear + tmpMonth + date) % 7;
    printf("%04d/%02d/%02d  %s  %02d:%02d:%02d\n\n", 
           year, month, date, strDay[tmpDay], hour, minute, second);
    
    close(m_rtcDevNode);

    return NO_ERROR_MSG;
}
 
ErrorMsg * RTCDDAPI :: setDate(INT32 year, INT32 month, INT32 date)
{
    INT32   retVal;
    struct  rtc_time rtc_tm;

    if ((m_rtcDevNode = open(RTC_DEV, O_RDONLY)) < 0)
    {
        return returnErrorMsg((ERRORCODE)(ERROR_RTC_DDAPI_SET_DATE | ERROR_RTC_DDAPI_RTC_DEVICE_OPEN_FAIL), NULL, NULL);
    }

	// Read the RTC time/date
    if(ioctl(m_rtcDevNode, RTC_RD_TIME, &rtc_tm))
    {
        close(m_rtcDevNode);

        sprintf(m_tmpString, "RTC IOCTL Error: RTC_RD_TIME  %s\n", strerror(errno));
        return returnErrorMsg((ERRORCODE)(ERROR_RTC_DDAPI_SET_DATE | ERROR_RTC_DDAPI_RTC_IOCTL_ACCESS_FAIL), m_tmpString, NULL);
    }

    rtc_tm.tm_year = year - 1900;
    rtc_tm.tm_mon = month - 1;
    rtc_tm.tm_mday = date;

	// Set the RTC date
    if(ioctl(m_rtcDevNode, RTC_SET_TIME, &rtc_tm))
    {
        close(m_rtcDevNode);

        sprintf(m_tmpString, "RTC IOCTL Error: RTC_SET_TIME  %s\n", strerror(errno));
        return returnErrorMsg((ERRORCODE)(ERROR_RTC_DDAPI_SET_DATE | ERROR_RTC_DDAPI_RTC_IOCTL_ACCESS_FAIL), m_tmpString, NULL);
    }
        
    close(m_rtcDevNode);
    
    return NO_ERROR_MSG;
}

ErrorMsg * RTCDDAPI :: setTime(INT32 hour, INT32 minute, INT32 second)
{
    INT32   retVal;
    struct  rtc_time rtc_tm;

    if ((m_rtcDevNode = open(RTC_DEV, O_RDONLY)) < 0)
    {
        return returnErrorMsg((ERRORCODE)(ERROR_RTC_DDAPI_SET_DATE | ERROR_RTC_DDAPI_RTC_DEVICE_OPEN_FAIL), NULL, NULL);
    }

	// Read the RTC time/date
    if(ioctl(m_rtcDevNode, RTC_RD_TIME, &rtc_tm))
    {
        close(m_rtcDevNode);

        sprintf(m_tmpString, "RTC IOCTL Error: RTC_RD_TIME  %s\n", strerror(errno));
        return returnErrorMsg((ERRORCODE)(ERROR_RTC_DDAPI_SET_DATE | ERROR_RTC_DDAPI_RTC_IOCTL_ACCESS_FAIL), m_tmpString, NULL);
    }

    rtc_tm.tm_hour = hour;
    rtc_tm.tm_min = minute;
    rtc_tm.tm_sec = second;
    
	// Set the RTC time
    if(ioctl(m_rtcDevNode, RTC_SET_TIME, &rtc_tm))
    {
        close(m_rtcDevNode);

        sprintf(m_tmpString, "RTC IOCTL Error: RTC_SET_TIME  %s\n", strerror(errno));
        return returnErrorMsg((ERRORCODE)(ERROR_RTC_DDAPI_SET_DATE | ERROR_RTC_DDAPI_RTC_IOCTL_ACCESS_FAIL), m_tmpString, NULL);
    }
        
    close(m_rtcDevNode);
    
    return NO_ERROR_MSG;
}

// RTC 
ErrorMsg * RTCDDAPI :: setRTCRate(FLOAT rate)
{
    FLOAT       fValue;
    UINT32      u32Power2 = 1;
    
    if (rate >= 122.07 && rate <= 500000)
    {
        fValue = 1000000 / rate;

        m_rtcRate = (UINT32)fValue;

        if ((fValue - m_rtcRate) >= 0.5)
            m_rtcRate++;

        while (m_rtcRate > u32Power2)
            u32Power2 *= 2;
            
        if (m_rtcRate != u32Power2)
        {
            return returnErrorMsg((ERRORCODE)(ERROR_RTC_DDAPI_SET_RTC_TIMER_RATE | ERROR_BAD_PARAMETER_ONE), "rate isn't support", NULL);
        }
    }
    else
        return returnErrorMsg((ERRORCODE)(ERROR_RTC_DDAPI_SET_RTC_TIMER_RATE | ERROR_BAD_PARAMETER_ONE), "rate isn't support", NULL);
        
    return NO_ERROR_MSG;
}

ErrorMsg * RTCDDAPI :: startRTC(void)
{
    if ((m_rtcDevNode = open(RTC_DEV, O_RDWR)) < 0)
    {
        return returnErrorMsg((ERRORCODE)(ERROR_RTC_DDAPI_START_RTC_TIMER | ERROR_RTC_DDAPI_RTC_DEVICE_OPEN_FAIL), NULL, NULL);
    }

    if(ioctl(m_rtcDevNode, RTC_IRQP_SET, m_rtcRate))
    {
        sprintf(m_tmpString, "RTC IOCTL Error: RTC_IRQP_SET  %s\n", strerror(errno));
        return returnErrorMsg((ERRORCODE)(ERROR_RTC_DDAPI_START_RTC_TIMER | ERROR_RTC_DDAPI_RTC_IOCTL_ACCESS_FAIL), m_tmpString, NULL);
    }

    if(ioctl(m_rtcDevNode, RTC_PIE_ON, 0))
    {
        sprintf(m_tmpString, "RTC IOCTL Error: RTC_PIE_ON  %s\n", strerror(errno));
        return returnErrorMsg((ERRORCODE)(ERROR_RTC_DDAPI_START_RTC_TIMER | ERROR_RTC_DDAPI_RTC_IOCTL_ACCESS_FAIL), m_tmpString, NULL);
    }

    return NO_ERROR_MSG;
}

void RTCDDAPI :: stopRTC(void)
{
    ioctl(m_rtcDevNode, RTC_PIE_OFF, 0);
    
    if (m_rtcDevNode >= 0)
    {    
        close(m_rtcDevNode);
        m_rtcDevNode = -1;
    }
}

ErrorMsg * RTCDDAPI :: waitRTCInterrupt()
{
    UINT32      data;
        
    if (read(m_rtcDevNode, &data, 4) == -1)
    {
        stopRTC();
        sprintf(m_tmpString, "read RTC timer status fail  %s\n", strerror(errno));
        return returnErrorMsg((ERRORCODE)(ERROR_RTC_DDAPI_READ_RTC_TIMER_STATUS | ERROR_RTC_DDAPI_RTC_DEVICE_READ_FAIL), m_tmpString, NULL);
    }

    return NO_ERROR_MSG;

}

// 8254 Timer
void RTCDDAPI :: startPIT(UINT32 timerCount)
{
    UINT8 u8Value;

    setPITControlWord(PIT_MODE3);

    u8Value = (UINT8)(timerCount & 0xFF);
    outb(u8Value, PIT_SYSTEM_COUNTER_IOPORT);
    u8Value = (UINT8)((timerCount & 0xFF00) >> 8);
    outb(u8Value, PIT_SYSTEM_COUNTER_IOPORT);
    
    m_PITMode = PIT_MODE3;
}

void RTCDDAPI :: stopPIT(void)
{
    setPITControlWord(PIT_MODE0);
    
    outb(0, PIT_SYSTEM_COUNTER_IOPORT);
    outb(0, PIT_SYSTEM_COUNTER_IOPORT);
}

void RTCDDAPI :: setPITControlWord(PIT_TIMER_MODE PITMode)
{
    UINT8 u8Value;
    
    u8Value = (UINT8)PITMode << 1;
    u8Value += 3 << 4;
    outb(u8Value, PIT_CONTROL_COUNTER_IOPORT);
}

void RTCDDAPI :: convertPITTimeToCount(FLOAT usTime, UINT32* timerCount)
{   
    FLOAT   fValue;
    UINT32  u32Value;

    fValue = (usTime * 1000) / 838;
    
    u32Value = (UINT32)fValue;    
    
    if ((fValue - (UINT32)fValue) >= 0.5)
        u32Value = (UINT32)fValue + 1;
        
    *timerCount = u32Value;

}

void RTCDDAPI :: readPIT(UINT32* currentValue)
{
    UINT8 u8Value;
    
    outb(0, PIT_CONTROL_COUNTER_IOPORT);
    
    u8Value = inb(PIT_SYSTEM_COUNTER_IOPORT);
    *currentValue = (UINT32)u8Value;
    u8Value = inb(PIT_SYSTEM_COUNTER_IOPORT);
    *currentValue += (u8Value << 8);
}

ErrorMsg * RTCDDAPI :: rtcWDTTest(void)
{
	return returnErrorMsg((ERRORCODE)(ERROR_RTC_DDAPI_RTC_WDT_TEST | ERROR_FUNCTION_NOT_SUPPORTED));
}

