/// @file       SASIOCTL.cpp
/// @brief      LSI SAS controller ioctl class derived from IOCTL class
/// @details    This is an base class which define SCSI operation API
/// @version    1.0.0
/// @date       2014.02.20
/// @author     JerryLee
/// @copyright  Copyright 2014 Hon Hai Technology Group. All Rights Reserved.

#include "SASIOCTL.hpp"

SASIOCTL::SASIOCTL(STRING in_fileName) : IOCTL(in_fileName)
{
    // TODO Auto-generated constructor stub
    myIOCTLArgument = NULL;
}

SASIOCTL::~SASIOCTL()
{
    // TODO Auto-generated destructor stub
}

/// @fn         SASIOCTL::allocateIOCTLArgument
/// @brief      allocate every IOCTL command's argument
/// @param      IOCTL OP code
/// @exception  Nothing
/// @return     error code
ERRORCODE SASIOCTL::allocateIOCTLArgument(UINT8 IOCTLOperation)
{
    switch(IOCTLOperation)
    {
        case EXECUTE_LOOPBACK:
            if((myIOCTLArgument = (sas_loopback_param*) malloc(sizeof(sas_loopback_param))) != NULL)
                memset(myIOCTLArgument, 0, sizeof(sas_loopback_param));

            break;
        case CLEAR_ERR_LOG:
        case CLEAR_ERR_LOG_BY_PHY:
        case EXECUTE_IOC_SMP:
            if((myIOCTLArgument = (exec_ctrl *) malloc(sizeof(exec_ctrl))) != NULL)
                memset(myIOCTLArgument, 0, sizeof(exec_ctrl));

            break;
        case EXECUTE_IOC_HARD_RESET:
            if((myIOCTLArgument = (mpt3_ioctl_diag_reset*) malloc(sizeof(mpt3_ioctl_diag_reset))) != NULL)
                memset(myIOCTLArgument, 0, sizeof(mpt3_ioctl_diag_reset));

            break;
        case IOC_INFORMATION:
            if((myIOCTLArgument = (mpt3_ioctl_iocinfo*) malloc(sizeof(mpt3_ioctl_iocinfo))) != NULL)
                memset(myIOCTLArgument, 0, sizeof(mpt3_ioctl_iocinfo));

            break;
        case SHOW_ATTACHED_DEVICE:
        case ERROR_STATISTICS:
        case PHY_ERROR_COUNT:
        case LINK_STATUS:
        case LINK_SPEED:
        case CURRENT_TEMP:
        case IOC_SAS_ADDRESS:
        case PHY_SAS_ADDRESS:
            if((myIOCTLArgument = (get_info *) malloc(sizeof(get_info))) != NULL)
                memset(myIOCTLArgument, 0, sizeof(get_info));

            break;
        default:
            break;
    }

    if(myIOCTLArgument == NULL)
    {
        this->closeFileHandle();
        return ERROR_SAS_IOCTL_MEM_ALLOC_FAILED;
    }

    return NO_ERROR;
}

/// @fn         SASIOCTL::setIOCTLOperationArgument
/// @brief      set argument by IOCTL OP code
/// @param      IOCTL OP code
/// @exception  Nothing
/// @return     error code
ERRORCODE SASIOCTL::setIOCTLOperationArgument(UINT8 IOCTLOperation, void* data)
{
    if(myIOCTLArgument == NULL)
        return ERROR_SAS_IOCTL_MEM_ALLOC_FAILED;

    switch(IOCTLOperation)
    {
        case EXECUTE_LOOPBACK: // normal disk IO for SAS IOCTL
            static_cast<exec_ctrl*>(myIOCTLArgument)->exec_action = EXEC_LOOPBACK_EXEC;
            static_cast<exec_ctrl*>(myIOCTLArgument)->test_exc.ioc_num = targetControllerID;
            static_cast<exec_ctrl*>(myIOCTLArgument)->test_exc.random = 1; // use random field to convey the test level
            static_cast<exec_ctrl*>(myIOCTLArgument)->test_exc.dataPattern = 0xAA;
            static_cast<exec_ctrl*>(myIOCTLArgument)->test_exc.limit = 0; // use limit field to convey initiator phy#
            static_cast<exec_ctrl*>(myIOCTLArgument)->test_exc.loopCount = 10;

            break;
        case CLEAR_ERR_LOG:
            static_cast<exec_ctrl*>(myIOCTLArgument)->task_exc.ioc_num = targetControllerID;
            static_cast<exec_ctrl*>(myIOCTLArgument)->exec_action = EXEC_IOC_CLEAR_PHY_ERR_LOG;

            break;
        case CLEAR_ERR_LOG_BY_PHY:
            static_cast<exec_ctrl*>(myIOCTLArgument)->task_exc.ioc_num = targetControllerID;
            static_cast<exec_ctrl*>(myIOCTLArgument)->task_exc.phy_num = targetPHYNumber;
            static_cast<exec_ctrl*>(myIOCTLArgument)->exec_action = EXEC_IOC_CLEAR_SINGLE_PHY_ERR_LOG;

            break;
        case EXECUTE_IOC_SMP:
            static_cast<exec_ctrl*>(myIOCTLArgument)->task_exc.ioc_num = targetControllerID;
            static_cast<exec_ctrl*>(myIOCTLArgument)->exec_action = EXEC_IOC_EXP_SMP_TEST;

            break;
        case CURRENT_TEMP:
            static_cast<get_info*>(myIOCTLArgument)->get_action = GET_IOC_TEMP;
            static_cast<get_info*>(myIOCTLArgument)->temp_info.ioc_num = targetControllerID;
            static_cast<get_info*>(myIOCTLArgument)->temp_info.temperature = INVALID_TEMPERATURE;

            break;
        case EXECUTE_IOC_HARD_RESET:
            static_cast<mpt3_ioctl_diag_reset*>(myIOCTLArgument)->hdr.ioc_number = targetControllerID;

            break;
        case IOC_INFORMATION:
            static_cast<mpt3_ioctl_iocinfo*>(myIOCTLArgument)->hdr.ioc_number = targetControllerID;

            break;
        case IOC_SAS_ADDRESS:
            static_cast<get_info*>(myIOCTLArgument)->get_action = GET_IOC_WWID;
            static_cast<get_info*>(myIOCTLArgument)->ioc_num = targetControllerID;

            break;
        case SHOW_ATTACHED_DEVICE:
            static_cast<get_info*>(myIOCTLArgument)->get_action = GET_IOC_PHY_ATT_DEV_INFO;
            static_cast<get_info*>(myIOCTLArgument)->ioc_num = targetControllerID;
            static_cast<get_info*>(myIOCTLArgument)->phy_att_dev_info.ioc_num = targetControllerID;
            static_cast<get_info*>(myIOCTLArgument)->phy_att_dev_info.phy_num = targetPHYNumber;

            break;
        case ERROR_STATISTICS:
        case PHY_ERROR_COUNT:
            static_cast<get_info*>(myIOCTLArgument)->get_action = GET_IOC_PHY_ERR;
            static_cast<get_info*>(myIOCTLArgument)->ioc_num = targetControllerID;
            static_cast<get_info*>(myIOCTLArgument)->phy_err.ioc_num = targetControllerID;
            static_cast<get_info*>(myIOCTLArgument)->phy_err.phy_num = targetPHYNumber;

            break;
        case LINK_STATUS:
        case LINK_SPEED:
            static_cast<get_info*>(myIOCTLArgument)->get_action = GET_IOC_PHY_SPEED;
            static_cast<get_info*>(myIOCTLArgument)->ioc_num = targetControllerID;
            static_cast<get_info*>(myIOCTLArgument)->phy_speed.ioc_num = targetControllerID;
            static_cast<get_info*>(myIOCTLArgument)->phy_speed.phy_num = targetPHYNumber;

            break;
    }

    return NO_ERROR;
}

/// @fn         SASIOCTL::executeIOCTL
/// @brief      sent IOCTL to specific SAS IOC
/// @param      IOCTL OP code
/// @exception  Nothing
/// @return     error code
ERRORCODE SASIOCTL::executeIOCTL(UINT8 in_IOCTLOperation)
{
    INT8 status;

    switch(in_IOCTLOperation)
    {
        case EXECUTE_LOOPBACK:
            status = ioctl(myFileHandle, SASDIAG_EXEC_LOOPBACK, myIOCTLArgument);

            if (status == IOC_NUM_INVALID)
                return ERROR_IOCTL_SAS_INVALID_IOC_NUMBER;
            else if (status != 0)
                return ERROR_SAS_IOCTL_ACCESS_FAIL;

            break;
        case CLEAR_ERR_LOG:
        case CLEAR_ERR_LOG_BY_PHY:
        case EXECUTE_IOC_SMP:
            status = ioctl(myFileHandle, SASDIAG_EXEC, myIOCTLArgument);

            if (status == IOC_NUM_INVALID)
                return ERROR_IOCTL_SAS_INVALID_IOC_NUMBER;
            else if (status != 0)
                return ERROR_SAS_IOCTL_ACCESS_FAIL;

            break;
        case EXECUTE_IOC_HARD_RESET:
            status = ioctl(myFileHandle, MPT3HARDRESET, myIOCTLArgument);

            if (status == IOC_NUM_INVALID)
                return ERROR_IOCTL_SAS_INVALID_IOC_NUMBER;
            else if (status != 0)
                return ERROR_SAS_IOCTL_ACCESS_FAIL;

            break;
        default:
            return ERROR_IOCTL_UNSUPPORT_ARGUMENT;
    }

    return NO_ERROR;
}
/// @fn         SASIOCTL::executeIOCTL
/// @brief      sent IOCTL to specific SAS IOC
/// @param      IOCTL OP code
/// @param      returned IOCTL argument include several information we need
/// @exception  Nothing
/// @return     error code
ERRORCODE SASIOCTL::executeIOCTL(UINT8 IOCTLOperation, void* IOCTLArgument)
{
    INT8 status;

    switch(IOCTLOperation)
    {
        case CURRENT_TEMP:
            status = ioctl(myFileHandle, SASDIAG_GET_INFO, myIOCTLArgument);

            if(status == IOC_NUM_INVALID)
                return ERROR_IOCTL_SAS_INVALID_IOC_NUMBER;
            else if(status != 0)
                return ERROR_SAS_IOCTL_ACCESS_FAIL;
            else
                *(static_cast<UINT32*>(IOCTLArgument)) = static_cast<get_info*>(myIOCTLArgument)->temp_info.temperature;

            break;
        case IOC_SAS_ADDRESS:
            status = ioctl(myFileHandle, SASDIAG_GET_INFO, myIOCTLArgument);

            if(status == IOC_NUM_INVALID)
                return ERROR_IOCTL_SAS_INVALID_IOC_NUMBER;
            else if(status != 0)
                return ERROR_SAS_IOCTL_ACCESS_FAIL;
            else
                memcpy(static_cast<UINT32*>(IOCTLArgument), static_cast<get_info*>(myIOCTLArgument)->ioc_wwid,
                        sizeof(static_cast<get_info*>(myIOCTLArgument)->ioc_wwid));

            break;
        case IOC_INFORMATION:
            status = ioctl(myFileHandle, MPT3IOCINFO, myIOCTLArgument);

            if(status == IOC_NUM_INVALID)
                return ERROR_IOCTL_SAS_INVALID_IOC_NUMBER;
            else if(status != 0)
                return ERROR_SAS_IOCTL_ACCESS_FAIL;
            else
            {
                strcpy((char*) static_cast<mpt3_ioctl_iocinfo*>(IOCTLArgument)->driver_version,
                       (const char*) static_cast<mpt3_ioctl_iocinfo*>(myIOCTLArgument)->driver_version);

                static_cast<mpt3_ioctl_iocinfo*>(IOCTLArgument)->firmware_version =
                        static_cast<mpt3_ioctl_iocinfo*>(myIOCTLArgument)->firmware_version;
                static_cast<mpt3_ioctl_iocinfo*>(IOCTLArgument)->bios_version =
                        static_cast<mpt3_ioctl_iocinfo*>(myIOCTLArgument)->bios_version;
                static_cast<mpt3_ioctl_iocinfo*>(IOCTLArgument)->pci_information.u.bits.bus =
                        static_cast<mpt3_ioctl_iocinfo*>(myIOCTLArgument)->pci_information.u.bits.bus;
                static_cast<mpt3_ioctl_iocinfo*>(IOCTLArgument)->pci_information.u.bits.device =
                        static_cast<mpt3_ioctl_iocinfo*>(myIOCTLArgument)->pci_information.u.bits.device;
                static_cast<mpt3_ioctl_iocinfo*>(IOCTLArgument)->pci_information.u.bits.function =
                        static_cast<mpt3_ioctl_iocinfo*>(myIOCTLArgument)->pci_information.u.bits.function;
                static_cast<mpt3_ioctl_iocinfo*>(IOCTLArgument)->adapter_type =
                        static_cast<mpt3_ioctl_iocinfo*>(myIOCTLArgument)->adapter_type;
                static_cast<mpt3_ioctl_iocinfo*>(IOCTLArgument)->hw_rev =
                        static_cast<mpt3_ioctl_iocinfo*>(myIOCTLArgument)->hw_rev;
                static_cast<mpt3_ioctl_iocinfo*>(IOCTLArgument)->pci_id =
                        static_cast<mpt3_ioctl_iocinfo*>(myIOCTLArgument)->pci_id;
            }

            break;
        case SHOW_ATTACHED_DEVICE:
            status = ioctl(myFileHandle, SASDIAG_GET_INFO, myIOCTLArgument);

            if(status == IOC_NUM_INVALID)
                return ERROR_IOCTL_SAS_INVALID_IOC_NUMBER;
            else if(status != 0)
                return ERROR_SAS_IOCTL_ACCESS_FAIL;
            else
            {
                if((static_cast<get_info*>(myIOCTLArgument)->phy_att_dev_info.att_dev_info & MPI2_SAS_DEVICE_INFO_MASK_DEVICE_TYPE)
                     == MPI2_SAS_DEVICE_INFO_EDGE_EXPANDER)
                    *(static_cast<string*>(IOCTLArgument)) = "Edge Expander";

                if((static_cast<get_info*>(myIOCTLArgument)->phy_att_dev_info.att_dev_info & MPI2_SAS_DEVICE_INFO_MASK_DEVICE_TYPE)
                     == MPI2_SAS_DEVICE_INFO_FANOUT_EXPANDER)
                    *(static_cast<string*>(IOCTLArgument)) = "FanOut Expander";

                if(static_cast<get_info*>(myIOCTLArgument)->phy_att_dev_info.att_dev_info & MPI2_SAS_DEVICE_INFO_SSP_INITIATOR)
                {
                    if(static_cast<get_info*>(myIOCTLArgument)->phy_att_dev_info.att_dev_info & MPI2_SAS_DEVICE_INFO_SSP_TARGET)
                        *(static_cast<string*>(IOCTLArgument)) = "SAS Initiator and Target";
                    else
                        *(static_cast<string*>(IOCTLArgument)) = "SAS Initiator";
                }
                else if(static_cast<get_info*>(myIOCTLArgument)->phy_att_dev_info.att_dev_info & MPI2_SAS_DEVICE_INFO_SSP_TARGET)
                    *(static_cast<string*>(IOCTLArgument)) = "SAS Target";

                if(static_cast<get_info*>(myIOCTLArgument)->phy_att_dev_info.att_dev_info & MPI2_SAS_DEVICE_INFO_SATA_HOST)
                {
                    if(static_cast<get_info*>(myIOCTLArgument)->phy_att_dev_info.att_dev_info & MPI2_SAS_DEVICE_INFO_SATA_DEVICE)
                        *(static_cast<string*>(IOCTLArgument)) = "SATA Initiator and Target";
                    else
                        *(static_cast<string*>(IOCTLArgument)) = "SATA Initiator";
                }
                else if(static_cast<get_info*>(myIOCTLArgument)->phy_att_dev_info.att_dev_info & MPI2_SAS_DEVICE_INFO_SATA_DEVICE)
                    *(static_cast<string*>(IOCTLArgument)) = "SATA Target";

                if(static_cast<get_info*>(myIOCTLArgument)->phy_att_dev_info.att_dev_info & MPI2_SAS_DEVICE_INFO_ATAPI_DEVICE)
                    *(static_cast<string*>(IOCTLArgument)) = "ATAPI";

                if(!static_cast<get_info*>(myIOCTLArgument)->phy_att_dev_info.att_dev_info)
                    *(static_cast<string*>(IOCTLArgument)) = "not present";
            }

            break;
        case PHY_ERROR_COUNT:
            status = ioctl(myFileHandle, SASDIAG_GET_INFO, myIOCTLArgument);

            if(status == IOC_NUM_INVALID)
                return ERROR_IOCTL_SAS_INVALID_IOC_NUMBER;
            else if(status != 0)
                return ERROR_SAS_IOCTL_ACCESS_FAIL;
            else
            {
                static_cast<ErrorCountInfo*>(IOCTLArgument)->DisparityError =
                        static_cast<get_info*>(myIOCTLArgument)->phy_err.RunningDisparityErrorCount;
                static_cast<ErrorCountInfo*>(IOCTLArgument)->InvalidDwordError =
                        static_cast<get_info*>(myIOCTLArgument)->phy_err.InvalidDwordCount;
                static_cast<ErrorCountInfo*>(IOCTLArgument)->PhyResetProblem =
                        static_cast<get_info*>(myIOCTLArgument)->phy_err.PhyResetProblemCount;
                static_cast<ErrorCountInfo*>(IOCTLArgument)->LossDwordSync =
                        static_cast<get_info*>(myIOCTLArgument)->phy_err.LossDwordSynchCount;
            }

            break;
        case LINK_STATUS:
            status = ioctl(myFileHandle, SASDIAG_GET_INFO, myIOCTLArgument);

            if(status == IOC_NUM_INVALID)
                return ERROR_IOCTL_SAS_INVALID_IOC_NUMBER;
            else if(status != 0)
                return ERROR_SAS_IOCTL_ACCESS_FAIL;
            else
            {
                if(static_cast<get_info*>(myIOCTLArgument)->phy_speed.physical_link_speed == MPI2_SAS_NEG_LINK_RATE_NEGOTIATION_FAILED ||
                   static_cast<get_info*>(myIOCTLArgument)->phy_speed.physical_link_speed == MPI2_SAS_NEG_LINK_RATE_UNKNOWN_LINK_RATE)
                    *(static_cast<string*>(IOCTLArgument)) =  "UNKNOWN";
                else if(static_cast<get_info*>(myIOCTLArgument)->phy_speed.physical_link_speed == MPI2_SAS_NEG_LINK_RATE_PHY_DISABLED)
                    *(static_cast<string*>(IOCTLArgument)) = "DOWN";
                else
                    *(static_cast<string*>(IOCTLArgument)) = "UP";
            }

            break;
        case LINK_SPEED:
            status = ioctl(myFileHandle, SASDIAG_GET_INFO, myIOCTLArgument);

            if(status == IOC_NUM_INVALID)
                return ERROR_IOCTL_SAS_INVALID_IOC_NUMBER;
            else if(status != 0)
                return ERROR_SAS_IOCTL_ACCESS_FAIL;
            else
            {
                if(static_cast<get_info*>(myIOCTLArgument)->phy_speed.physical_link_speed == MPI2_SAS_NEG_LINK_RATE_1_5)
                    *(static_cast<string*>(IOCTLArgument)) = "1.5G";
                else if(static_cast<get_info*>(myIOCTLArgument)->phy_speed.physical_link_speed == MPI2_SAS_NEG_LINK_RATE_3_0)
                    *(static_cast<string*>(IOCTLArgument)) = "3G";
                else if(static_cast<get_info*>(myIOCTLArgument)->phy_speed.physical_link_speed == MPI2_SAS_NEG_LINK_RATE_6_0)
                    *(static_cast<string*>(IOCTLArgument)) = "6G";
                else if(static_cast<get_info*>(myIOCTLArgument)->phy_speed.physical_link_speed == MPI25_SAS_NEG_LINK_RATE_12_0)
                    *(static_cast<string*>(IOCTLArgument)) = "12G";
                else if(static_cast<get_info*>(myIOCTLArgument)->phy_speed.physical_link_speed == MPI2_SAS_NEG_LINK_RATE_NEGOTIATION_FAILED)
                    *(static_cast<string*>(IOCTLArgument)) = "Speed negotiation failed";
                else if(static_cast<get_info*>(myIOCTLArgument)->phy_speed.physical_link_speed == MPI2_SAS_NEG_LINK_RATE_UNKNOWN_LINK_RATE)
                    *(static_cast<string*>(IOCTLArgument)) = "Link rate unknown";
                else if(static_cast<get_info*>(myIOCTLArgument)->phy_speed.physical_link_speed == MPI2_SAS_NEG_LINK_RATE_PHY_DISABLED)
                    *(static_cast<string*>(IOCTLArgument)) = "PHY disabled";
            }

            break;
        default:
            return ERROR_IOCTL_UNSUPPORT_ARGUMENT;
    }

    return NO_ERROR;
}

