#ifndef STORAGE_INFO_HPP
#define STORAGE_INFO_HPP

#include <iostream>
#include <sstream>
#include <string>

#include "BaseClassDDAPI.hpp"

#define STORAGE_SECTOR_COUNT_PER_RDWR_ACCESS 2048 // 2048 blocks are about 1MB
#define STORAGE_USB_SECTOR_COUNT_PER_RDWR_ACCESS 128 // 128 blocks are about 64KB
#define STORAGE_SCSICMD_BUFFER_SIZE 2048 // 128 blocks are about 2048 Bytes

using namespace std;

enum StorageControllerType
{
     eStorageControllerTypeSAS = 0,
     eStorageControllerTypeSATA,
     eStorageControllerTypeUSB,
     eStorageControllerTypeNVME,
     eStorageControllerTypeUnknown,
};

// SCSI ERROR LOG definition
typedef struct
{
    struct {
            UINT64 fastECC;
            UINT64 delayedECC;
            UINT64 reReadWrite;
            UINT64 totalCorrected;
            UINT64 algoInvocation;
            string processedGigaBytes;
            UINT64 totalUnCorrected;
    } read;

    struct {
            UINT64 fastECC;
            UINT64 delayedECC;
            UINT64 reReadWrite;
            UINT64 totalCorrected;
            UINT64 algoInvocation;
            string processedGigaBytes;
            UINT64 totalUnCorrected;
    } write;

    struct {
            UINT64 fastECC;
            UINT64 delayedECC;
            UINT64 reReadWrite;
            UINT64 totalCorrected;
            UINT64 algoInvocation;
            string processedGigaBytes;
            UINT64 totalUnCorrected;
    } verify;

    UINT64 nmErrorCount;
} SCSI_ERROR_LOG;

typedef struct StorageSASSMARTData
{
    SCSI_ERROR_LOG errorLog;
    string strHealthStatus;
} StorageSASSMARTDataStruct;

typedef struct StorageSMARTAttribute
{
     UINT32 u32IDNumber;
     string strName;
     string strCurrentValue;
     string strWorstValue;
     string strThresholdValue;
     string strReadableValue;
     string strRawData;
     string strType;
     string strUpdate;
     string strHealthStatus;
     string strPastHealthStatus;
} StorageSMARTAttributeStruct;

typedef struct StorageDrive
{
     // basic info & data element
     UINT32 u32ControllerType; //0:SAS 1:SATA 2:USB 3:NVME
     UINT32 u32DefinedIndex;

     UINT32 u32PortNumber;
     UINT32 u32PHYNumber;
     UINT32 u32SlotNumber;

     string strDefinedName;
     string strDeviceNodeName;
     string strSGDeviceNodeName;

     string strSerialNumber;
     string strVendorName;
     string strModelName;
     UINT32 u32LogicalBlockSize;
     UINT32 u32PhysicalBlockSize;
     UINT64 u64SASAddress;

     UINT64 u64CapacityMB;
     string strLinkStatus;
     string strLinkSpeed;
     string strFirmwareVersion;
     string strTemperature;
     string strRotationRate;
     string strFormFactor;
     string strProtocalType;
     string strUSBVersion;

     // SMART info
     string strSMARTSupport;
     string strSMARTDiskHealthStatus;
     list <StorageSMARTAttributeStruct> SMARTAttributeList;
     StorageSASSMARTDataStruct SASSMARTData;

} StorageDriveStruct;

#endif
