#ifndef TPMDDAPI_H
#define TPMDDAPI_H

#include <fcntl.h>
#include <unistd.h>//for close() to be used in c++
#include <sys/io.h>

#include <string>
#include <vector>
#include <sys/ioctl.h>
#include <iostream>

#include "BaseClassDDAPI.hpp"
#include "TPMStructureDefinitions.hpp"

using namespace std;

#define TPM_UNUSED     0xff

#define TPM_DEV  "/dev/tpm0"

#define TPM_MAX_RSP_BUFFER_SIZE     64
#define TPM_MAX_STRING_LENGTH       22

#define ERR_COMMUNICATION   -1

#define TPM_HEADER_SIZE 10 //header size in bytes

typedef struct _TPM_CMD_INPUT
{
    UINT16  u16Tag; // TPMI_ST_COMMAND_TAG
    UINT32  u32CmdSize;
    UINT32  u32CmdCode; // TPM_CC

    union
    {
        UINT16      u16StartupType; // TPM_SU
        UINT8       u8YesNoData; // TPMI_YES_NO
        //TPM_ALG*    pAlgorithms; // TPML_ALG
        UINT32      u32Capability; // TPM_CAP
    }uParm1;

    union
    {
        UINT32 u32Value;
    }uParm2;

    union
    {
        UINT32 u32Value;
    }uParm3;

}TPM_CMD_INPUT;


class TPMDDAPI : public BaseClassDDAPI
{
    public:
         TPMDDAPI(STRING);
        ~TPMDDAPI();

        ERRORCODE   openDevice();
        void        closeDevice();

        ErrorMsg*   commandInterface(UINT8*   in_u8pCmdBuffer,
                                     UINT32   in_u32CmdBufferSize,
                                     UINT8*   out_u8pRspBuffer,
                                     UINT32   in_u32RspBufferSize,
                                     UINT32*  out_u32pRspSize); //size the TPM feedback

        ErrorMsg*   getCmdSize(TPM_CMD_INPUT  in_cmdList,
                               UINT32*        out_u32pCmdSize);

        ErrorMsg*   createCmdList(TPM_CMD_INPUT     in_cmdList,
                                  UINT8*            out_u8pCmdList);

        ErrorMsg*   sendCommandManually(STRING    in_cmdFileName,
                                        UINT8*    out_u8pRspBuffer,
                                        UINT32    in_u32RspBufferSize,
                                        UINT32*   out_u32pRspSize); //size TPM feedback


    private:
        int             m_tpmDevice;
        ErrorMsg*       m_pErrorMsg;
        ERRORCODE       m_eErrorCode;
        
        void converDWordToBytes(UINT32     in_u32Data,
                                UINT8*     out_u8pBuffer);

        //void convertStringToDWord(char*     in_string,
        //                          UINT32*   out_u32pData);

};

#endif
