#include "USBDDAPI.hpp"

USBDDAPI::USBDDAPI(STRING className) : BaseClassDDAPI(className)
{

    list<string> TempUSBDeviceBusPortList;
    TempUSBDeviceBusPortList.clear();

     memset(m_senseBuff, 0, sizeof(m_senseBuff));
    //memset(USBConfig, "", sizeof(USBConfig));

    for(UINT32 index = 0; index < MAX_USB_DEVICE_NO; index++)
    {
        USBConfig[index].Index          = USB_ID_UNDEFINED;
        USBConfig[index].WellKnownName[0]   = (CHAR)NULL;
        USBConfig[index].DeviceNode[0]      = (CHAR)NULL;
        USBConfig[index].BusPortAddress[0]  = (CHAR)NULL;
    }
    for(UINT32 index = 0; index < MAX_DATA_ELEMENT_NUMBER; index++)
    {
        USBElementMap[index].DateElement  = MAX_DATA_ELEMENT_NUMBER;
        USBElementMap[index].FileName[0]  = (CHAR)NULL;
    }

    for(UINT32 index = 0; index < MAX_DATA_ELEMENT_NUMBER; index++)
    {
        USB_DATA_ELEMENT eCurrent = USB_DATA_ELEMENT(index);
        USBElementMap[eCurrent].DateElement = USB_DATA_ALL_INFO;
        switch(eCurrent)
        {
            case USB_DATA_ALL_INFO:
                USBElementMap[eCurrent].FileName[0] = (CHAR)NULL;
                break;
            case USB_DATA_NODE_INFO:
                USBElementMap[eCurrent].FileName[0]= (CHAR)NULL;
                break;
            case USB_DATA_SN_INFO:
                sprintf(USBElementMap[eCurrent].FileName,"%s", "serial");
                break;
            case USB_DATA_VENDOR_INFO:
                sprintf(USBElementMap[eCurrent].FileName,"%s", "manufacturer");
                break;
            case USB_DATA_CAPACITY_INFO:
                sprintf(USBElementMap[eCurrent].FileName,"%s", "size");
                break;
            case USB_DATA_LINK_INFO:
                sprintf(USBElementMap[eCurrent].FileName,"%s", "version");
                break;
            case USB_DATA_SPEED_INFO:
                sprintf(USBElementMap[eCurrent].FileName,"%s", "speed");
                break;
            case USB_DATA_PHYSICAL_BLOCK_SIZE_INFO:
                sprintf(USBElementMap[eCurrent].FileName,"%s", "queue/physical_block_size");
                break;
            case USB_DATA_LOGICAL_BLOCK_SIZE_INFO:
                sprintf(USBElementMap[eCurrent].FileName,"%s", "queue/logical_block_size");
                break;
            case USB_DATA_SG_NODE_INFO:
            	sprintf(USBElementMap[eCurrent].FileName,"%s", "device/scsi_generic/");
                break;
            case MAX_DATA_ELEMENT_NUMBER:
                break;
            default:
                break;

        }
    }





    searchUSBInterfaceAddress(&TempUSBDeviceBusPortList);
    confirmUSBDeviceType(&TempUSBDeviceBusPortList, USBConfig);

    /*  For Debug
    while(!TempUSBDeviceBusPortList.empty())
    {
        for Debug
        printf("Link:%s\n", TempUSBDeviceBusPortList.front().c_str());
        TempUSBDeviceBusPortList.pop_front();
    }
    */




    /*
    searchUSBStorageDeviceNode(&DeviceNodeList);
    UINT32 intIndex = 0;

    //while(!DeviceNodeList.empty())
    for(list<string>::iterator ListIndex=DeviceNodeList.begin(); ListIndex != DeviceNodeList.end(); ListIndex++)
    {
        USBConfig[intIndex].LogicalID = intIndex;
        sprintf(USBConfig[intIndex].WellKnownName,"usb%d", intIndex);
        sprintf(USBConfig[intIndex].DeviceNode, "%s", (*ListIndex).c_str());
        //DeviceNodeList.pop_front();
        //printf("WellKnownName=%s, DeviceNode=%s\n", USBConfig[intIndex].WellKnownName, USBConfig[intIndex].DeviceNode);
        intIndex++;
    }
    */
}

USBDDAPI::~USBDDAPI()
{
    USBDeviceBusPortList.clear();
}

ErrorMsg *USBDDAPI::getUSBDevicesList(list<string> *pUSBList)
{
    //ErrorMsg*   ErrorMsg = NO_ERROR_MSG;
    //INT32 intIndex = 0;
/*
    for(list<string>::iterator ListIndex = USBDeviceBusPortList.begin() ;
    ListIndex != USBDeviceBusPortList.end() ; ListIndex++)
    {
        pUSBList->push_back(USBConfig[intIndex].WellKnownName);
        intIndex++;
    }
    */
    for(INT32 intIndex = 0; intIndex < MAX_USB_DEVICE_NO ; intIndex++)
    {
        //printf("%s\n", USBConfig[intIndex].WellKnownName);
        if(USBConfig[intIndex].WellKnownName[0] == (CHAR)NULL)
            continue;
        pUSBList->push_back(USBConfig[intIndex].WellKnownName);
    }
    if(pUSBList->empty())
    {
        pUSBList->push_back("Not Attach Any USB Storage Device");
    }

    return NO_ERROR_MSG;
}


ErrorMsg*  USBDDAPI::searchUSBInterfaceAddress(list<string> *returnDataInfo)
{
    //ErrorMsg* pErrorMsg = NULL;
    CHAR chDirPath[256];
    struct dirent *pEntry = NULL;
    struct stat attr;
    DIR *pDir;
    //CHAR DeviceNodeName[256];

    for(UINT32 i=0 ; i < USB_VENDOR_SPEC; i++  )
    {
        //printf("%d\n",i);
        USB_INTERFACE_TYPE eCurrent = USB_INTERFACE_TYPE(i);
        switch(eCurrent)
        {
            case USB_HID:
                //printf("USB_HID");
                strcpy(chDirPath, "/sys/bus/usb/drivers/usbhid/");
                break;
            case USB_STORAGE:
                //printf("USB_STORAGE");
                strcpy(chDirPath, "/sys/bus/usb/drivers/usb-storage/");
                break;
            default:
                strcpy(chDirPath, "");
                break;
        }

        pDir = opendir(chDirPath);
        if(pDir == NULL)
            continue;
        while((pEntry = readdir(pDir)) != NULL)
        {

            char chFileName[256];
            //INT16 u32Link = 0;


            memset(chFileName, '\0', sizeof(chFileName));
            snprintf(chFileName, sizeof(chFileName), "%s%s", chDirPath,
                     pEntry->d_name);

            lstat(chFileName, &attr);
            if( S_ISLNK(attr.st_mode) && !strstr(chFileName,"module") )
            {
                //printf("%s\n", chFileName);
                returnDataInfo->push_back(pEntry->d_name);
            }
        }
        closedir(pDir);

    }
    return NO_ERROR_MSG;
}





ErrorMsg*  USBDDAPI::confirmUSBDeviceType(list<string> *USBInterfaceAddressList,
                                          USB_CONFIGURATION *returnUSBConfiguration)
{

    UINT32 InterfaceClass;
    UINT32 InterfaceProtocol;
    USB_INTERFACE_TYPE eCurrentType;
    UINT32 USBKeyBoardCount = 0;
    UINT32 USBMouseCount = 0;
    UINT32 USBStorageCount = 0;

    UINT32 LogicalID = 0;

    CHAR  tempString[256];
    CHAR  *tempAddress;

    string  tempWellKnownName;
    tempWellKnownName.clear();

    string tempUSBStorageNode;
    tempUSBStorageNode.clear();



    //CHAR chDirPath[256];


    string tempStr;
    tempStr.clear();
    //strcpy(chDirPath, "/sys/bus/usb/devicess/");


    //popen();

    while(!USBInterfaceAddressList->empty())
    {
        tempStr.clear();
        tempWellKnownName.clear();
        //Get Interface Class Value
        getUSBDeviceInterfaceElement(USBInterfaceAddressList->front() ,"bInterfaceClass", &tempStr);
        //Get Interface Protocol value
        //printf("ResultString=%s\n", tempStr.c_str());
        InterfaceClass = atoi(tempStr.c_str());
        //printf("bInterfaceClass=%d\n", InterfaceClass);
        /* For Debug
        printf("String : %s\n", strDirPath.c_str());
        */
        tempStr.clear();
        getUSBDeviceInterfaceElement(USBInterfaceAddressList->front() ,"bInterfaceProtocol", &tempStr);
        //Get Interface Protocol value
        //printf("ResultString=%s\n", tempStr.c_str());
        InterfaceProtocol = atoi(tempStr.c_str());
        //printf("bInterfaceProtocol=%d\n", InterfaceProtocol);

        eCurrentType = USB_INTERFACE_TYPE(InterfaceClass);
        switch(eCurrentType)
        {
            case USB_HID:
                if(InterfaceProtocol == 1)
                {
                    tempWellKnownName.append("KEYBOARD");
                    sprintf(tempString, "%d", USBKeyBoardCount++);
                    tempWellKnownName.append(tempString);


                    //printf("!!!%s\n", USBInterfaceAddressList->front().c_str());
                    sprintf(tempString, "%s", USBInterfaceAddressList->front().c_str());
                    //printf("tempString=%s\n", tempString);
                    tempAddress = tempString;//strtok(tempString, ":");

                    //printf("tempAddress=%s\n", tempAddress);
                    addUSBDeviceEntry(LogicalID++, tempWellKnownName, "", tempAddress, returnUSBConfiguration);

                }
                else if(InterfaceProtocol == 2)
                {
                    tempWellKnownName.append("MOUSE");
                    sprintf(tempString, "%d", USBMouseCount++);
                    tempWellKnownName.append(tempString);


                    //printf("!!!%s\n", USBInterfaceAddressList->front().c_str());
                    sprintf(tempString, "%s", USBInterfaceAddressList->front().c_str());
                    //printf("tempString=%s\n", tempString);
                    tempAddress = tempString;//strtok(tempString, ":");
                    //printf("tempAddress=%s\n", tempAddress);
                    addUSBDeviceEntry(LogicalID++, tempWellKnownName, "", tempAddress, returnUSBConfiguration);
                }
                break;
            case USB_STORAGE:
                tempWellKnownName.append("USB");
                sprintf(tempString, "%d", USBStorageCount++);
                tempWellKnownName.append(tempString);
                //printf("!!!%s\n", USBInterfaceAddressList->front().c_str());
                sprintf(tempString, "%s", USBInterfaceAddressList->front().c_str());
                //printf("tempString=%s\n", tempString);
                tempAddress = tempString;//strtok(tempString, ":");

                //printf("tempAddress=%s\n", tempAddress);
                searchUSBStorageDeviceNode(tempAddress, &tempUSBStorageNode);
                addUSBDeviceEntry(LogicalID++, tempWellKnownName, tempUSBStorageNode, tempAddress, returnUSBConfiguration);

                break;
            default:
                break;
        }
        //printf("\n");
        USBInterfaceAddressList->pop_front();
    }

    /*//For Debug
    for(INT32 i = 0 ; i < LogicalID ; i++)
    {
        printf("LogicalID=%d\n", returnUSBConfiguration[i].LogicalID);
        printf("DeviceNode=%s\n", returnUSBConfiguration[i].DeviceNode);
        printf("WellKnownName=%s\n", returnUSBConfiguration[i].WellKnownName);
        printf("BusPortAddress=%s\n", returnUSBConfiguration[i].BusPortAddress);
        printf("\n");
    }
    */


    return NO_ERROR_MSG;
}



ErrorMsg*  USBDDAPI::getUSBDeviceInterfaceElement(string InterfaceAddress,
                                                  string strType,
                                                  string *returnString)
{
    FILE *fp = NULL;
    CHAR tempStr[256];

    string strDirPath;
    string strCommand;
    strDirPath.clear();
    strCommand.clear();


    strDirPath.append("/sys/bus/usb/devices/");
    strDirPath.append(InterfaceAddress);
    strDirPath.append("/");
    strCommand.append("cat ");
    strCommand.append(strDirPath);
    strCommand.append(strType);


    fp = popen(strCommand.c_str(), "r");
    fgets(tempStr, sizeof(tempStr), fp);
    if(tempStr[strlen(tempStr) - 1] == '\n')
        tempStr[strlen(tempStr) - 1] = '\0';

    returnString->append(tempStr);
    //printf("returnString=%s", returnString->c_str());

    strDirPath.clear();
    strCommand.clear();


    pclose(fp);
    return NO_ERROR_MSG;
}





ErrorMsg*  USBDDAPI::addUSBDeviceEntry(UINT32 Index,
                                       string WellKnownName,
                                       string DeviceNode,
                                       string BusPortAddress,
                                       USB_CONFIGURATION *returnUSBConfiguration)
{
    //printf("Index=%d\n", Index);
    //printf("WellKnownName=%s\n", WellKnownName.c_str());
    //printf("DeviceNode=%s\n", DeviceNode.c_str());
    //printf("BusPortAddress=%s\n", BusPortAddress.c_str());
    //printf("returnUSBConfiguration=%x\n", returnUSBConfiguration);

    returnUSBConfiguration[Index].Index = Index;
    sprintf(returnUSBConfiguration[Index].WellKnownName, "%s", WellKnownName.c_str());
    sprintf(returnUSBConfiguration[Index].DeviceNode, "%s", DeviceNode.c_str());
    sprintf(returnUSBConfiguration[Index].BusPortAddress,"%s", BusPortAddress.c_str());



    return NO_ERROR_MSG;
}








/* *************************************************************************
 * TraceUsbDeviceNode
 *  Search USB Device Node
 *
 * Input:   Node List
 *
 * Returns: ERRORCODE
 *
 * ************************************************************************* */
ErrorMsg* USBDDAPI::searchUSBStorageDeviceNode(string BusPortAddress, string* pDeviceNodeList)
{
    ErrorMsg* pErrorMsg = NULL;

    struct dirent *pEntry = NULL;
    struct stat attr;
    CHAR chDirPath[256];
    CHAR DeviceNodeName[256];

    strcpy(chDirPath, "/sys/block/");
    m_errorCode = NO_ERROR_MSG;
    //CHAR* p;
    DIR *pDir = opendir(chDirPath);

    if(pDir == NULL)
    {
        //sprintf(m_tempString,"open dirctory %s error when searching USB device name\n",m_dirPath);
        //m_errorCode = (ERRORCODE)(ERROR_USB_DDAPI_SCAN_USB_DEVICE);
        //return m_errorCode;
        return returnErrorMsg((ERRORCODE) (ERROR_USB_DDAPI_SCAN_USB_DEVICE),
                              NULL, pErrorMsg);
    }
    while((pEntry = readdir(pDir)) != NULL)
    {

        char chFileName[256];
        INT16 u32Link = 0;

        memset(chFileName, '\0', sizeof(chFileName));
        snprintf(chFileName, sizeof(chFileName), "%s%s", chDirPath,
                 pEntry->d_name);

        lstat(chFileName, &attr);

        if(S_ISDIR(attr.st_mode))
        {

        }
        else if(S_ISLNK(attr.st_mode))
        {
            char chTemp[256];

            memset(chTemp, '\0', sizeof(chTemp));

            u32Link = readlink(chFileName, chTemp, sizeof(chTemp) - 1);
            //printf("chTemp=%s\n", chTemp);
            if(u32Link != -1)
            {
                chTemp[u32Link] = '\0';
            }
            else
            {

            }
            if(strstr(chTemp, BusPortAddress.c_str()))
            {
                //strcpy(DeviceNodeName, "/dev/");
                strcpy(DeviceNodeName, "");
                strcat(DeviceNodeName, pEntry->d_name);

                //printf("DeviceNodeName=%s\n",DeviceNodeName);
                //DeviceNode->push_back(pEntry->d_name);
                //pDeviceNodeList->push_back(DeviceNodeName);
                pDeviceNodeList->clear();
                pDeviceNodeList->append(DeviceNodeName);
            }
        }
        else if(S_ISREG(attr.st_mode))
        {

        }
    }
    closedir(pDir);
    /*
     if(DeviceNode->empty())
     m_errorCode = (ERRORCODE)(ERROR_USB_DRIVER_NO_USB_DEVICE);
     */
    return returnErrorMsg(NO_ERROR, NULL, NULL);
}

ErrorMsg* USBDDAPI::getIndexByName(STRING WellKnownName, UINT32 *Index)
{

    ErrorMsg* pErrorMsg = NULL;

    CHAR tempName[(USB_DEVICE_NAME_LENGTH + 1)];

    // we need upper-case names, but don't change the callers input string
    strncpy(tempName, WellKnownName, sizeof(tempName));
    ConvertStringToUpperCase(tempName);

    for (UINT32 USBIndex = 0; USBIndex < MAX_USB_DEVICE_NO; USBIndex++ )
    {
        if (USBConfig[USBIndex].WellKnownName == (CHAR)NULL)
            continue;
        if (strcmp(USBConfig[USBIndex].WellKnownName, tempName) == 0)
        {
            *Index = USBConfig[USBIndex].Index;
            return NO_ERROR_MSG;
            break;
        }
    }

    return returnErrorMsg((ERRORCODE)(ERRIR_USB_DDAPI_GET_LOGICAL_ID_BY_NAME | ERROR_USB_DDAPI_DEVICE_NAME_NOT_FOUND), NULL, pErrorMsg);
}

ErrorMsg* USBDDAPI::getElementByIndex(UINT32 Index,
                                USB_DATA_ELEMENT USBDataElement,
                                list<string> *returnDataInfo)
{
     ErrorMsg* pErrorMsg = NULL;
    switch(USBDataElement)
    {
        case USB_DATA_ALL_INFO:
            //printf("%s USB_DATA_ALL_INFO\n", __func__);
            break;
        case USB_DATA_NODE_INFO:
            pErrorMsg = getUSBDeviceNode(Index, returnDataInfo);
            if(pErrorMsg != NO_ERROR_MSG)
            {
                return returnErrorMsg(ERRIR_USB_DDAPI_GET_DATA_ID_BY_ID, NULL, pErrorMsg);
            }
            break;
        // attention !!
        case USB_DATA_SN_INFO:
        case USB_DATA_VENDOR_INFO:
        case USB_DATA_LINK_INFO:
        case USB_DATA_SPEED_INFO:
            pErrorMsg = getUSBInfoFromSysBus(Index,
                                             USBDataElement,
                                             returnDataInfo);
            if(pErrorMsg != NO_ERROR_MSG)
            {
                return returnErrorMsg(ERRIR_USB_DDAPI_GET_DATA_ID_BY_ID, NULL, pErrorMsg);
            }
            break;

        // attention !!

        case USB_DATA_CAPACITY_INFO:
        case USB_DATA_PHYSICAL_BLOCK_SIZE_INFO:
        case USB_DATA_LOGICAL_BLOCK_SIZE_INFO:

            if(USBConfig[Index].DeviceNode[0] == (CHAR)NULL)
            {
               returnDataInfo->push_front("None");
            }
            else
            {
                pErrorMsg = getUSBInfoFromSysBlock(Index,
                                                   USBDataElement,
                                                   returnDataInfo);
                if(pErrorMsg != NO_ERROR_MSG)
                {
                    return returnErrorMsg(ERRIR_USB_DDAPI_GET_DATA_ID_BY_ID, NULL, pErrorMsg);
                }
            }
            break;
        case USB_DATA_SG_NODE_INFO:
            if(USBConfig[Index].DeviceNode[0] == (CHAR)NULL)
            {
               returnDataInfo->push_front("None");
            }
            else
            {
                pErrorMsg = getUSBSGDeviceNode(Index,
                							   USBDataElement,
											   returnDataInfo);
                if(pErrorMsg != NO_ERROR_MSG)
                {
                    return returnErrorMsg(ERRIR_USB_DDAPI_GET_DATA_ID_BY_ID, NULL, pErrorMsg);
                }
            }
            break;
        default:
            return returnErrorMsg((ERRORCODE)(ERRIR_USB_DDAPI_GET_DATA_ID_BY_ID | ERROR_USB_DDAPI_DATA_ELEMENT_NOT_FOUND), NULL, pErrorMsg);
            break;
    }
    return NO_ERROR_MSG;
}

ErrorMsg*  USBDDAPI::getUSBSysBusAddress(UINT32   Index,
                                         string* returnDataInfo)
{
    //printf("%s\n", __func__);

    //string  TargetDeviceAddress;
    //TargetDeviceAddress.clear();

    for(UINT32 USBIndex = 0 ; USBIndex < MAX_USB_DEVICE_NO ; USBIndex++)
    {
        //printf("[%d]:%s\n", USBIndex, USBConfig[USBIndex].BusPortAddress);

        if(USBConfig[USBIndex].Index == Index)
        {
            //returnDataInfo->push_front(USBConfig[USBIndex].DeviceNode);
            *returnDataInfo = USBConfig[USBIndex].BusPortAddress;
            //printf("returnDataInfo=%s\n", (*returnDataInfo).c_str());
            break;
        }
    }
    return NO_ERROR_MSG;
}








ErrorMsg* USBDDAPI::getUSBDeviceNode(UINT32 Index,
                                     list<string> *returnDataInfo)
{
    string USBDeviceNode;
    USBDeviceNode.clear();
    ErrorMsg* pErrorMsg = NULL;

    for(UINT32 USBIndex = 0 ; USBIndex < MAX_USB_DEVICE_NO ; USBIndex++)
    {
        if(USBConfig[USBIndex].Index == Index)
        {
            if(USBConfig[USBIndex].DeviceNode[0] == (CHAR)NULL)
            {
                returnDataInfo->push_front("None");
            }
            else
            {
                //USBDeviceNode.append("Node:       ");
                USBDeviceNode.append("/dev/");
                USBDeviceNode.append(USBConfig[USBIndex].DeviceNode);
                //returnDataInfo->push_front(USBConfig[USBIndex].DeviceNode);
                returnDataInfo->push_front(USBDeviceNode);
            }
            return NO_ERROR_MSG;
            break;
        }
    }

    return returnErrorMsg((ERRORCODE)(ERRIR_USB_DDAPI_GET_USB_DEVICE_NODE | ERROR_USB_DDAPI_DEVICE_NAME_NOT_FOUND), NULL, pErrorMsg);
}

ErrorMsg* USBDDAPI::getUSBInfoFromSysBus(UINT32 Index,
                                         USB_DATA_ELEMENT   DataElement,
                                         list<string>* returnDataInfo)

{
    CHAR    tempString[256];
    string  strResult;
    ErrorMsg* pErrorMsg = NULL;
    FILE    *fp = NULL;

    memset(tempString, 0, sizeof(tempString));
    strResult.clear();

    //printf("%s\n", __func__);
    string TempUSBAddress;
    string SysBusFilePath;
    TempUSBAddress.clear();
    SysBusFilePath.clear();

    SysBusFilePath.append("cat ");
    SysBusFilePath.append(USB_SYSTEM_BUS_PATH);
    getUSBSysBusAddress(Index, &TempUSBAddress);
    sprintf(tempString, "%s", TempUSBAddress.c_str());
    strtok(tempString,":");
    SysBusFilePath.append(tempString);
    SysBusFilePath.append("/");
    SysBusFilePath.append(USBElementMap[DataElement].FileName);
    //printf("!!!%s\n",SysBusFilePath.c_str());

    memset(tempString, 0, sizeof(tempString));
    fp=popen(SysBusFilePath.c_str(),"r");
    if (fp == NULL)
    {
        return returnErrorMsg((ERRORCODE)(ERRIR_USB_DDAPI_GET_USB_INFO_FROM_SYS_BUS | ERROR_OPEN_FILE), NULL, pErrorMsg);
    }
    fgets(tempString, sizeof(tempString), fp);
    if(tempString[strlen(tempString) - 1] == '\n')
        tempString[strlen(tempString) - 1] = '\0';

    //printf(" USBDDAPI :: getDataByName\n");
    //printf("WellKnownName=%s, USBDataElement=%d\n", WellKnownName, USBDataElement);
    switch(DataElement)
    {
        case USB_DATA_SN_INFO:
            //strResult.append("SN:         ");
            strResult.append(tempString);
            break;
        case USB_DATA_VENDOR_INFO:
            //strResult.append("Vendor:     ");
            strResult.append(tempString);
            break;
        case USB_DATA_LINK_INFO:
            //strResult.append("Link:      ");
            strResult.append(tempString);
            break;
        case USB_DATA_SPEED_INFO:
            //strResult.append("Speed:      ");
            strResult.append(tempString);
            //strResult.append(" M/s");
            break;
        default:
            return returnErrorMsg((ERRORCODE)(ERRIR_USB_DDAPI_GET_USB_INFO_FROM_SYS_BUS | ERROR_USB_DDAPI_DATA_ELEMENT_NOT_FOUND), NULL, pErrorMsg);
            break;
    }
    returnDataInfo->push_front(strResult);

    if(pclose(fp) != 0)
    {
        return returnErrorMsg((ERRORCODE)(ERRIR_USB_DDAPI_GET_USB_INFO_FROM_SYS_BUS | ERROR_CLOSE_FILE), NULL, pErrorMsg);
    }

    return NO_ERROR_MSG;
}

ErrorMsg*  USBDDAPI::getUSBInfoFromSysBlock(UINT32   Index,
                                            USB_DATA_ELEMENT   DataElement,
                                            list<string>* returnDataInfo)
{
    CHAR    tempString[256];
    string  strResult;
    FILE    *fp = NULL;
    ErrorMsg* pErrorMsg = NULL;

    memset(tempString, 0, sizeof(tempString));

    string TempUSBAddress;
    string SysBlockFilePath;
    TempUSBAddress.clear();
    SysBlockFilePath.clear();


    SysBlockFilePath.append("cat ");
    SysBlockFilePath.append(USB_SYSTEM_BLPCK_PATH);
    SysBlockFilePath.append(USBConfig[Index].DeviceNode);
    SysBlockFilePath.append("/");
    SysBlockFilePath.append(USBElementMap[DataElement].FileName);

    //printf("SysBlockFilePath=%s\n", SysBlockFilePath.c_str());

    memset(tempString, 0, sizeof(tempString));
    fp=popen(SysBlockFilePath.c_str(),"r");
    if(fp == NULL)
    {
        return returnErrorMsg((ERRORCODE)(ERRIR_USB_DDAPI_GET_USB_INFO_FROM_SYS_BLOCK | ERROR_OPEN_FILE), NULL, pErrorMsg);
    }
    fgets(tempString, sizeof(tempString), fp);

    if(tempString[strlen(tempString) - 1] == '\n')
        tempString[strlen(tempString) - 1] = '\0';

    //strResult.append(tempString);
    //memset(tempString, 0, sizeof(tempString));

    switch(DataElement)
    {
        case USB_DATA_CAPACITY_INFO:
            //strResult.append("Capacity:   ");

            sprintf(tempString,"%.2f\n", (double)((atol(tempString)*USB_BLOCK_SIZE)/MEGA_BYTES_SIZE));
            if(tempString[strlen(tempString) - 1] == '\n')
                tempString[strlen(tempString) - 1] = '\0';
            strResult.append(tempString);
            //strResult.append(" MB");
            break;
        case USB_DATA_PHYSICAL_BLOCK_SIZE_INFO:
        case USB_DATA_LOGICAL_BLOCK_SIZE_INFO:
            if(tempString[strlen(tempString) - 1] == '\n')
                tempString[strlen(tempString) - 1] = '\0';
            strResult.append(tempString);
            break;
        default:
            return returnErrorMsg((ERRORCODE)(ERRIR_USB_DDAPI_GET_USB_INFO_FROM_SYS_BLOCK | ERROR_USB_DDAPI_DATA_ELEMENT_NOT_FOUND), NULL, pErrorMsg);
            break;
    }
    returnDataInfo->push_front(strResult);

    if(pclose(fp) != 0)
    {
        return returnErrorMsg((ERRORCODE)(ERRIR_USB_DDAPI_GET_USB_INFO_FROM_SYS_BLOCK | ERROR_CLOSE_FILE), NULL, pErrorMsg);
    }

    return NO_ERROR_MSG;
}
ErrorMsg*  USBDDAPI::getUSBSGDeviceNode(UINT32   Index,
										USB_DATA_ELEMENT   DataElement,
										list<string>* returnDataInfo)
{
    CHAR    tempString[256];
    string  strResult;
    DIR    	*pDir = NULL;
    ErrorMsg* pErrorMsg = NULL;
    ERRORCODE errorCode = NO_ERROR;
    struct dirent *pDirent;
    struct stat fileAttribute;

    string USBSGDeviceNode;

    memset(tempString, 0, sizeof(tempString));

    string TempUSBAddress;
    string SysBlockFilePath;
    TempUSBAddress.clear();
    SysBlockFilePath.clear();


    //SysBlockFilePath.append("cat ");
    SysBlockFilePath.append(USB_SYSTEM_BLPCK_PATH);
    SysBlockFilePath.append(USBConfig[Index].DeviceNode);
    SysBlockFilePath.append("/");
    SysBlockFilePath.append(USBElementMap[DataElement].FileName);
    pDir = opendir(SysBlockFilePath.c_str());
	if(pDir == NULL)
	{
	  return returnErrorMsg((ERRORCODE)(ERRIR_USB_DDAPI_GET_USB_SG_DEVICE_NODE | ERROR_USB_DDAPI_OPEN_DIR_FAIL), NULL, pErrorMsg);
	}

    while((pDirent = readdir(pDir)) != NULL)
    {

		CHAR szFileName[256] = "";
		snprintf(szFileName, sizeof(szFileName), "%s%s", SysBlockFilePath.c_str(), pDirent->d_name);

		lstat(szFileName, &fileAttribute);
		if(S_ISDIR(fileAttribute.st_mode))
		{
			if(!strcmp(pDirent->d_name, ".") || !strcmp(pDirent->d_name, ".."))
			{
				continue;
			}
			USBSGDeviceNode.append("/dev/");
			USBSGDeviceNode.append(pDirent->d_name);
			returnDataInfo->push_front(USBSGDeviceNode);
			   // This directory pDirent is another directory
		}
	}
    if(closedir(pDir) != 0)
    {
        return returnErrorMsg((ERRORCODE)(ERRIR_USB_DDAPI_GET_USB_INFO_FROM_SYS_BLOCK | ERROR_CLOSE_FILE), NULL, pErrorMsg);
    }

	return NO_ERROR_MSG;
}




ErrorMsg* USBDDAPI::PREAD(UINT32 fileDescription,
                          UINT32* buffer,
                          UINT32 logicalSector,
                          UINT32 testSize)
{
    //UINT32      blockSize;
    //UINT32      numberOfBlocksToRead;
    SCSI_BLOCK_CDB scsiCDB;
    UINT32      xferBlocks;

    UINT64      lba;

    m_errorCode = NO_ERROR_MSG;
    // clear the SCSI CDB
    memset(&scsiCDB, 0, sizeof(SCSI_BLOCK_CDB));

    lba = logicalSector;

    xferBlocks = testSize/512 ;

    // select the SCSI command based upon the amount of continous data to read
    if (lba & ~0xffffffffULL)

    {

        scsiCDB.cdbType = SCSI_CDB_16_BYTE;
        scsiCDB.cdb.cdb16.opCode = SCSI_OPCODE_READ_16;
        scsiCDB.cdb.cdb16.allocationLengthByte10 = (UINT8)((xferBlocks >> 24) & 0xff);
        scsiCDB.cdb.cdb16.allocationLengthByte11 = (UINT8)((xferBlocks >> 16) & 0xff);
        scsiCDB.cdb.cdb16.allocationLengthByte12 = (UINT8)((xferBlocks >> 8) & 0xff);
        scsiCDB.cdb.cdb16.allocationLengthByte13 = (UINT8)(xferBlocks & 0xff);

        scsiCDB.cdb.cdb16.logicalBlockByte2 = (UINT8)((lba >> 56) & 0xff);
        scsiCDB.cdb.cdb16.logicalBlockByte3 = (UINT8)((lba >> 48) & 0xff);
        scsiCDB.cdb.cdb16.logicalBlockByte4 = (UINT8)((lba >> 40) & 0xff);
        scsiCDB.cdb.cdb16.logicalBlockByte5 = (UINT8)((lba >> 32) & 0xff);
        scsiCDB.cdb.cdb16.additionalCBPByte6 = (UINT8)((lba >> 24) & 0xff);
        scsiCDB.cdb.cdb16.additionalCBPByte7 = (UINT8)((lba >> 16) & 0xff);
        scsiCDB.cdb.cdb16.additionalCBPByte8 = (UINT8)((lba >> 8) & 0xff);
        scsiCDB.cdb.cdb16.additionalCBPByte9 = (UINT8)(lba & 0xff);
    }
    else
    {

        scsiCDB.cdbType = SCSI_CDB_10_BYTE;
        scsiCDB.cdb.cdb10.opCode = SCSI_OPCODE_READ_10;
        scsiCDB.cdb.cdb10.allocationLengthByte7 = (UINT8)((xferBlocks >> 8) & 0xff);
        scsiCDB.cdb.cdb10.allocationLengthByte8 = (UINT8)(xferBlocks & 0xff);

        //printf("allocationLengthByte8 = %d\n", (UINT8)(xferBlocks & 0xff));

        scsiCDB.cdb.cdb10.logicalBlockByte2 = (UINT8)((lba >> 24) & 0xff);
        scsiCDB.cdb.cdb10.logicalBlockByte3 = (UINT8)((lba >> 16) & 0xff);
        scsiCDB.cdb.cdb10.logicalBlockByte4 = (UINT8)((lba >> 8) & 0xff);
        scsiCDB.cdb.cdb10.logicalBlockByte5 = (UINT8)(lba & 0xff);
    }


    m_errorCode = SCSIRead(fileDescription, (UINT8*)&scsiCDB, (UINT8*)buffer, testSize);

    if (m_errorCode != NO_ERROR_MSG)
    {
        return returnErrorMsg(ERRIR_USB_DDAPI_PREAD, NULL, m_errorCode);
    }
/*
    sprintf(m_tempString, "Read logicalSector %llx done\n", logicalSector);
    displayDebugString(DEBUG_LEVEL_2, m_tempString);
*/


    return NO_ERROR_MSG;
}







ErrorMsg* USBDDAPI::PWRITE(UINT32 fileDescription,
                           UINT32* buffer,
                           UINT32 logicalSector,
                           UINT32 testSize)
{
    SCSI_BLOCK_CDB    scsiCDB;
    UINT32      xferBlocks;
    UINT64      lba;

    // clear the SCSI CDB
    memset(&scsiCDB, 0, sizeof(SCSI_BLOCK_CDB));

    lba = logicalSector;

    xferBlocks = testSize/512 ;
    if (lba & ~0xffffffffULL)
    {
        scsiCDB.cdbType = SCSI_CDB_16_BYTE;
        scsiCDB.cdb.cdb16.opCode = SCSI_OPCODE_WRITE_16;
        scsiCDB.cdb.cdb16.allocationLengthByte10 = (UINT8)((xferBlocks >> 24) & 0xff);
        scsiCDB.cdb.cdb16.allocationLengthByte11 = (UINT8)((xferBlocks >> 16) & 0xff);
        scsiCDB.cdb.cdb16.allocationLengthByte12 = (UINT8)((xferBlocks >> 8) & 0xff);
        scsiCDB.cdb.cdb16.allocationLengthByte13 = (UINT8)(xferBlocks & 0xff);

        scsiCDB.cdb.cdb16.logicalBlockByte2 = (UINT8)((lba >> 56) & 0xff);
        scsiCDB.cdb.cdb16.logicalBlockByte3 = (UINT8)((lba >> 48) & 0xff);
        scsiCDB.cdb.cdb16.logicalBlockByte4 = (UINT8)((lba >> 40) & 0xff);
        scsiCDB.cdb.cdb16.logicalBlockByte5 = (UINT8)((lba >> 32) & 0xff);
        scsiCDB.cdb.cdb16.additionalCBPByte6 = (UINT8)((lba >> 24) & 0xff);
        scsiCDB.cdb.cdb16.additionalCBPByte7 = (UINT8)((lba >> 16) & 0xff);
        scsiCDB.cdb.cdb16.additionalCBPByte8 = (UINT8)((lba >> 8) & 0xff);
        scsiCDB.cdb.cdb16.additionalCBPByte9 = (UINT8)(lba & 0xff);
    }
    else
    {
        scsiCDB.cdbType = SCSI_CDB_10_BYTE;
        scsiCDB.cdb.cdb10.opCode = SCSI_OPCODE_WRITE_10;
        scsiCDB.cdb.cdb10.allocationLengthByte7 = (UINT8)((xferBlocks >> 8) & 0xff);
        scsiCDB.cdb.cdb10.allocationLengthByte8 = (UINT8)(xferBlocks & 0xff);

        scsiCDB.cdb.cdb10.logicalBlockByte2 = (UINT8)((lba >> 24) & 0xff);
        scsiCDB.cdb.cdb10.logicalBlockByte3 = (UINT8)((lba >> 16) & 0xff);
        scsiCDB.cdb.cdb10.logicalBlockByte4 = (UINT8)((lba >> 8) & 0xff);
        scsiCDB.cdb.cdb10.logicalBlockByte5 = (UINT8)(lba & 0xff);
    }

    m_errorCode = SCSIWrite(fileDescription, (UINT8*)&scsiCDB, (UINT8*)buffer, testSize);
    if (m_errorCode != NO_ERROR_MSG)
    {
        return returnErrorMsg(ERRIR_USB_DDAPI_PWRITE, NULL, m_errorCode);
    }


    return NO_ERROR_MSG;
}








ErrorMsg* USBDDAPI::SCSIRead(UINT32 fileDescription,
                             UINT8 *cdbP,
                             UINT8 *buffer,
                             UINT32 bufferSize)
{
    int retVal=0;
    SCSI_BLOCK_CDB* scsiCDB;
    scsiCDB = (SCSI_BLOCK_CDB*) cdbP;

    struct sg_io_hdr io_hdr;
    UINT8 cdbSize;

    switch(scsiCDB->cdbType)
    {
        case 0x0:
            cdbSize = 6;
            break;
        case 0x01:
            cdbSize = 10;
            break;
        case 0x02:
            cdbSize = 12;
            break;
        case 0x03:
            cdbSize = 16;
            break;
        default:
            return returnErrorMsg((ERRORCODE)ERROR_USB_DDAPI_SEEK_WRITE,"NULL pointer", NULL);
    }

  //printf("bufferSize=%d\n",bufferSize);

    memset(&io_hdr, 0, sizeof(struct sg_io_hdr));
    io_hdr.interface_id = 'S';
    io_hdr.cmd_len = cdbSize;
    io_hdr.cmdp = (UINT8*)&scsiCDB->cdb;
    io_hdr.dxfer_direction = SG_DXFER_FROM_DEV;
    io_hdr.dxfer_len = bufferSize;
    io_hdr.dxferp = buffer;
    io_hdr.mx_sb_len = SENSE_BUFF_LEN;
    io_hdr.sbp = m_senseBuff;
    io_hdr.timeout = 20000;//DEF_TIMEOUT;
    io_hdr.pack_id = 0;//m_packIdRead++;


    //printf("1:bufferSize=%d\n",bufferSize);
    if ( (retVal=ioctl(fileDescription, SG_IO, &io_hdr)) < 0)
    {
         return returnErrorMsg((ERRORCODE)(ERRIR_USB_DDAPI_SCSI_READ | ERROR_USB_DRIVER_IOCTL_FAIL),"NULL pointer", NULL);
    }


//    if (io_hdr.status != 0/*SCSI_GOOD*/)
//        m_errorCode = (ERRORCODE) (((UINT32) m_senseBuff[0]) << 24 |
//                                   ((UINT32) m_senseBuff[1]) << 16 |
//                                   ((UINT32) m_senseBuff[2]) <<  8 |
//                                   ((UINT32) m_senseBuff[3]));


    return NO_ERROR_MSG;
}






ErrorMsg* USBDDAPI::SCSIWrite(UINT32 fileDescription,
                              UINT8 *cdbP,
                              UINT8 *buffer,
                              UINT32 bufferSize)
{
    //ErrorMsg* pErrorMsg = NULL;


    INT16   retVal;
    struct sg_io_hdr io_hdr;
    UINT8 cdbSize;
    SCSI_BLOCK_CDB*    scsiCDB;

    scsiCDB = (SCSI_BLOCK_CDB*)cdbP;

    switch (scsiCDB->cdbType)
    {
        case 0x0:
            cdbSize = 6;
            break;
        case 0x01:
            cdbSize = 10;
            break;
        case 0x02:
            cdbSize = 12;
            break;
        case 0x03:
            cdbSize = 16;
            break;
        default:
            return returnErrorMsg((ERRORCODE)ERROR_USB_DDAPI_SEEK_WRITE,"NULL pointer", NULL);
    }

    memset(&io_hdr, 0, sizeof(struct sg_io_hdr));
    io_hdr.interface_id = 'S';
    io_hdr.cmd_len = cdbSize;
    //io_hdr.cmdp = cdbP;
    io_hdr.cmdp =(UINT8*)&scsiCDB->cdb;
    io_hdr.dxfer_direction = SG_DXFER_TO_DEV;
    io_hdr.dxfer_len = bufferSize;
    io_hdr.dxferp = buffer;
    io_hdr.mx_sb_len = SENSE_BUFF_LEN;
    io_hdr.sbp = m_senseBuff;
    io_hdr.timeout = DEF_TIMEOUT;
    io_hdr.pack_id = 0;

    while (((retVal = ioctl(fileDescription, SG_IO, &io_hdr)) < 0) && (EINTR == errno));

    if (retVal < 0)
        return returnErrorMsg((ERRORCODE)(ERRIR_USB_DDAPI_SCSI_WRITE | ERROR_USB_DRIVER_IOCTL_FAIL),"NULL pointer", NULL);

/*
//    if (io_hdr.status != SCSI_GOOD)
//        pErrorMsg = (ERRORCODE) (((UINT32) m_senseBuff[0]) << 24 |
//                                   ((UINT32) m_senseBuff[1]) << 16 |
//                                   ((UINT32) m_senseBuff[2]) <<  8 |
//                                   ((UINT32) m_senseBuff[3]));
*/


    return NO_ERROR_MSG;
}








