 /*
  * PMC-Sierra SPC 8001 SAS/SATA based host adapters driver
  *
  * Copyright (c) 2008-2009 USI Co., Ltd.
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions
  * are met:
  * 1. Redistributions of source code must retain the above copyright
  *    notice, this list of conditions, and the following disclaimer,
  *    without modification.
  * 2. Redistributions in binary form must reproduce at minimum a disclaimer
  *    substantially similar to the "NO WARRANTY" disclaimer below
  *    ("Disclaimer") and any redistribution must be conditioned upon
  *    including a substantially similar Disclaimer requirement for further
  *    binary redistribution.
  * 3. Neither the names of the above-listed copyright holders nor the names
  *    of any contributors may be used to endorse or promote products derived
  *    from this software without specific prior written permission.
  *
  * Alternatively, this software may be distributed under the terms of the
  * GNU General Public License ("GPL") version 2 as published by the Free
  * Software Foundation.
  *
  * NO WARRANTY
  * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
  * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
  * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR
  * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
  * HOLDERS OR CONTRIBUTORS BE LIABLE FOR SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
  * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
  * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
  * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
  * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
  * IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
  * POSSIBILITY OF SUCH DAMAGES.
  *
  */

#ifndef PM8001_CTL_H_INCLUDED
#define PM8001_CTL_H_INCLUDED



#ifdef __KERNEL__
#include <linux/miscdevice.h>
#endif


#define PM8001_DEV_NAME                 "pm80xxctl"


#define PM8001_MINOR                    254
#define PM8001_MAGIC_NUMBER             'S'
#define PM8001_IOCTL_DEFAULT_TIMEOUT    (10) /* in seconds */

/**
 * IOCTL opcodes
 */
#define IOCTL_FW_CONTROL              _IOWR(PM8001_MAGIC_NUMBER, 81, int)
#define IOCTL_NVMD_GET                _IOWR(PM8001_MAGIC_NUMBER, 82, int)
#define IOCTL_NVMD_SET                _IOWR(PM8001_MAGIC_NUMBER, 83, int)
#define IOCTL_GET_EVENT_LOG1          _IOWR(PM8001_MAGIC_NUMBER, 84, int)
#define IOCTL_GET_EVENT_LOG2          _IOWR(PM8001_MAGIC_NUMBER, 85, int)
#define IOCTL_GET_CORE_DUMP           _IOWR(PM8001_MAGIC_NUMBER, 86, int)
#define IOCTL_GET_VPD_INFO            _IOWR(PM8001_MAGIC_NUMBER, 87, int)
#define IOCTL_SET_PHY_STATE           _IOWR(PM8001_MAGIC_NUMBER, 88,int)
#define IOCTL_SAS_DIAG_START_END      _IOWR(PM8001_MAGIC_NUMBER, 89,int)
#define IOCTL_SAS_DIAG_EXECUTE        _IOWR(PM8001_MAGIC_NUMBER, 90,int)
#define IOCTL_SMP_CMND_PASSTHROUGH    _IOWR(PM8001_MAGIC_NUMBER, 91,int)

#define IOCTL_BUF_SIZE                  4096
#define HEADER_LEN                      28
#define SIZE_OFFSET                     16

/*
 * IOCTL Status Codes
 */
#define IOCTL_STATUS_OK                 0x00
#define IOCTL_STATUS_MORE_DATA          0x01
#define IOCTL_STATUS_NO_MORE_DATA       0x02
#define IOCTL_STATUS_INVALID_CODE       0x03

/*
 * IOCTL Return Codes
 */
#define IOCTL_CALL_SUCCESS              0x00
#define IOCTL_CALL_FAIL                 0x01
#define IOCTL_CALL_PENDING              0x02
#define IOCTL_CALL_INVALID_CODE         0x03

#define WWNOFFSET                       2052
#define SERIALNO_OFFSET                 2120
#define BIOSOFFSET                      56
#define BIOS_OFFSET_LIMIT               61  
#define FLASH_OK                        0x000000
#define FAIL_OPEN_BIOS_FILE             0x000100
#define FAIL_FILE_SIZE                  0x000a00
#define FAIL_PARAMETERS                 0x000b00
#define FAIL_OUT_MEMORY                 0x000c00
#define FLASH_IN_PROGRESS               0x001000

#define IB_OB_READ_TIMES                256
#define SYSFS_OFFSET                    1024
#define PM80XX_IB_OB_QUEUE_SIZE         32 * 1024
#define PM8001_IB_OB_QUEUE_SIZE         16 * 1024

#define ADPT_IOCTL_CALL_SUCCESS         0x00
#define ADPT_IOCTL_CALL_FAILED          0x01
#define ADPT_IOCTL_CALL_INVALID_CODE    0x03
#define ADPT_IOCTL_CALL_INVALID_DEVICE  0x04
#define ADPT_IOCTL_CALL_TIMEOUT         0x08

/************************************************************
* SGPIO Function and Register type
************************************************************/
#define READ_SGPIO_REGISTER                             0x02
#define WRITE_SGPIO_REGISTER                            0x82

#define SMP_FRAME_REQ                                   0x40
#define SMP_FRAME_RESP                                  0x41

#define SGPIO_CONFIG_REG                                0x0
#define SGPIO_DRIVE_BY_DRIVE_RECEIVE_REG                0x1
#define SGPIO_GENERAL_PURPOSE_RECEIVE_REG               0x2
#define SGPIO_DRIVE_BY_DRIVE_TRANSMIT_REG               0x3
#define SGPIO_GENERAL_PURPOSE_TRANSMIT_REG              0x4

/************************************************************
* SGPIO Function result
************************************************************/
#define SGPIO_COMMAND_SUCCESS                          0x00 
#define SGPIO_CMD_ERROR_WRONG_FRAME_TYPE               0x01
#define SGPIO_CMD_ERROR_WRONG_REG_TYPE                 0x02
#define SGPIO_CMD_ERROR_WRONG_REG_INDEX                0x03
#define SGPIO_CMD_ERROR_WRONG_REG_COUNT                0x04
#define SGPIO_CMD_ERROR_WRONG_FRAME_REG_TYPE           0x05
#define SGPIO_CMD_ERROR_WRONG_FUNCTION                 0x06
#define SGPIO_CMD_ERROR_WRONG_FRAME_TYPE_REG_INDEX     0x19
#define SGPIO_CMD_ERROR_WRONG_FRAME_TYPE_REG_CNT       0x81
#define SGPIO_CMD_ERROR_WRONG_REG_TYPE_REG_INDEX       0x1A
#define SGPIO_CMD_ERROR_WRONG_REG_TYPE_REG_COUNT       0x82
#define SGPIO_CMD_ERROR_WRONG_REG_INDEX_REG_COUNT      0x83
#define SGPIO_CMD_ERROR_WRONG_FRAME_REG_TYPE_REG_INDEX 0x1D
#define SGPIO_CMD_ERROR_WRONG_ALL_HEADER_PARAMS        0x9D

/*********************************************************
 * Device Reset Exapnder
 * ******************************************************/
#define DEV_RESET_EXPANDER                              0x01


struct ioctl_header {
    u32 io_controller_num;
    u32 length;
    u32 return_code;
    u32 timeout;
    u16 direction;
};

struct ioctl_drv_info {
    u8  sz_name[64];
    u16 usmajor_revision;
    u16 usminor_revision;
    u16 usbuild_revision;
    u16 reserved0;
    u32 maxdevices;
    u32 maxoutstandingIO;
    u32 reserved[16];   
};

#define GPIO_READ       0
#define GPIO_WRITE      1
#define GPIO_PINSETUP   2
#define GPIO_EVENTSETUP 3

struct pm8001_gpio {
    u32 operation;
    u32 mask;
    u32 rd_wr_val;
    u32 input_enable;
    u32 pinsetup1;
    u32 pinsetup2;
    u32 event_level;
    u32 event_rising_edge;
    u32 event_falling_edge;
};

#define MAX_SGPIO_REQ_PAYLOAD   12
#define MAX_SGPIO_RESP_PAYLOAD  13

struct read_write_req_resp {
        u8 frame_type; /* =0x40 */
        u8 function; /* 0x02 for read, 0x82 for write */
        u8 register_type;
        u8 register_index;
        u8 register_count;
        u8 function_result;
        u32 read_write_data[MAX_SGPIO_RESP_PAYLOAD];
};

struct sgpio_cfg_0 {
        u8 reserved;
        u8 version:4;
        u8 reserved1:4;
        u8 gp_register_count:4;
        u8 cfg_register_count:3;
        u8 enable:1;
        u8 supported_drive_cnt;
};

struct sgpio_cfg_1{
        u8 reserved;
        u8 blink_gen_a:4;
        u8 blink_gen_b:4;
        u8 max_act_on:4;
        u8 forced_act_off:4;
        u8 stretch_act_on:4;
        u8 stretch_act_off:4;
};

struct ioctl_info_buffer {
    struct ioctl_header header;
    struct ioctl_drv_info   information;
};

struct gpio_buffer {
    struct ioctl_header header;
    struct pm8001_gpio  gpio_payload;
};

struct sgpio_buffer {
        struct ioctl_header     header;
        struct read_write_req_resp  sgpio_req;
};

struct device_info {
    unsigned char iproto;
    unsigned char tproto;
    unsigned char sas_address[8];
    unsigned char linkrate;
    unsigned int  phy_id;
    unsigned int  luns;
}__attribute__((packed));

struct phy_profile {
        char phy_id;
        unsigned int phys:4;
        unsigned int nlr:4;
        unsigned int plr:4;
        unsigned int reserved1:12;
        unsigned char port_id;
        unsigned int prts:4;
        unsigned int reserved2:20;
}__attribute__((packed));

struct phy_errcnt{
        unsigned int  InvalidDword;
        unsigned int  runningDisparityError;
        unsigned int  codeViolation;
        unsigned int  LossOfSyncDW;
        unsigned int  phyResetProblem;
        unsigned int  inboundCRCError;
};

struct phy_prof_resp {
    union {
        struct phy_profile status;
        struct phy_errcnt errcnt;
    }phy;
};

struct tmf_ioctl {
    unsigned long long sas_address;;
    unsigned int lun_number;
    unsigned int iomb_status; 
    unsigned int resp_code;
};

#define ADPT_IOCTL_INFO             _IOR(ADPT_MAGIC_NUMBER, 0, struct ioctl_info_buffer *) 
#define ADPT_IOCTL_GPIO             _IOWR(ADPT_MAGIC_NUMBER, 1, struct  gpio_buffer *)
#define ADPT_IOCTL_SGPIO            _IOWR(ADPT_MAGIC_NUMBER, 2, struct sgpio_buffer *)
#define ADPT_IOCTL_SCAN             _IO(ADPT_MAGIC_NUMBER, 3)
#define ADPT_IOCTL_RESET_LUN        _IOW(ADPT_MAGIC_NUMBER, 4, struct tmf_ioctl *)
#define ADPT_IOCTL_RESET_DEVICE     _IOW(ADPT_MAGIC_NUMBER, 5 ,u8 *)
#define ADPT_IOCTL_LIST_DEVICES     _IOWR(ADPT_MAGIC_NUMBER, 6, unsigned char *)
#define ADPT_IOCTL_SOFT_RESET       _IOW(ADPT_MAGIC_NUMBER, 7, unsigned long)
#define ADPT_IOCTL_GET_PHY_PROFILE _IOWR(ADPT_MAGIC_NUMBER, 8, struct phy_profile*)
#define ADPT_IOCTL_GET_PHY_ERR_CNT _IOWR(ADPT_MAGIC_NUMBER, 9, struct phy_err*)
#define ADP80XX_SMP_COMMAND         0x6
#define ADPT_COMMAND                _IOWR(ADPT_MAGIC_NUMBER,20,struct adpt_ioctl_command)

/*
 * IOCTLs to support sldiag like functionalities
 */
#define ADPT_IOCTL_DIAG_LOOPBACK       		_IOWR(ADPT_MAGIC_NUMBER, 21, int)
#define ADPT_IOCTL_DIAG_INTERNAL_LOOPBACK     	_IOWR(ADPT_MAGIC_NUMBER, 22, int)
#define ADPT_IOCTL_DIAG_EXTERNAL_LOOPBACK     	_IOWR(ADPT_MAGIC_NUMBER, 23, int)
#define ADPT_IOCTL_SAS_CONNECTION_STATE		_IOWR(ADPT_MAGIC_NUMBER, 24, int)
#define ADPT_IOCTL_DIAG_ENABLE			_IOWR(ADPT_MAGIC_NUMBER, 25, int)
#define ADPT_IOCTL_DIAG_DISABLE			_IOWR(ADPT_MAGIC_NUMBER, 26, int)
#define ADPT_IOCTL_ADAPTER_REVISION             _IOW(ADPT_MAGIC_NUMBER, 27, u8 *)
#define ADPT_IOCTL_GET_VPD                      _IOR(ADPT_MAGIC_NUMBER, 28, pmcsas_VPD_t *)

#define ADPT_IOCTL_START_STOP_PHY_DIAG		_IOWR(ADPT_MAGIC_NUMBER, 50, int)
#define ADPT_IOCTL_START_PHY_NORMAL		_IOWR(ADPT_MAGIC_NUMBER, 51, int)
#define ADPT_IOCTL_STOP_PHY_NORMAL		_IOWR(ADPT_MAGIC_NUMBER, 52, int)

#define ADPT_MAGIC_NUMBER           'm'

typedef struct _adpt_ioctl_header {
    unsigned int     iocnum;    /* IOC unit number */
    unsigned int     port;      /* IOC port number */
    int      maxDataSize;   /* Maximum Num. bytes to transfer on read */
} adpt_ioctl_header;


struct adpt_ioctl_command {
    adpt_ioctl_header hdr;
    int     timeout;    /* optional (seconds) */
    char        __user *replyFrameBufPtr;
    char        __user *dataInBufPtr;
    char        __user *dataOutBufPtr;
    char        __user *senseDataPtr;
    int     maxReplyBytes;
    int     dataInSize;
    int     dataOutSize;
    int     maxSenseBytes;
    int     dataSgeOffset;
    char        MF[1];
};

#ifndef MPI_POINTER
#define MPI_POINTER     *
#endif

typedef struct _SGE_SIMPLE_UNION
{
    u32                     FlagsLength;
    union
    {
        u32                 Address32;
        u64                 Address64;
    }u;
} SGE_SIMPLE_UNION, MPI_POINTER PTR_SGE_SIMPLE_UNION,
  SGESimpleUnion_t, MPI_POINTER pSGESimpleUnion_t;

typedef struct _MSG_SMP_PASSTHROUGH_REQUEST
{
    u8                      PassthroughFlags;   /* 00h */
    u8                      PhysicalPort;       /* 01h */
    u8                      ChainOffset;        /* 02h */
    u8                      Function;           /* 03h */
    u16                     RequestDataLength;  /* 04h */
    u8                      ConnectionRate;     /* 06h */
    u8                      MsgFlags;           /* 07h */
    u32                     MsgContext;         /* 08h */
    u32                     Reserved1;          /* 0Ch */
    u64                     SASAddress;         /* 10h */
    u32                     Reserved2;          /* 18h */
    u32                     Reserved3;          /* 1Ch */
    SGE_SIMPLE_UNION        SGL;                /* 20h */
} MSG_SMP_PASSTHROUGH_REQUEST, MPI_POINTER PTR_MSG_SMP_PASSTHROUGH_REQUEST,
  SmpPassthroughRequest_t, MPI_POINTER pSmpPassthroughRequest_t;

struct PM8001_SMP_PASSTHROUGH_REQUEST
{
    u8                      PassthroughFlags;   /* 0x00 */
    u8                      PhysicalPort;       /* 0x01 */
    u8                      ChainOffset;        /* 0x02 */
    u8                      Function;           /* 0x03 */
    u16                     RequestDataLength;  /* 0x04 */
    u8                      SGLFlags;           /* 0x06 */
    u8                      MsgFlags;           /* 0x07 */
    u8                      VP_ID;              /* 0x08 */
    u8                      VF_ID;              /* 0x09 */
    u16                     Reserved1;          /* 0x0A */
    u32                     Reserved2;          /* 0x0C */
    u64                     sas_addr;           /* 0x10 */
    u32                     Reserved3;          /* 0x18 */
    u32                     Reserved4;          /* 0x1C */
    u8                      SGL;                /* 0x20 */
};

struct smp_pass_through_req
{
 u8 exp_sas_addr[8];
 u8 *smp_req;
 u8 *smp_resp;
 u32 smp_req_len;
 u32 smp_resp_len;
};

/************************************************
 *     TYPEDEFS AND STRUCTURES - NetApp
 *     *************************************************/
typedef struct {
    uint32_t       magic_number;                        /* Big-endian */
    uint8_t        base_sas_address[8];
    uint8_t        manufacturer[16];
    uint8_t        manufacturer_part_number[16];
    uint8_t        manufacturer_revision_number[16];
    uint8_t        serial_number[16];
    uint8_t        date_code[16];
    uint8_t        netapp_part_number[16];
    uint8_t        netapp_revision_number[16];
    uint32_t       checksum;                            /* Big-endian */
} pmcsas_VPD_t;

#define PORT_CONNECTION_STATE		32

#define DIAG_IOCTL_SUCCESS                    	0x00
#define DIAG_IOCTL_BUSY                    	0x01
#define DIAG_IOCTL_INVALID_PARM             	0x02
#define DIAG_IOCTL_INVALID_STATE            	0x03
#define DIAG_IOCTL_NOT_ENABLED              	0x04
#define DIAG_IOCTL_INVALID_REQUEST          	0x05
#define DIAG_IOCTL_LOOPBACK_ERROR			0x06
#define DIAG_IOCTL_FAILED          			0x07

typedef struct diag_ioctl_status {
	u8    valid;      /* if this is 1, other fields are valid */
	u8    status;	/* diag ioctl status 		*/
	u8    res0;		/* not used				*/
	u8    res1;		/* not used				*/
	u32   info;       /* diag specific 			*/
} diag_ioctl_status_t;

typedef struct loopback_enable_status {
	u32 port_num;
	u32 loopback_status;
}loopback_enable_status_t;

typedef struct sas_connection_state {
	u32 port_num;
	char state[PORT_CONNECTION_STATE];
	diag_ioctl_status_t	diag;
}sas_connection_state_t;

typedef struct loopback_test_status {
	u32 port_num;
	u32 loopback_status;
}loopback_test_status_t;

typedef struct diag_ioctl	{
	diag_ioctl_status_t	diag_status;
	union {
		loopback_enable_status_t lp_status;
		sas_connection_state_t	sas_state;
	};
} diag_ioctl_t;

#endif /* PM8001_CTL_H_INCLUDED */

